
#include "Game.H"
#include "GfxModule.H"
#include "GameStateModule.H"

#include "_ResGatherBrain.H" // used brain implementation
#include "_EventHandler.H"


blazi::EventHandler::EventHandler( GameState &state_ )
    : state(state_)
{
    brain = new Brain( state.gsm->get_game() );
}


blazi::EventHandler::~EventHandler()
{
    delete brain;
}


bool blazi::EventHandler::handle_event( const Event &e_ )
{
    if( e_.get_who() == GameStateModule::FROM )
    {
        if( e_.get_what() == GameStateModule::VIEW_MSG )
        {
            brain->compute_actions();
            state.gsm->send_actions();

            if( state.gfxm )
            {
                state.gfxm->process_changes();

                const Game &game = state.gsm->get_game();
                sint4 vf = game.get_view_frame();
                sint4 af = game.get_action_frame();

                if( abs( vf - af ) < 10 )
                {
                    state.gfxm->draw();
                    state.just_drew = true;
                }
            }

            return true;
        }

        if( e_.get_what() == GameStateModule::STOP_MSG )
        {
            state.quit = true;
            return true;
        }

        if( e_.get_what() == GameStateModule::READ_ERROR_MSG )
        {
            state.quit = state.error = true;
            return true;
        }
    }

    return false;
}

