
#ifndef _EVENTHANDLER_H
#define _EVENTHANDLER_H

#include "EventHandler.H"

class Event;


namespace blazi {

class Brain;


/** Representation of the game state
 *  maintained by an event handler. */
struct GameState
{
    GameStateModule *gsm;
    GfxModule *gfxm;        // != NULL means gfx is used

    bool just_drew;         // just drew a gfx frame
    bool quit;              // true => quit game
    bool error;             // true <=> error occured

    GameState( GameStateModule *gsm_, GfxModule *gfxm_ )
        : gsm(gsm_), gfxm(gfxm_)
    {
        assert( gsm );
        just_drew = quit = error = false;
    }
};


/** Responsible for the communication between the
 *  server and the client. Receiving actions and
 *  using the "brain" to find out the response. */
class EventHandler : public ::EventHandler
{
public:

    EventHandler( GameState &state_ );
    ~EventHandler();

    bool handle_event( const Event &e_ );

private:

    GameState &state;
    Brain *brain;
};


} /* namespace blazi */


#endif /* _EVENTHANDLER_H */

