
#include "Game.H"

#include "_Terrain.H"
#include "_Obj.H"

#include <set>


std::set<const blazi::Obj*> blazi::Mineral::_objs;


sint4 blazi::Worker::MAX_SAFETY = 40;

const std::string blazi::Worker::moveAction( "move" );
const std::string blazi::Worker::miningComponent( "pickaxe" );
const std::string blazi::Worker::miningAction( "mine" );
const std::string blazi::Worker::returnResourcesAction( "return_resources" );


bool blazi::Mineral::has_free_space() const
{
    assert( 0 <= _workers_num && _workers_num <= MAX_WORKERS );
    return MAX_WORKERS != _workers_num;
}


bool blazi::Mineral::reserve( const Obj *obj_ )
{
    DEBUG( "RESM " << obj_ );
    assert( _objs.insert(obj_).second );

    if( !has_free_space() ) return false;
    assert( _workers_num < MAX_WORKERS );
    ++_workers_num;
    assert( _workers_num <= MAX_WORKERS );
    return true;
}


void blazi::Mineral::release( const Obj *obj_ )
{
    DEBUG( "RELM " << obj_ );
    assert( _objs.erase(obj_) );

    assert( 0 < _workers_num );
    --_workers_num;
    assert( 0 <= _workers_num );
}


blazi::Worker::Worker( const Terrain &terrain_, GameObj *obj_ )
    : Obj(obj_, true), _terrain(terrain_)
{
    _to = loc();
    _path = NULL;
    _target = NULL;

    _prevState = _state = Idle;

    safety = MAX_SAFETY;

    dodge_left = false;
    dodge_right = false;
    dodge_wait = MAX_DODGE_WAIT;
    dodge_near_to_target = MAX_DODGE_NEAR_TO_TARGET;
}


blazi::Worker::~Worker()
{
    if( _path ) delete _path;
}


void blazi::Worker::set_target( Obj *obj_ )
{
    if( target() == obj_ ) return;
    _oldMineral = mineralTarget();
    _target = obj_;
}


void blazi::Worker::set_state( State state_ )
{
    DEBUG( "STATEchg " << this << "  " << _state << " -> " << state_ );

    if( Idle == state_ )
    {
        safety = MAX_SAFETY;
        dodge_near_to_target = MAX_DODGE_NEAR_TO_TARGET;
    }

    if( Mining != state_ ) _oldMineral = NULL;

    _prevState = _state;
    _state = state_;
}


void blazi::Worker::move( Path *path_, bool update_ )
{
    set_state( Moving );

    DEBUG( " muve? " << update_ );

    _to = loc();
    if( _path && _path != path_ ) delete _path;
    _path = path_;

    if( update_ ) update(); // start moving
}


void blazi::Worker::mine( ObjId mineralId_ )
{
    set_state( Mining );

    Vector<sint4> params( 1, mineralId_ );
    obj()->component( miningComponent )->set_action( miningAction, params );
}


void blazi::Worker::return_resources( ObjId ccId_ )
{
    set_state( ReturnResources );

    Vector<sint4> params( 1, ccId_ );
    obj()->set_action( returnResourcesAction, params );
}


void blazi::Worker::update()
{
    DEBUG( "          ub " << state()  );

    if( Mining != state() && 0 == --safety )
    {
        DEBUG( "SAFETY EXPIRED << " << state() );
        run_away( false );

        if( Mineral *mineral = mineralTarget() )
        {
            mineral->release( this );
            set_target( NULL );
        }
        
        set_state( Idle );
        return;
    }

    switch( state() )
    {
        case Idle :
        case Blocked :
        {
            break;
        }

        case Moving :
        {
            if( target() )
            {
                DEBUG( "worker: move to " << target()->obj()->bp_name()
                       << " " << target()->loc() << "  " << this );

                if( Game::distance( *obj(), *target()->obj() ) <= Worker::RANGE )
                {
                    if( path() )
                    {
                        delete _path;
                        _path = NULL;
                    }

                    DEBUG( "i am arrived, loc: " << loc() );

                    set_state( Idle );
                    break;
                }
            }

            if( !path() )
            {
                if( stopped() )
                {
                    if( 0 < dodge_near_to_target && target() &&
                        Game::distance( *obj(), *target()->obj() ) > RANGE )
                    {
                        DEBUG( "dodge near target: "
                               << Game::distance( *obj(), *target()->obj() ) );

                        Path *path = new Path();
//                        path->push_front( target()->loc() );
                        path->push_front( _terrain.min_realobjloc( loc(), target() ) );

                        --dodge_near_to_target;
                        move( path, false );
                        dodge();
                    }
                    else
                    {
                        DEBUG( "stopped with empty path, loc: " << loc() );
                        if( Mineral *mineral = mineralTarget() )
                        {
                            mineral->release( this );
                            set_target( NULL );
                        }

                        set_state( Idle );
                    }
                }
                break;
            }

            if( !target() )
            {
                if( stopped() ) set_state( Idle );
                break;
            }

            if( !path()->empty() &&
                !_terrain.check_free_pass( this,
                                           loc(),
                                           path()->front(),
                                   //        Loc::avg( loc(), path()->front() ),
                                           true ) )
            {
                DEBUG( "blocked fwd, locs: " << loc()
                       << " [" << _terrain.real2pos(loc()) << "] ? "
                       << path()->front() << " ["
                       << _terrain.real2pos(loc()) );

                set_state( Blocked );
                break;
            }

            if( path()->empty() )
            {
                delete _path;
                _path = NULL;

                if( target() &&
                    Game::distance( *obj(), *target()->obj() ) < REACH_RADIUS )
                {
                    DEBUG( "move to target" );
    //                move( target()->loc() );
                    move( Loc::avg( target()->loc(),
                                    _terrain.min_realobjloc( loc(), target() ) ) );
                }
                else
                {
                    if( Mineral *mineral = mineralTarget() )
                    {
                        mineral->release( this );
                        set_target( NULL );
                    }

                    set_state( Idle );
                }

                break;
            }

            DEBUG( "    moving , loc = " << loc() << " to " << _to );
            if( Loc::distance( loc(), _to ) > RANGE )
            {
                if( stopped() )
                {
                    DEBUG( "  blocked far, loc: " << loc()
                           << " , to: " << _to
                           << " , d = " << Loc::distance( loc(), _to ) );

     //               if( path()->empty() ||
     //                   path()->front() != _to ) path()->push_front( _to );

                    dodge();
          //          set_state( Blocked );
                }
       //         else DEBUG( "  keep walking, loc: " << loc() );

                break;
            }

            dodge_left = false;
            dodge_right = false;
            dodge_wait = MAX_DODGE_WAIT;

            _prev = _to;
            _to = path()->front();
            path()->pop_front();
            DEBUG( "move NEXT " << loc() << " -> " << _to );
            move( _to );

            break;
        }

        case Mining :
        {
            assert( _oldMineral );

            if( 0 == obj()->component( "pickaxe" )->get_int( "active" ) )
            {
                // move away a little from the mineral

                _oldMineral->release( this );
                move( Loc::away( _oldMineral->loc(), loc(), obj()->get_radius() ) );

                _oldMineral = NULL;
                set_state( Moving );
          //      set_state( Idle );
            }
            break;
        }

        case ReturnResources :
        {
            if( !obj()->is_pending_action() )
            {

                if( ControlCenter *cc = ccTarget() )
                {
                    move( Loc::away( cc->loc(), loc(), obj()->get_radius() ) );
                }
                else move( _to );

                set_state( Moving );
        //        set_state( Idle );
            }
            break;
        }
    }

    DEBUG( "          ue " << state()  );
}


inline bool blazi::Worker::stopped() const
{
    return !obj()->is_pending_action() &&
           0 == obj()->get_speed( NULL );
}


void blazi::Worker::move( const Loc &loc_ )
{
    DEBUG( "     move command to " << loc_ );
    Vector<sint4> params;
    params.push_back( loc_.x );
    params.push_back( loc_.y );
    params.push_back( obj()->get_max_speed() );

    obj()->set_action( moveAction, params );
}


void blazi::Worker::run_away( bool update_ )
{
    Terrain::PosSet ps;
    _terrain.fill_pos_set( ps,
                           _terrain.get_tile_points() / 2,
                           loc() );
    Loc avg(0,0);
    sint4 num = 0;
    FORALL( ps, pit )
    {
        FORALL( (*pit)->objs(), oit )
        {
            if( *oit != this )
            {
                DEBUG( " - " << (*oit)->loc()
                        << "  " << *oit );

                avg += (*oit)->loc();
                ++num;
            }
        }
    }

    Loc aw;
    if( num )
    {
        avg.x /= num;
        avg.y /= num;

        aw = Loc::away( avg, loc(), obj()->get_radius() );

        DEBUG( "avg: " << avg << ", loc: " << loc() << ", aw: " << aw );
    }

    if( num && loc() != aw )
    {
        Path *path = new Path();
        path->push_front( aw );
        move( path, update_ );
    }
    else
    {
        if( Mineral *mineral = mineralTarget() )
        {
            mineral->release( this );
            set_target( NULL );
        }
        
        set_state( Idle );
    }
}


void blazi::Worker::dodge()
{
    assert( path() );
    assert( target() );

    Loc to = target()->loc();
    if( !path()->empty() ) to = path()->front();

    DEBUG( "------ dodge " << dodge_wait << "  " << this );
    Loc dp = _terrain.dodge_point( this, _to, to );

    if( dp.valid() )
    {
        DEBUG( "dp: " << dp << " , to: " << to << " , _to: " << _to );

//        if( Loc::distance( to, _to ) < Loc::distance( to, dp ) )
//            path()->push_front( _to );

//        _path->push_front( dp );

        _to = dp;
        move( dp );
        set_state( Moving );
    }
    else
    if( 0 == dodge_wait )
    {
        run_away( false );

        if( Mineral *mineral = mineralTarget() )
        {
            mineral->release( this );
            set_target( NULL );
        }
        
        set_state( Idle );
    }

    if( 0 < dodge_wait ) --dodge_wait;

    /*
    else
    {
        DEBUG( "NO DODGE" );
        set_state( Blocked );
    }
    */
}

