
#ifndef _OBJ_H
#define _OBJ_H

#include <string>

#include "Action.H"
#include "_Loc.H"


namespace blazi {

class Path;
class Terrain;


/** Superclass of all the objects in the game. */
class Obj
{
public:

    Obj( GameObj *obj_, bool mobile_ ) : _obj(obj_), _mobile(mobile_) {}
    virtual ~Obj() {}

    GameObj* obj() { return _obj; }
    const GameObj* obj() const { return _obj; }

    Loc loc() const { return Loc( *obj()->sod.x, *obj()->sod.y ); }
    Loc locTP() const { return Loc( *obj()->sod.x1, *obj()->sod.y1 ); }
    Loc locBR() const { return Loc( *obj()->sod.x2, *obj()->sod.y2 ); }

    bool mobile() const { return _mobile; }

private:

    GameObj *_obj;
    bool _mobile;
};


class Mineral : public Obj
{
public:
    static const sint4 MAX_WORKERS = 1;

public:

    Mineral( GameObj *obj_ ) : Obj(obj_, false) { _workers_num = 0; }

    bool has_free_space() const;
    bool reserve( const Obj *obj_ );
    void release( const Obj *obj_ );

    static const sint4 INIT_DIST_BASE = 100000;
    static const sint4 INIT_RATE = 10;

    Loc loc_from_cc;
    sint4 dist_from_cc;

private:

    static std::set<const Obj*> _objs;

    sint4 _workers_num;
};


class Sheep : public Obj
{
public:

    Sheep( GameObj *obj_ ) : Obj(obj_, true) {}
};


class ControlCenter : public Obj
{
public:

    ControlCenter( GameObj *obj_ ) : Obj(obj_, false) {}
};


class Worker : public Obj
{
public:

    static const std::string moveAction;
    static const std::string miningComponent;
    static const std::string miningAction;
    static const std::string returnResourcesAction;

    static const sint4 RANGE = 2;
    static const sint4 REACH_RADIUS = 30;

    enum State { Idle, Moving, Blocked, Mining, ReturnResources };

public:

    Worker( const Terrain &terrain_, GameObj *obj_ );
    ~Worker();

    State state() const { return _state; }
    State prevState() const { return _prevState; }
    void set_state( State state_ );

    Path* path() { return _path; }
    const Path* path() const { return _path; }

    Obj* target() { return _target; }
    const Obj* target() const { return _target; }
    void set_target( Obj *obj_ );

    ControlCenter* ccTarget()
                   { return dynamic_cast<ControlCenter*>( target() ); }
    const ControlCenter* ccTarget() const
                         { return dynamic_cast<const ControlCenter*>( target() ); }
    Mineral* mineralTarget()
             { return dynamic_cast<Mineral*>( target() ); }
    const Mineral* mineralTarget() const
                   { return dynamic_cast<const Mineral*>( target() ); }

    void dodge();
    void move( Path *path_, bool update_ = true );
    void mine( ObjId mineralId_ );
    void return_resources( ObjId ccId_ );
    void run_away( bool update_ = true );

    void update();

    // dodge variables
    mutable bool dodge_left;
    mutable bool dodge_right;

    static const sint4 MAX_DODGE_WAIT = 2;
    mutable sint4 dodge_wait;

    static const sint4 MAX_DODGE_NEAR_TO_TARGET = 1;
    mutable sint4 dodge_near_to_target;

protected:

    static sint4 MAX_SAFETY;
    sint4 safety;

    bool stopped() const;
    void move( const Loc &loc_ );

private:

    const Terrain &_terrain;

    Loc _prev;
    Loc _to;
    Path *_path;
    Obj *_target;
    Mineral *_oldMineral;

    State _state;
    State _prevState;
};


} /* namespace blazi */


#endif /* _OBJ_H */

