
#ifndef _PATHFIND_H
#define _PATHFIND_H

#include "_Terrain.H"


namespace blazi {


/** Support of path finding between two locations on a specified terrain. */
class PathFind
{
public:

    typedef Terrain::Distance Distance;

    struct Result
    {
        Result( bool success_ = false, Path *path_ = NULL )
            : success(success_), path(path_) {}

        bool success;
        Path *path;
    };

public:

    PathFind( const Terrain &terrain_ );

    const Terrain& terrain() const { return _terrain; }

    /** A* searches. */

    Result search_path( const Obj *obj_,
                        const Loc &from_,
                        const Loc &to_,
                        Distance freeCheckRadius_ = 0 ) const;

    Result search_path( const Obj *from_,
                        const Obj *to_,
                        Distance freeCheckRadius_ = 0,
                        const Obj *checkedObj_ = NULL ) const;

protected:

    typedef Terrain::ObjLocations LocSeq;

    Result search_engine( const Obj *obj_,
                          const LocSeq &from_,
                          const LocSeq &to_,
                          Distance freeCheckRadius_ ) const;

private:

    const Terrain &_terrain;
};


} /* namespace blazi */


#endif /* _PATHFIND_H */

