
#include "Game.H"
#include "GameObj.H"
#include "Options.H"

#include "_Global.H"
#include "_Obj.H"
#include "_Terrain.H"
#include "_PathFind.H"
#include "_ResGatherBrain.H"


blazi::Worker *myman, *myman2;


blazi::Brain::Brain( const Game &game_ )
    : _game(game_)
{
     DEBUG( "map size: " << get_map().get_width() << "x"
                         << get_map().get_height() );

     DEBUG( "tile_points: " << get_tile_points() );

     _terrain = new Terrain( get_map(), get_tile_points() );
     _pf = new PathFind( *_terrain );

     _cc = NULL;
}


blazi::Brain::~Brain()
{
    delete _pf;
    delete _terrain;

    if( _cc ) delete _cc;
    while( !_minerals.empty() ) { delete _minerals.back(); _minerals.pop_back(); }
    while( !_workers.empty() ) { delete _workers.back(); _workers.pop_back(); }
    while( !_sheeps.empty() ) { delete _sheeps.back(); _sheeps.pop_back(); }
}


void blazi::Brain::compute_actions()
{
    if( !_cc )
    {
        const Game::ObjCont &my_objs = _game.get_objs( _game.get_client_player() );
        FORALL( my_objs, it )
        {
            GameObj  *gob = (*it)->get_GameObj();
            if( !gob ) continue;

            if( "controlCenter" == gob->bp_name() )
            {
                assert( !_cc ); // only one cc can exist now

                DEBUG( "cc tp : (" << *gob->sod.x1 << "," << *gob->sod.y1 << ")" );
                DEBUG( "cc br : (" << *gob->sod.x2 << "," << *gob->sod.y2 << ")" );

                _cc = new ControlCenter( gob );
                terrain()->update_obj( _cc );
            }
            else
            if( "worker" == gob->bp_name() )
            {
                _workers.push_back( new Worker( *terrain(), gob ) );
                // terrain is updated later
            }
            else DEBUG( "ignored gob: " << gob->bp_name() );
        }

        // return if the info is not available yet
        if( !_cc ) return;

        const Game::ObjCont &neutral_objs = _game.get_objs( _game.get_player_num() );
        FORALL( neutral_objs, it )
        {
            GameObj *gob = (*it)->get_GameObj();
            if( !gob ) continue;

            if( "mineral" == gob->bp_name() )
            {
                Mineral *mineral = new Mineral( gob );
                _minerals.push_back( mineral );
                terrain()->update_obj( mineral );
            }
            else
            if( "sheep" == gob->bp_name() )
            {
                _sheeps.push_back( new Sheep( gob ) );
                // terrain is updated later
            }
            else DEBUG( "ignored gob: " << gob->bp_name() );
///*
            // updating mineral info
            FORALL( _minerals, it )
            {
                (*it)->loc_from_cc = Loc(); // undefined loc shows guess distance
                (*it)->dist_from_cc = Mineral::INIT_DIST_BASE +
                                      terrain()->distance( _cc, *it );

            }
//*/
        }
    }

    DEBUG( "=======new-round===========" );

    // collapsing, vanishing, ... should be handled

    // balanced update of the exact mineral distances
    {
        static std::list<Mineral*>::iterator init_mit = _minerals.begin();
        if( _minerals.end() != init_mit )
        {
            assert( !(*init_mit)->loc_from_cc.valid() );

            static sint4 rate_counter = Mineral::INIT_RATE;

            if( 0 == --rate_counter )
            {
                PathFind::Result res = _pf->search_path( _cc,
                                                         *init_mit,
                                                         0,
                                                         *(_workers.begin()) );
                if( res.success )
                {
                    DEBUG( "mineral updated, dist: " << res.path->length() );

                    (*init_mit)->loc_from_cc = res.path->front();
                    (*init_mit)->dist_from_cc = res.path->length();
                }
                else DEBUG( "unreachable mineral" );

                ++init_mit;
                rate_counter = Mineral::INIT_RATE;
            }
        }
    }

    // terrain updates for mobile objects
    FORALL( _workers, it ) terrain()->update_obj( *it );
    FORALL( _sheeps , it ) terrain()->update_obj( *it );

    // update all workers
    FORALL( _workers, it )
    {
//        if( myman != *it && myman2 != *it ) continue;
//        DEBUG( ".. " << *it );
        (*it)->update();
    }

    // set new actions

    PathFind::Result res;

//    myman = *(_workers.begin());
//    myman2 = *(++(++(++(++_workers.begin()))));

    FORALL( _workers, it )
    {
        Worker *worker = *it;

//        if( myman != worker && myman2 != worker ) continue;

//        DEBUG( "-------------------- " << worker
//               << " (" << worker->state() << "         my: " << ((worker != myman)+1) );

        switch( worker->state() )
        {
            case Worker::Idle :
            {
                if( worker->target() )
                {
                    if( Mineral *mineral = worker->mineralTarget() )
                    {
                        DEBUG( "target = mineral" );
                        if( Game::distance( *worker->obj(),
                                            *mineral->obj() ) <= Worker::RANGE )
                        {
                            PlayerInfo &pi = _game.get_cplayer_info();
                            worker->set_target( _cc );
                            worker->mine( pi.get_id( mineral->obj() ) );
                        }
                        else
                        {
                            // blocked close to the mineral
                            DEBUG( "cannot reach mineral: "
                                   << Game::distance( *worker->obj(),
                                                      *mineral->obj() ) );

                            mineral->release( worker );
                            worker->set_target( NULL );
                            worker->set_state( Worker::Idle );
                        }
                    }
                    else
                    if( ControlCenter *cc = worker->ccTarget() )
                    {
                        DEBUG( "target = cc, d = "
                               << Game::distance( *worker->obj(), *cc->obj() ) );

                        if( Game::distance( *worker->obj(),
                                            *cc->obj() ) <= Worker::RANGE )
                        {
                            PlayerInfo &pi = _game.get_cplayer_info();
                            worker->return_resources( pi.get_id( cc->obj() ) );
                            worker->set_target( NULL );
                        }
                        else
                        {
                            DEBUG( "PF: cc" );
                            res = _pf->search_path( worker, _cc, PF_RADIUS );
                            if( res.success )
                            {
                                DEBUG( "move to cc" );
                                worker->move( res.path );
                            }
                            else
                            {
                                DEBUG( "unreachable cc" );
                                worker->run_away();
                            }
                        }
                    }
                    else ERR( "illegal target: " );
                }
                else
                {
                   choose_mineral( worker );
                }
                break;
            }

            case Worker::Blocked :
            {
                path_slicing( worker );
                break;
            }

            default : break; // just carry on
        }
    }
}


void blazi::Brain::choose_mineral( Worker *worker_ )
{
    assert( worker_ );
/*
    int idx;
    if( myman == worker_ )
    {
        Options::get( "-dmidx", idx );
    }
    else
    {
        Options::get( "-dmidx2", idx );
    }
    std::list<Mineral*>::iterator mi = _minerals.begin();
    for( int i = 0; i < idx; ++i ) ++mi;

    DEBUG( "chosen mineral " << *mi << "  " << (myman != worker_)+1 );

    PathFind::Result res = _pf->search_path( worker_,
                                             terrain()->posloc( worker_ ),
                                             terrain()->posloc( *mi ),
                                             PF_RADIUS );
    if( res.success )
    {
        if( (*mi)->reserve() )
        {
            DEBUG( "worker: goto mineral: " << (*mi)->loc() );
            worker_->set_target( *mi );
            worker_->move( res.path );
        }
    }
    else worker_->run_away();
    return;
*/
///*
    std::map<sint4, Mineral*> ordered;

    if( Game::distance( *worker_->obj(), *_cc->obj() ) > PF_CC_MINERAL )
    {
        FORALL( _minerals, mit )
        {
            if( (*mit)->has_free_space() )
            {
                ordered.insert(
                    std::make_pair(
                        terrain()->distance( worker_, *mit ),
                        *mit ) );
            }
        }
    }
    else
    {
        FORALL( _minerals, mit )
        {
            if( (*mit)->has_free_space() )
            {
                sint4 dist_around_cc =
                    std::min( Loc::hamilton_distance( worker_->loc(),
                                                      _cc->locTP() )
                              + Loc::hamilton_distance( _cc->locTP(),
                                                        (*mit)->loc_from_cc ),

                              Loc::hamilton_distance( worker_->loc(),
                                                      _cc->locBR() )
                              + Loc::hamilton_distance( _cc->locBR(),
                                                        (*mit)->loc_from_cc ) );

                ordered.insert(
                    std::make_pair( dist_around_cc + (*mit)->dist_from_cc, *mit ) );
            }
        }
    }

    assert( !ordered.empty() );

    PathFind::Result res;
    FORALL( ordered, mit )
    {
        DEBUG( "PF: mineral" );
        res = _pf->search_path( worker_,
                                terrain()->posloc( worker_ ),
                                terrain()->posloc( mit->second ),
                                PF_RADIUS );
        if( res.success )
        {
            if( mit->second->reserve( worker_ ) )
            {
                DEBUG( "worker: goto mineral , dist = " << mit->first  );
                worker_->set_target( mit->second );
                worker_->move( res.path );
                return;
            }
        }
    }

    // no free mineral has been found ??
    worker_->run_away();
//*/
}


void blazi::Brain::path_slicing( Worker *worker_ )
{
    assert( worker_ );

    Path *path = worker_->path();

    if( !path || 1 >= path->size() )
    {
        DEBUG( "slice: empty path!" );
        if( Mineral *mineral = worker_->mineralTarget() )
        {
            mineral->release( worker_ );
            worker_->set_target( NULL );
        }
        worker_->set_state( Worker::Idle );
        return;
    }

    path->pop_front();

    DEBUG( "PF: slice" );
    PathFind::Result res = _pf->search_path( worker_,
                                             terrain()->posloc( worker_ ),
                                             terrain()->real2pos( path->front() ),
                                             PF_RADIUS );

    if( res.success && res.path->length() < MAX_SLICE_LENGTH )
    {
        worker_->move( &res.path->concat(path) );
    }
    else
    {
        if( Mineral *mineral = worker_->mineralTarget() )
        {
            mineral->release( worker_ );
            worker_->set_target( NULL );
        }
        worker_->set_state( Worker::Idle );
    }
}


