
#ifndef _RESGATHERBRAIN_H
#define _RESGATHERBRAIN_H

#include <list>

#include "Map.H"
#include "GameTile.H"

class Game;


namespace blazi {


class Loc;
class Terrain;
class PathFind;
class Obj;
class Mineral;
class Worker;
class Sheep;
class ControlCenter;


/** The brain of resource gathering. */
class Brain
{
public:

    static const sint4 PF_RADIUS = 30;
    static const sint4 PF_CC_MINERAL = 10;
    static const sint4 MAX_SLICE_LENGTH = 80;

public:

    Brain( const Game &game_ );
    ~Brain();

    void compute_actions();

protected:

    Terrain* terrain() { return _terrain; }
    const Terrain* terrain() const { return _terrain; }

    const Map<GameTile>& get_map() const { return _game.get_map(); }
    sint4 get_tile_points() const { return _game.get_tile_points(); }

    void choose_mineral( Worker *worker_ );
    void path_slicing( Worker *worker_ );

private:

    const Game &_game;

    Terrain *_terrain;
    PathFind *_pf;

    // cc
    ControlCenter *_cc;

    // minerals
    std::list<Mineral*> _minerals;

    // workers
    std::list<Worker*> _workers;

    // sheeps
    std::list<Sheep*> _sheeps;
};


} /* namespace blazi */


#endif /* _RESGATHERBRAIN_H */

