
#include "Game.H"
#include "SDLinit.H"
#include "Options.H"
#include "ShutDown.H"
#include "LoopFunctor.H"
#include "GameStateModule.H"
#include "GfxGlobal.H"
#include "GfxModule.H"

#include "_Global.H"
#include "_EventHandler.H"

static char *VERSION = "blazi v0.1";

using namespace std;


/** Loop representation for the gfx module. */
class Looper : public LoopFunctor
{
public:

    Looper( blazi::GameState *state_ )
        : state(state_) {}

    void loop() const
    {
        if( state->gfxm )
        {
            if( state->gfxm->get_quit() )
            {
                state->quit = true;
                exit( 0 );
            }

            bool refresh;
            Options::get( "-refresh", refresh );

            if( refresh )
            {
                if( state->just_drew )
                    state->just_drew = false;
                else
                    state->gfxm->draw();
            }
        }

        if( state->gsm->exit() ) exit( 0 );
        if( !state->gsm->recv_view() ) SDL_Delay( 1 );
    }

private:

    blazi::GameState *state;
};


/** Environment initalizations
 *  and running the main loop. */
int main( int argc, char **argv )
{
    signals_shut_down( true );
    atexit( SDL_Quit );

    Options opt( "blazi" );
    opt.put( "-usegfx", "enable graphics" );
    opt.put( "-dmidx", (sint4)0, "index of debugged mineral" );
    opt.put( "-dmidx2", (sint4)0, "index of debugged mineral" );

    GameStateModule::Options::add();
    Game::Options::add();
    GfxModule::Options::add();

    if( Options::process( argc, argv, cerr, VERSION ) ) exit( 20 );

    SDLinit::video_init();
    SDLinit::network_init();

    GameStateModule::Options gsmo;
    GameStateModule gsm( gsmo );

    GfxModule::Options gfxmo;
    GfxModule *gfxm = NULL;
    {
        bool use_gfx;
        Options::get( "-usegfx", use_gfx );
        if( use_gfx ) gfxm = new GfxModule();
    }

    if( !gsm.connect() ) ERR( "connection problems" );

    blazi::GameState state( &gsm, gfxm );
    blazi::EventHandler evehan( state );
    gsm.add_handler( &evehan );

    Looper looper( &state );

    if( gfxm )
    {
        glutInit( &argc, argv );

        gfxm->init( gsm.get_game(),
                    gsm.get_changes(),
                    gsm.get_action_changes(),
                    gfxmo );

        gfxm->start_loop( &looper );
    }
    else
    {
        while( !state.quit ) looper.loop();
    }

    delete gfxm;
    return 0;
}

