#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "main.h"


void
on_new1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  new_x = 32;
  new_y = 32;
  new_z = 1;

  GtkWidget *dialog = create_newdialog();

  gtk_dialog_run (GTK_DIALOG(dialog));
  
  gtk_widget_destroy(dialog);

  gtk_window_set_title( GTK_WINDOW (mainwindow), "ORTS Map Editor");
}


void
on_open1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *dialog = create_open_selection();
  gint response = gtk_dialog_run (GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
  if (response != GTK_RESPONSE_OK) return;

  char temp[255];
  char *t1, *t2;

  t1 = t2 = ofname;

  while (t2) {
    t1 = t2;
    t2 = index(t2, '/');
    if (t2) t2++;
  }

  strcpy(temp, "ORTS Map Editor - ");
  strcat(temp, t1);

  gtk_window_set_title( GTK_WINDOW (mainwindow), temp);
}


void
on_save1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (ofname[0]) {

    FILE *f = fopen(ofname, "w");

    fwrite(&map_x, 4, 1, f);
    fwrite(&map_y, 4, 1, f);
    fwrite(&z_sight, 4, 1, f);

    fwrite(height_map, 1, map_x*map_y, f);
    fwrite(type_map,   1, map_x*map_y, f);
    fwrite(ramp_opts,  1, map_x*map_y, f);

    fclose(f);
  } else {
    GtkWidget *dialog = create_save_selection();
    gint response = gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
    if (response != GTK_RESPONSE_OK) return;
  }

  char temp[255];
  char *t1, *t2;

  t1 = t2 = ofname;

  while (t2) {
    t1 = t2;
    t2 = index(t2, '/');
    if (t2) t2++;
  }

  strcpy(temp, "ORTS Map Editor - ");
  strcat(temp, t1);

  gtk_window_set_title( GTK_WINDOW (mainwindow), temp);
}


void
on_save_as1_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *dialog = create_save_selection();
  gint response = gtk_dialog_run (GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
  if (response != GTK_RESPONSE_OK) return;

  char temp[255];
  char *t1, *t2;

  t1 = t2 = ofname;

  while (t2) {
    t1 = t2;
    t2 = index(t2, '/');
    if (t2) t2++;
  }

  strcpy(temp, "ORTS Map Editor - ");
  strcat(temp, t1);

  gtk_window_set_title( GTK_WINDOW (mainwindow), temp);
}


void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_main_quit();
}


void
on_clear1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  memset(height_map, 0, map_x*map_y);
  memset(type_map, 2, map_x*map_y);
  memset(ramp_opts, 0, map_x*map_y);
}

void
draw_tile(int x1, int y1) {
  x1 = x1 - (brush_size-1)/2;
  y1 = y1 - (brush_size-1)/2;
  
  int x,y;

  for(x = x1; x < x1+brush_size; x++) {
    for (y = y1; y < y1+brush_size; y++) {
      if (x < 0 || x >= map_x) continue;
      if (y < 0 || y >= map_y) continue;

      int tind = y * map_x + x;

      if (brush_type == 1) {
	height_map[tind] = brush_z; //= 0;
      }
      if (brush_type == 2) {
	height_map[tind] = brush_z;
      }
      
      type_map[tind] = brush_type;
      
      if (x < map_x-1) {
	if ((height_map[tind+1] != brush_z)||(type_map[tind+1] != brush_type))
	  ramp_opts[tind+1] |= 2;
	else
	  ramp_opts[tind+1] &= ~2;
      }
      if (x > 0) {
	if ((height_map[tind-1] != brush_z)||(type_map[tind-1] != brush_type))
	  ramp_opts[tind] |= 2;
	else
	  ramp_opts[tind] &= ~2;
      }
      if (y < map_y-1) {
	if ((height_map[tind+map_x] != brush_z)||(type_map[tind+map_x] != brush_type))
	  ramp_opts[tind+map_x] |= 1;
	else
	  ramp_opts[tind+map_x] &= ~1;
      }
      if (y > 0) {
	if ((height_map[tind-map_x] != brush_z)||(type_map[tind-map_x] != brush_type))
	  ramp_opts[tind] |= 1;
	else
	  ramp_opts[tind] &= ~1;
      }
    }
  }
}

gboolean
on_drawingarea1_button_press_event     (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  int x,y;
  x = (int)event->x / 16;
  y = (int)event->y / 16;

  px = x;
  py = y;

  if (x >= 0 && x < map_x && y >= 0 && y < map_y && event->button == 1) {
    draw_tile(x,y);

    GdkWindow *win = gtk_widget_get_parent_window(widget);

    GdkRectangle rect;
    rect.x = x*16+16;
    rect.y = y*16+16;
    rect.width = 16;
    rect.height = 16;

    gdk_window_invalidate_rect(win, &rect, TRUE);
  }

  return TRUE;
}

gboolean
on_drawingarea1_expose_event           (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  int i,j;


  gtk_drawing_area_size(GTK_DRAWING_AREA(widget), map_x*16, map_y*16);


  GdkColor c;
  c.red = 0;
  c.green = 0;
  c.blue = 0;

  GdkColor red;
  red.red = 0;//65535;
  red.green = 0;
  red.blue = 0;

  GdkColor white;
  white.red = 65535;
  white.green = 65535;
  white.blue = 65535;

  GdkColor black;
  black.red = 0;
  black.green = 0;
  black.blue = 0;

  gdk_window_clear_area(widget->window,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state],
			    &event->area);

  c.red = c.green = c.blue = 256*64;

  gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &c);
  gdk_draw_rectangle(widget->window, widget->style->fg_gc[widget->state], TRUE, 0, 0, 16*map_x-1, 16*map_y-1);

  gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &black);
  gdk_draw_rectangle(widget->window, widget->style->fg_gc[widget->state], FALSE, 0, 0, 16*map_x-1, 16*map_y-1);

#define LW 5

  for (i = 0; i < map_x; ++i) {
    for (j = 0; j < map_y; ++j) {
      int tind = j*map_x+i;

      if (type_map[tind] == 2) {

	switch (height_map[tind]) {
	case 0:  c.red =   0; c.green =   0; c.blue =   0; break;
	case 1:  c.red =   0; c.green =  64; c.blue =   0; break;
	case 2:  c.red =   0; c.green =  96; c.blue =   0; break;
	case 3:  c.red =   0; c.green = 140; c.blue =   0; break;
	case 4:  c.red =   0; c.green = 200; c.blue =   0; break;
	case 5:  c.red = 150; c.green = 220; c.blue =   0; break;
	case 6:  c.red = 220; c.green = 200; c.blue =   0; break;
	case 7:  c.red = 200; c.green = 140; c.blue =   0; break;
	case 8:  c.red = 180; c.green = 100; c.blue =   0; break;
	case 9:  c.red = 140; c.green =  80; c.blue =   0; break;
	case 10: c.red = 120; c.green =  80; c.blue =  80; break;
	case 11: c.red =  80; c.green =  80; c.blue =  80; break;
	case 12: c.red = 100; c.green = 100; c.blue = 100; break;
	case 13: c.red = 150; c.green = 150; c.blue = 150; break;
	case 14: c.red = 200; c.green = 200; c.blue = 200; break;
	case 15: c.red = 250; c.green = 250; c.blue = 250; break;
	}
	c.red   *= 256;
	c.green *= 256;
	c.blue  *= 256;

      }

      if (type_map[tind] == 1) {

	switch (height_map[tind]) {
	case 0:  c.red =   0; c.green =   0; c.blue =  96; break;
	case 1:  c.red =   0; c.green =   0; c.blue = 140; break;
	case 2:  c.red =   0; c.green =   0; c.blue = 190; break;
	case 3:  c.red =   0; c.green =   0; c.blue = 250; break;
	case 4:  c.red =   0; c.green =  90; c.blue = 250; break;
	case 5:  c.red =   0; c.green = 150; c.blue = 250; break;
	case 6:  c.red =   0; c.green = 200; c.blue = 250; break;
	case 7:  c.red =   0; c.green = 250; c.blue = 250; break;
	case 8:  c.red =  96; c.green = 250; c.blue = 200; break;
	case 9:  c.red = 100; c.green = 200; c.blue = 160; break;
	case 10: c.red =  80; c.green = 130; c.blue = 100; break;
	case 11: c.red = 100; c.green =  70; c.blue = 120; break;
	case 12: c.red = 100; c.green =  40; c.blue = 100; break;
	case 13: c.red = 140; c.green =  20; c.blue = 140; break;
	case 14: c.red = 190; c.green =   0; c.blue = 190; break;
	case 15: c.red = 250; c.green =   0; c.blue = 250; break;
	}
	c.red   *= 256;
	c.green *= 256;
	c.blue  *= 256;

      }

      gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &c);
      gdk_draw_rectangle(widget->window, widget->style->fg_gc[widget->state], TRUE, 16*i, 16*j, 15, 15);

      gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &red);
      if (ramp_opts[tind] & 1) gdk_draw_rectangle(widget->window, widget->style->fg_gc[widget->state], TRUE, 16*i-1, 16*j-(LW/2+1), 17, LW);
      if (ramp_opts[tind] & 2) gdk_draw_rectangle(widget->window, widget->style->fg_gc[widget->state], TRUE, 16*i-(LW/2+1), 16*j-1, LW, 17);

    }
  }

  gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &white);

  for (i = 0; i < map_x; ++i) {
    for (j = 0; j < map_y; ++j) {
      int tind = j*map_x+i;
      if (ramp_opts[tind] & 1) gdk_draw_line(widget->window, widget->style->fg_gc[widget->state], 16*i-1, 16*j-1, 16*(i+1)-1, 16*j-1);
      if (ramp_opts[tind] & 2) gdk_draw_line(widget->window, widget->style->fg_gc[widget->state], 16*i-1, 16*j-1, 16*i-1, 16*(j+1)-1);
    }
  }


  gdk_gc_set_rgb_fg_color(widget->style->fg_gc[widget->state], &black);

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state], NULL);
  
  return TRUE;
}


gboolean
on_drawingarea1_motion_notify_event    (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  int x,y;
  x = (int)event->x / 16;
  y = (int)event->y / 16;

  if (x >= 0 && x < map_x && y >= 0 && y < map_y && (event->state & GDK_BUTTON2_MASK)) {
    if (x > px) {
      ramp_opts[y*map_x+px+1] |= 2;
    }
    if (x < px) {
      ramp_opts[y*map_x+px] |= 2;
    }
    if (y > py) {
      ramp_opts[(py+1)*map_x+x] |= 1;
    }
    if (y < py) {
      ramp_opts[py*map_x+x] |= 1;
    }
  }

  if (x >= 0 && x < map_x && y >= 0 && y < map_y && (event->state & GDK_BUTTON3_MASK)) {
    if (x > px) {
      ramp_opts[y*map_x+px+1] &= ~2;
    }
    if (x < px) {
      ramp_opts[y*map_x+px] &= ~2;
    }
    if (y > py) {
      ramp_opts[(py+1)*map_x+x] &= ~1;
    }
    if (y < py) {
      ramp_opts[py*map_x+x] &= ~1;
    }
  }


  px = x;
  py = y;

  if (x >= 0 && x < map_x && y >= 0 && y < map_y && (event->state & GDK_BUTTON1_MASK)) {
    draw_tile(x,y);

    GdkWindow *win = gtk_widget_get_parent_window(widget);

    GdkRectangle rect;
    rect.x = x*16+16;
    rect.y = y*16+16;
    rect.width = 16;
    rect.height = 16;

    gdk_window_invalidate_rect(win, &rect, TRUE);
  }

  return TRUE;
}

void
on_spinbutton1_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  brush_z = gtk_spin_button_get_value_as_int(spinbutton);
}


void
on_button1_pressed                     (GtkButton       *button,
                                        gpointer         user_data)
{
  brush_type = 1;
}


void
on_button2_pressed                     (GtkButton       *button,
                                        gpointer         user_data)
{
  brush_type = 2;
}

void
on_spinbutton2_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  new_x = gtk_spin_button_get_value_as_int(spinbutton);
}


void
on_spinbutton3_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  new_y = gtk_spin_button_get_value_as_int(spinbutton);
}

void
on_newdialog_response                  (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  switch (response_id) {
  case GTK_RESPONSE_OK:
    map_x = new_x;
    map_y = new_y;
    z_sight = new_z;
    free(height_map);
    free(type_map);
    free(ramp_opts);
    height_map = (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);
    type_map =   (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);
    ramp_opts =  (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);
    memset(height_map, 0, map_x*map_y);
    memset(type_map, 2, map_x*map_y);
    memset(ramp_opts, 0, map_x*map_y);

    ofname[0] = 0;

    break;
  default:
    break;
  }

}


void
on_fixed2_size_request                 (GtkWidget       *widget,
                                        GtkRequisition  *requisition,
                                        gpointer         user_data)
{
  requisition->width = (map_x+2) * 16;
  requisition->height = (map_y+2) * 16;
}


void
on_drawingarea1_size_request           (GtkWidget       *widget,
                                        GtkRequisition  *requisition,
                                        gpointer         user_data)
{
  requisition->width = map_x * 16;
  requisition->height = map_y * 16;
}


void
on_save_selection_response             (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  fname = (char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));

  if (!fname) {
    printf ("no file name\n");
    return;
  }

  FILE *f = 0;

  switch (response_id) {

  case GTK_RESPONSE_OK:
    f = fopen(fname, "w");

    if (!f) {
      printf ("no file name\n");
      return;    
    }

    strncpy(ofname, fname, 255);

    fwrite(&map_x, 4, 1, f);
    fwrite(&map_y, 4, 1, f);
    fwrite(&z_sight, 4, 1, f);

    fwrite(height_map, 1, map_x*map_y, f);
    fwrite(type_map,   1, map_x*map_y, f);
    fwrite(ramp_opts,  1, map_x*map_y, f);

    fclose(f);

    break;
  default:
    break;
  }
}


void
on_open_selection_response             (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  fname = (char*)gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));

  FILE *f = 0;

  switch (response_id) {

  case GTK_RESPONSE_OK:
    f = fopen(fname, "r");
    
    if (!f) {
      printf ("no file name\n");
      return;    
    }

    strncpy(ofname, fname, 255);

    fread(&new_x, 4, 1, f);
    fread(&new_y, 4, 1, f);
    fread(&z_sight, 4, 1, f);

    map_x = new_x;
    map_y = new_y;

    free(height_map);
    free(type_map);
    free(ramp_opts);

    height_map = (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);
    type_map =   (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);
    ramp_opts =  (unsigned char*)malloc(sizeof(unsigned char)*map_x*map_y);

    fread(height_map, 1, map_x*map_y, f);
    fread(type_map,   1, map_x*map_y, f);
    fread(ramp_opts,  1, map_x*map_y, f);

    fclose(f);

    break;
  default:
    
    break;
  }
}

void
on_window1_destroy                     (GtkObject       *object,
                                        gpointer         user_data)
{
  gtk_main_quit();
}


void
on_prop_dialog_response                (GtkDialog       *dialog,
                                        gint             response_id,
                                        gpointer         user_data)
{
  switch (response_id) {

  case GTK_RESPONSE_OK:
    z_sight = new_z;
    break;
  default:
    break;
  }
}


void
on_spinbutton4_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  new_z = gtk_spin_button_get_value_as_int(spinbutton);
}


void
on_properties1_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  new_z = z_sight;

  GtkWidget *dialog = create_prop_dialog();

  gtk_dialog_run (GTK_DIALOG(dialog));
  
  gtk_widget_destroy(dialog);
}


void
on_spinbutton4_realize                 (GtkWidget       *widget,
                                        gpointer         user_data)
{
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), (float)z_sight);
}


void
on_button1_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  brush_type = 1;
}


void
on_button2_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  brush_type = 2;
}


void
on_spinbutton5_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  brush_size = gtk_spin_button_get_value_as_int(spinbutton);
}

