// $Id: osl_orts_main.C,v 1.2 2004/10/29 17:03:01 orts_jyip Exp $

// This is an ORTS file (c) Jonathan Yip, licensed under the GPL

#include "GfxGlobal.H"
#include "kernel_init.H"
#include "serverclient_init.H"
#include "gfxclient_init.H"
#include "Options.H"
#include "GfxClient.H"
#include "ShutDown.H"
#include "SDLinit.H"

// OSL headers
#include "OSL.H"
//#include "OSL_Opcodes.H"
//#include "OSL_RecStream.H"
//#include "Lexer.H"
#include "OSL_Parser.H"

// OSL ORTS headers
#include "MyHandler.H"

using namespace std;

const char *OSL_ORTS_VERSION = "OSL Client version 0.0 (2004-Oct-20)";

#define USE_EFENCE 0

extern int EF_PROTECT_FREE, EF_FILL, EF_PROTECT_BELOW, EF_ALLOW_MALLOC_0;

static void add_options(Options &o) {
  o.put("-oslfile", string("scripts/simple.osl"), "osl script file");
}

int main(int argc, char **argv)
{
  Options opt;

#if USE_EFENCE
  // aggressive efence checks
  EF_PROTECT_FREE = 1;
  EF_FILL = 0xee;
  EF_PROTECT_BELOW = 1;
  EF_ALLOW_MALLOC_0 = 1;
#endif

  signals_shut_down(true);

  kernel_add_options(opt);
  serverclient_add_options(opt);
  gfxclient_add_options(opt);
  add_options(opt);
  
  // process command line arguments
  if (opt.process(argc, argv, cerr, OSL_ORTS_VERSION)) exit(20);

  gfxclient_init_rnd(opt);

#if 0  
  sint4 decomp;
  opt.get("-decomp", decomp);

  if (decomp) {
    Decomp::test(opt);
    return 0;
  }
#endif
  
  SDLinit::video_init();
  SDLinit::network_init();
  
  GfxClient client(opt);

  OSL::Program prog;
  OSL::Parser p(prog, "apps/osl_orts/scripts/simple.osl");
  p.parse();

  cout << "Done parsing" << endl;

  // add update event handlers here like so:
  MyHandler h(&client, &prog); 
  client.add_update_handler(&h);


  
  client.connect();

  return 0;
}
