#include "benchcam.h"

int benchcam::num_frames = 5000;

double benchcam::pi_o_180 = 3.14159265/180;

benchcam::benchcam(int _mode) : _x(0), _y(0), _z(0) { init(_mode); }
benchcam::benchcam(int _mode, double x, double y, double z) : _x(x), _y(y), _z(z) { init(_mode); }

benchcam::~benchcam() { delete tracking_info; }

void benchcam::apply_rotation() {

  if(mode==RECORD){
	glRotatef(pitch,1,0,0);
	glRotatef(yaw,0,1,0);
  }
  else{
	glRotatef( *(tracking_info+tracking_counter*5),1,0,0);
	glRotatef( *(tracking_info+tracking_counter*5+1),0,1,0);
  }
}

void benchcam::apply_translation() {
    switch(mode) {

    case RECORD:
	glTranslatef(-_x,-_y,-_z);

	if(recording) {
	    if(tracking_counter < num_frames) {

		*(tracking_info+tracking_counter*5)   = pitch;
		*(tracking_info+tracking_counter*5+1) = yaw;
		*(tracking_info+tracking_counter*5+2) = _x;
		*(tracking_info+tracking_counter*5+3) = _y;
		*(tracking_info+tracking_counter*5+4) = _z;

		tracking_counter++;
	    }
	    else {
		FILE *f = fopen("tracking.bin", "wb");
		fwrite(tracking_info,sizeof(double),tracking_counter*5,f);
		fclose(f);
		exit(0);
	    }
	}

	break;

    case PLAYBACK:
	_x = *(tracking_info+tracking_counter*5+2);
	_y = *(tracking_info+tracking_counter*5+3);
	_z = *(tracking_info+tracking_counter*5+4);

	glTranslatef( -_x, -_y, -_z);

	if(tracking_counter < num_frames)
	    tracking_counter++;
	else
	    exit(0);

	break;
    }


}

void benchcam::apply_transform() {

    switch(mode) {

    case RECORD:

	if(recording) {
	    if(tracking_counter < num_frames) {

		*(tracking_info+tracking_counter*5)   = pitch;
		*(tracking_info+tracking_counter*5+1) = yaw;
		*(tracking_info+tracking_counter*5+2) = _x;
		*(tracking_info+tracking_counter*5+3) = _y;
		*(tracking_info+tracking_counter*5+4) = _z;

		tracking_counter++;
	    }
	    else {
		FILE *f = fopen("tracking.bin", "wb");
		fwrite(tracking_info,sizeof(double),tracking_counter*5,f);
		fclose(f);
		exit(0);
	    }
	}

	break;

    case PLAYBACK:
	pitch=*(tracking_info+tracking_counter*5);
	yaw=*(tracking_info+tracking_counter*5+1);
	_x = *(tracking_info+tracking_counter*5+2);
	_y = *(tracking_info+tracking_counter*5+3);
	_z = *(tracking_info+tracking_counter*5+4);
	
	if(tracking_counter < num_frames)
	    tracking_counter++;
	else
	    exit(0);

	break;
    }
    glRotatef(pitch,1,0,0);
    glRotatef(yaw,0,1,0);
    glTranslatef(-_x,-_y,-_z);

}

void benchcam::update() {
  static int framecount=0;
  const int count=10;

  if(++framecount==count)
  {
    timeb timeB;
    int ms_diff;
    ftime(&timeB);
    ms_diff=1000*(timeB.time-timeA.time)+timeB.millitm-timeA.millitm;
    fps=(double)count/(double(ms_diff)/1000.0);
    timeA=timeB;
    framecount=0;
  }

    if(mode == RECORD) {
	if(keys['W'] || keys['w'])
	    vz -= speed/fps;
	else if(keys['S'] || keys['s'])
	    vz += speed/fps;

	if(keys['A'] || keys['a'])
	    vx -= speed/fps;
	else if(keys['D'] || keys['d'])
	    vx += speed/fps;

	if(keys['Q'] || keys['q'])
	    vy -= speed/fps;
	else if(keys['Z'] || keys['z'])
	    vy += speed/fps;

	if(keys['R'] || keys['r'])
	    recording = true;

	_x += vx*cos(yaw*pi_o_180)-vz*sin(yaw*pi_o_180);
	_z += vx*sin(yaw*pi_o_180)+vz*cos(yaw*pi_o_180);
	_y -= vy;

	yaw += vyaw;
	pitch += vpitch;

	vx *= 0.99;
	vy *= 0.99;
	vz *= 0.99;

	vyaw *= 0.99;
	vpitch *= 0.99;
    }
}

void benchcam::update_mouse_moved(int x, int y) {

    if(mouse_differential_reset) {
	mx = x;
	my = y;
	mouse_differential_reset = false;
    }
    else {
	prev_mx = mx;
	prev_my = my;

	mx = x;
	my = y;

	if(rotation_active) {
	    vyaw += (mx-prev_mx)/300;
	    vpitch += (my-prev_my)/300;
	}
    }
}

void benchcam::update_mouse_pressed(int button, int state,
					 int x, int y) {

    if(state==GLUT_DOWN)
	mouse_differential_reset = true;

    if(button==GLUT_LEFT_BUTTON && state==GLUT_DOWN)
	rotation_active = true;
    else
	rotation_active = false;
}

void benchcam::set_keyboard_array(unsigned char *k) { keys = k; }

bool benchcam::is_moving() const {
    if(fabs(vx)-0.001>0 ||
       fabs(vy)-0.001>0 ||
       fabs(vz)-0.001>0 ||
       fabs(vyaw)-0.000000001>0 ||
       fabs(vpitch)-0.000000001>0) {
	return true;
    }
    else {
	return false;
    }
}

void benchcam::init(int _mode) {
   
    mode = _mode;

    fps=60;
      
    mx = 0;
    my = 0;

    prev_mx = 0;
    prev_my = 0;

    vx = 0;
    vy = 0;
    vz = 0;

    yaw = 0;
    vyaw = 0;

    pitch = 0.0;
    vpitch = 0;

    rotation_active = false;
    mouse_differential_reset = false;

    tracking_info = (double *) malloc(sizeof(double)*num_frames*5);
    tracking_counter = 0;

    if(mode==PLAYBACK) {
	FILE *f = fopen("tracking.bin", "rb");
	fread(tracking_info,sizeof(double),num_frames*5,f);
	fclose(f);
    }

    recording = false;
}
