#ifndef PathProcessor_H
#define PathProcessor_H

// $Id: PathProcessor.H,v 1.7 2005/06/11 20:06:47 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GameObj.H"
#include "Decomp.H"

class PolyGraph;

class Corridor {
public:
  Point2i p1;
  Point2i p2;

  Corridor(Point2i a, Point2i b) {
    p1 = a;
    p2 = b;
  }
};

class UnitPath
{
public:
    
  GameObj* unit;
  Vector<Point2i> path;
  Vector<Decomp::Rect> tunnel;
  Vector<Corridor> doors;
    
  Point2i source;
  Point2i target;
    
  int current_target;

  bool active;

  UnitPath() : unit(NULL), current_target(1) {
    active = false;
  };

  UnitPath(GameObj* obj) {
    current_target = 1;
    unit = obj;
    active = false;
  }

  ~UnitPath() {
    if (!path.empty()) path.clear();
    if (!tunnel.empty()) tunnel.clear();
    if (!doors.empty()) doors.clear();
  }

  void set_path(Vector<Point2i> &v) {
    path = v;
    current_target = path.size()-2;
  }

  void set_tunnel(Vector<Decomp::Rect> &t) {
    tunnel = t;
  }

  void set_doors(Vector<Corridor> &c) {
    doors = c;
  }

  void print_path() {
    cout<<"/*****path*****/"<<endl;
    FORALL (path, i) {
      cout<<"waypoint"<<i->x<<" "<<i->y<<endl;
    }
    cout<<"/*****end*****/"<<endl;
  }
};


class PathProcessor {
public:
  Vector<UnitPath> paths;

  sint4 scale;

  int count;
    
  PathProcessor() {
    count = 0;
  }
  ~PathProcessor() {
    if (!paths.empty()) {
      paths.clear();
    }
  }

  void process_paths() {

    remove_dead_paths();

	  
    for(Vector<UnitPath>::iterator i = paths.begin();
	i!= paths.end(); i++) {
      UnitPath &cur_path = *i;
#if 0
      cout<<"number:"<<count<<endl;
      cout<<cur_path.path.size()<<endl;
      cout<<cur_path.current_target<<endl;
#endif	
      count++;
		
      if (cur_path.unit->sod.is_moving) {
	cout<<"unit stop moving at:"<<cur_path.unit->get_int("x")<<" "<<
	  cur_path.unit->get_int("y")<<endl;

		    
		    
	continue;
      } else {

		    
	sint4 x = cur_path.unit->get_int("x");
	sint4 y = cur_path.unit->get_int("y");

		    
	Point2i p(x,y);

	Point2i cur_goal = cur_path.path.at(cur_path.current_target);


#if 0
	cout<<"current"<<p.x<<" "<<p.y<<endl;

	cout<<cur_path.current_target<<endl;
		    
	cout<<"goal"<<cur_goal.x<<" "<<cur_goal.y<<endl;

	cur_path.print_path();

	cout<<"source "<<cur_path.source.x<<" "<<cur_path.source.y<<endl;

	cout<<"target "<<cur_path.target.x<<" "<<cur_path.target.y<<endl;
#endif
		    
	//temporary target reach
	if (p == cur_goal ) {
			  
	  //reach the goal
	  if (cur_path.current_target == 0) {

	    cout<<"done!"<<endl;
				
	    paths.erase(i);

	    i--;
				
	  } else {
	    cout<<"moving to the next target\n";
				
	    //go to the nest target
	    cout<<cur_path.current_target<<endl;
				
	    cur_path.current_target--;

	    Vector<sint4> params;

	    Point2i goal = cur_path.path.at(cur_path.current_target);

	    params.push_back(goal.x);
	    params.push_back(goal.y);
				
	    params.push_back(*cur_path.unit->sod.max_speed);

#if 0

	    cout << "sending unit to" << goal.x << " " << goal.y << endl;

	    cout<<cur_path.current_target<<endl;

#endif
				
	    cur_path.unit->set_action("move", params);

	    cur_path.active = true;
				
	  }
	} else {

	  if (cur_path.active) {
	    //cur_path.unit->clear_action();
	    continue;
	  }
	  Vector<sint4> params;
  
	  Point2i goal = cur_path.path.at(cur_path.current_target);
			  
	  params.push_back(goal.x);
	  params.push_back(goal.y);

	  params.push_back(*cur_path.unit->sod.max_speed);
			  
	  cur_path.unit->set_action("move", params);

	  cur_path.active = true;
	}
      }
		
    }

    count = 0;
  }

  void add_new_path(UnitPath &new_path) {
    //delete previous path for the same object
    for(Vector<UnitPath>::iterator i = paths.begin();
	i != paths.end(); i++) {

      if (i->unit == new_path.unit) {
	paths.erase(i);
	cout << "deleting previous path" << endl;
	break;
      }
    }

    paths.push_back(new_path);
  }

  void remove_dead_paths() {

    for(Vector<UnitPath>::iterator i = paths.begin();
	i != paths.end(); i++) {
		
      //remove dead object
      if (i->unit->is_dead()) {
	paths.erase(i);
	i--;
      }
    }
  }

  void remove_all_paths() {

    paths.clear();
  }

  void remove_all_ground_paths() {
    for(Vector<UnitPath>::iterator i = paths.begin();
	i != paths.end(); i++) {
		
      //remove ground object
      if (i->unit->get_int("zcat") == Object::ON_LAND) {
	paths.erase(i);
	i--;
      }
    }
  }

  void set_paths(Vector<UnitPath> &p) {
    paths = p;
  }
};

#endif
