#ifndef GLWIDGET_H
#define GLWIDGET_H

/**
 * glfractal.h
 *
 * Brief Definition of GLFractal
 *
 * This is a simple QGLWidget displaying an openGL Sierpinski Gasket
 */

#include <qgl.h>
#include <iostream>

#include "Worlded.H"
#include "Tile.H"
#include "Map.H"

const uint4 gWATER = 1;
const uint4 gGROUND = 2;
const uint4 gCLIFF = 3;

struct MyPoint {
  float x;
  float y;
}; // end Point

/**
 * GLFractal inherits from QGLWidget (http://doc.trolltech.com/2.3/qglwidget.html) <BR>
 * QGLWidget basically allows you to control OpenGL through method calls
 */
class GLWidget : public QGLWidget
{
  Q_OBJECT
public:
  GLWidget(QWidget* parent, const std::string &name);
  virtual ~GLWidget();



 public slots:
    void setXRotation(int degrees);
    void setYRotation(int degrees);
    void setZRotation(int degrees);
    // void setScale(int newscale);

protected:

    void initializeGL();
    void paintGL();
    void resizeGL(int w, int h );

  void drawTile(const Tile& t, const MyPoint& tlc, GLfloat width, GLfloat height);
  void drawGrid();
  void setColor(uint4 tileType, sint4 height);
  //void drawGrid();

  void mousePressEvent(QMouseEvent *e);
  void mouseReleaseEvent(QMouseEvent *e);
  void mouseMoveEvent(QMouseEvent *e);



private:
  
  bool leftMouseDown;
  bool middleMouseDown;
  bool rightMouseDown;

  GLint mapTileWidth, mapTileHeight;
  GLint viewportWidth, viewportHeight;
  GLint mouseXPosition, mouseYPosition;

    GLuint object;
    GLfloat xRot, yRot, zRot, scale;
};

#endif

