// $Id: ShooterAI.C,v 1.2 2007/03/24 22:44:42 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, Tim Furtak licensed under the GPL

#include "SquadCombatAI.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameObj.H"
#include "GameChanges.H"
#include "Profiler.H"

using namespace std;

//===================================================================

real8 SquadCombatAI::unitRank(GameObj* enemy)
{
  switch(target_strategy) {
    case SHOOT_RANDOM:    return unitRandom(/*enemy*/); 
    case SHOOT_WEAKEST:   return (-1) * (real8)unitHealth(enemy);
    case SHOOT_IMPORTANT:
    case SHOOT_IMPORTANT_2:
                          return unitImportance(enemy);
    default:
                          return 0;
  }
  return 0.0;
}

//-------------------------------------------------------------------

real8 SquadCombatAI::unitRandom(/*GameObj *enemy*/) {
  return drand48(); //(real8) (lrand48() % 100);
}

//-------------------------------------------------------------------

sint4 SquadCombatAI::unitHealth(GameObj *enemy) {
  stype *hpp = enemy->get_val_ptr("hp");
  if (!hpp) return -1;
  return (hpp->to_int());
}

//-------------------------------------------------------------------

sint4 SquadCombatAI::unitAttack(GameObj* unit) {
  ScriptObj *weapon = unit->component("weapon");
  if(weapon) 
    return ( weapon->get_val_ptr("min_damage")->to_int() +
                weapon->get_val_ptr("min_damage")->to_int() ) / 2;
  return -1;
}
  
//-------------------------------------------------------------------

real8 SquadCombatAI::unitImportance(GameObj *enemy)
{
  stype *hpp, *dmin, *dmax;
  sint4 enemy_hp, enemy_damage = 0;
  real8 q, a = 0.0, t;

  // get hitpoints
  hpp = enemy->get_val_ptr("hp");
  if (!hpp) return -1;
  enemy_hp = hpp->to_int();
  if(enemy_hp == 0) return 0;

  // get weapon damage - default to zero (buildings)
  ScriptObj *weapon = enemy->component("weapon");
  if (weapon) {
    dmin = enemy->component("weapon")->get_val_ptr("min_damage");
    dmax = enemy->component("weapon")->get_val_ptr("max_damage");
    enemy_damage = (dmin->to_int()+dmax->to_int()) / 2;
  }

  // sum of friendlies attack on this unit

  const Vector<GameObj::WPtr> &possible_attackers = shooter_data[enemy].incoming;
  FORALL (possible_attackers, unit) {
    const ShooterData &sd = shooter_data[*unit];
    if (sd.avg_damage > 0) {
      a += (real8)sd.avg_damage;
    }
  }
  if (fabs(a) <= 0.00001) return -1;
  
  // calculate time to death (t) and importance (q) 
  t = ceil((real8)enemy_hp / a);
  q = (real8)enemy_damage / t;  
  //cout <<"enemy# "<<enemy <<" d="<<enemy_damage<<" hp="<<enemy_hp<<" a="<<a<<" q="<<q<<endl;
  return q;
}

//-------------------------------------------------------------------

bool SquadCombatAI::cmpAttack(GameObj* a, GameObj* b) {
  return (unitAttack(a) < unitAttack(b));
}

//-------------------------------------------------------------------

real8 SquadCombatAI::avgEnemyImportance(GameObj* unit) {
  real8 avgEnemyRank = 0.0, enemyRank;
  const Vector<GameObj::WPtr> &enemies = shooter_data[unit].outgoing;
  FORALL (enemies, enemy) {
    enemyRank = unitAttack(*enemy) - shooter_data[unit].armor;
    avgEnemyRank = enemyRank;
    //if (unitRank > maxEnemyRank) {
      //maxEnemyRank = 
  }
  avgEnemyRank /= enemies.size();
  return avgEnemyRank;
}

//-------------------------------------------------------------------

bool SquadCombatAI::cmpEnemyImportance(GameObj* a, GameObj* b) {
  return (avgEnemyImportance(a) < avgEnemyImportance(b));
}

//-------------------------------------------------------------------
  
//choose attackers to reduce overkill
void SquadCombatAI::chooseAttackers(Vector<GameObj*> &attackers, GameObj* unit,
                                    const Vector<GameObj::WPtr> &possible_attackers_)
{
  sint4 health = (sint4) (unitHealth(unit) * (100+overkill_percent)/100.0);
  
  attackers.clear();

  // All attackers who can fire this turn
  Vector<GameObj*> possible_attackers;
  FORALL (possible_attackers_, it) {
    if (shooter_data[*it].time_to_shoot != 0 ) continue;
    possible_attackers.push_back(*it);
  }

  // All possible attackers who have no other choice, shoot at unit
/*  for(Vector<GameObj*>::iterator attacker=possible_attackers.begin(); attacker!=possible_attackers.end();) {
    if (shooter_data[*attacker].outgoing.size() == 1) {
      attackers.push_back(*attacker);
      health -= unitAttack(*attacker) - shooter_data[unit].armor;
      attacker = possible_attackers.erase(attacker);
      continue;
    }
    ++attacker;
  }
*/  
  // Choose units to minimize overkill
  if(target_strategy != SHOOT_IMPORTANT_2) {
    //sort(possible_attackers.begin(), possible_attackers.end(), &cmpAttack);
    FORALL (possible_attackers, attacker) {
      if(health > 0 /*&& shooter_data[*attacker].time_to_shoot == 0 */) {
        attackers.push_back(*attacker);
        health -= unitAttack(*attacker) - shooter_data[unit].armor;
      }
    }
  }

  // Choose attacking units based on secondary criteria to reduce overkill
  // (avg rank of neighboring targets OR last person who shot)
  else {
    real8 minImportance, importance;
    typeof(possible_attackers.begin()) importantUnit;
    while(health > 0 && possible_attackers.size()>0) {
      importantUnit = possible_attackers.begin();
      minImportance = /*shooter_data[*importantUnit].time_to_shoot;*/avgEnemyImportance(*importantUnit);//99999999;
      FORALL(possible_attackers, attacker) {
        importance = /*shooter_data[*attacker].time_to_shoot;*/avgEnemyImportance(*attacker);
        if(minImportance > importance /*&& shooter_data[*attacker].time_to_shoot == 0*/) {
          minImportance = importance;
          importantUnit = attacker;
        }
      }   
      attackers.push_back(*importantUnit);
      health -= unitAttack(*importantUnit) - shooter_data[unit].armor;
      possible_attackers.erase(importantUnit);  //FIXME!!!
    }
  }

  //cout << " possible # attackers="<<possible_attackers.size()
  //      << " actual # attackers="<<attackers.size()<<endl;
}

//-------------------------------------------------------------------

void SquadCombatAI::recompute_targets()
{
  PROFILE("Shooter - compute targets");
  if (!enabled_attack) return;
  if (target_strategy == SHOOT_NONE) return;

  real8 best_rank, rank;
  GameObj* best_enemy;

  // Reset targets 
  FORALL (friendlies->units, f) {
    shooter_data[*f].old_target = shooter_data[*f].cur_target;
    shooter_data[*f].cur_target = 0;
  }

  // Local decision for target selection
  if (strategy == LOCAL_STRATEGY) {
    FORALL (friendlies->units, f) {
      const Vector<GameObj::WPtr> &in_range_enemies = shooter_data[*f].outgoing;
      best_rank = -999999;
      best_enemy = 0;
      
      FORALL(in_range_enemies, enemy) {
        rank = unitRank(*enemy); 
        if (rank > best_rank) {
          best_rank = rank;
          best_enemy = *enemy;
        }
      }
      
      shooter_data[*f].cur_target = best_enemy;
    }
  }

  // Army-level decision; try to reduce overkill
  if (strategy == GLOBAL_STRATEGY) {

    set<GameObj*> enemy_data;
    if (enemies){
      FORALL (enemies->units, e) {
        enemy_data.insert(*e);
      }
    }

    while (!enemy_data.empty()) {
      best_rank = -999999.0;
      best_enemy = 0;
      
      // Choose highest ranked reachable target left in range map  
      FORALL(enemy_data, iter) {
        GameObj* enemy = *iter;
        rank = unitRank(enemy);
        if (rank > best_rank) {
          best_rank = rank;
          best_enemy = enemy;
        }
      }
      // Decide who shoots at target 
      //choose attackers to reduce overkill (FIXME)
      Vector<GameObj*> attackers;
      chooseAttackers(attackers, best_enemy, shooter_data[best_enemy].incoming);

      FORALL(attackers, it) {
        shooter_data[*it].cur_target = best_enemy;
        //cout<<"set attacker "<<(*it)<<" to target "<<best_enemy<<", best rank = "<< best_rank << " time_to_fire "<<shooter_data[*it].time_to_shoot<<endl;
        //cout<<"best enemy = "<<best_enemy<<endl;
      }
      
      // Revise range map
      removeNode(best_enemy);
      enemy_data.erase(best_enemy);    //remove enemy
      FORALL(attackers, it) {          //remove links from all attacking units
        removeNode(*it);
      }
    }
  }
}

//-------------------------------------------------------------------

/** Remove a GameObj from all outgoing and incoming lists. */
/** Remove a GameObj from its shooter's list of targets (incoming). */

void SquadCombatAI::removeNode(GameObj* node)
{
  GroupData *gd = get_group(node);
  GroupData *ed;

  if (gd == friendlies) {
    ed = enemies;
  } else {
    ed = friendlies;
  }

  ShooterData &sd0 = shooter_data[node];
  sd0.outgoing.clear();
  sd0.incoming.clear();
  
  FORALL (ed->units, u) {
    ShooterData &sd = shooter_data[*u];
    FORALL (sd.outgoing, i) {
      if (i->ptr() == node) {
        sd.outgoing.erase(i);
        break;
      }
    }
    FORALL (sd.incoming, i) {
      if (i->ptr() == node) {
        sd.incoming.erase(i);
        break;
      }
    }

  }
  
}

//===================================================================


