#ifndef BUILDTASKS_STATE_H
#define BUILDTASKS_STATE_H

#include <sys/time.h>
#include "TerrainBase.H"
#include "BuildTasksAI.h"
#include "TerrainModule.H"
#include "AutoGatherAI_G3.H"
#include "Random.H"
#include "TerrainClassifier.H"

namespace BSimul{
  inline real8 getCurrentTime(){
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return (tv.tv_sec + (tv.tv_usec / 1000000.0));
  }

  enum tileState{FREE, OCCUPIED, UNKNOWN};

  class BuildTasksAI;

  class BuildTasksState{
  public:
    BuildTasksState(BuildTasksAI* sim, GameStateModule &gsm_, TerrainModule &tm) {

      simAI = sim;
      for(int i=0; i<64; i++)
	for(int j=0; j<64; j++)
	  buildabletile[i][j]=BSimul::FREE;

      autogatherer=new AutoGatherAI(gsm_, tm);
      gsm = &gsm_;
      rand.srand(0);
    }

    ~BuildTasksState(){
      delete terrMap;
    }
            
    void init_map(){
      terrMap = new mapTerrain::TerrainMap(&(gsm->get_game()));
    }


    ObjectDataMap& getFriendlyBases(){return friendlyBases;}
    ObjectDataMap& getFriendlyUnits(){return friendlyUnits;}
    ObjectVector& getBarracks(){return friendlyBarracks;}
    ObjectVector& getConCens(){return friendlyControlCenters;}
    ObjectVector& getFactories(){return friendlyFactories;}
    ObjectVector& getWorkers(){return friendlyWorkers;}
    ObjectVector& getFighters(){return friendlyFighters;}
    ObjectVector& getNewBuildings(){return newBuildings;}
    ObjectVector& getHarassers(){return harassers;}
    ObjectVector& getMinerals(){return minerals;}
    ObjectVector& getEnemyWorkers(){return enemyWorkers;}
    ObjectVector& getEnemyFighters(){return enemyFighters;}
    ObjectVector& getEnemyBases(){return enemyBases;}
    
    bool incompleteBuildings(){return incompletes;}
    bool queued(){return bqueued;}
    bool gotEnough(){return currentMins>=targetMins;}
    bool frozen(){return freezeQueue;}

    uint4 needMarine(){return queueMarine;}
    uint4 needTank(){return queueTank;}
    uint4 needWorker(){return queueWorker;}
    uint4 getCurrentMins(){return currentMins;}
    uint4 trainingMarine(){return intrainMarine;}
    uint4 trainingWorker(){return intrainWorker;}
    uint4 trainingTank(){return intrainTank;}

    void toggle(){bqueued=(!bqueued);}
    void freeze(){freezeQueue=true;}
    void unfreeze(){freezeQueue=false;}

    void setCurrent(uint4 amount){currentMins=amount;} 
    void addMins(uint4 amount){currentMins+=amount;}
    bool spendMins(uint4 amount){if(amount>currentMins) return false; currentMins-=amount; return true;}

    void noGoal(){targetMins=0;}
    void uptrainMarine(){intrainMarine++;}
    void uptrainWorker(){intrainWorker++;}
    void uptrainTank(){intrainTank++;}
    void queueMarines(uint amount){queueMarine=queueMarine+amount;}
    void queueTanks(uint amount){queueTank=queueTank+amount;}
    void queueWorkers(uint amount){queueWorker=queueWorker+amount;}
    void marineCompleted(){if(queueMarine==0) return; queueMarine--;}
    void tankCompleted(){if(queueTank==0) return; queueTank--;}
    void workerCompleted(){if(queueWorker==0) return; queueWorker--;}
    void marineDone(){if(intrainMarine==0) return; intrainMarine--;}
    void tankDone(){if(intrainTank==0) return; intrainTank--;}
    void workerDone(){if(intrainWorker==0) return; intrainWorker--;}
    void setHarassers(ObjectVector harassers_){harassers=harassers_;}
    void clearHarassers(){harassers.clear();}

    void addEnemyWorker(GameObj* obj){ enemyWorkers.push_back(obj);}
    void addEnemyFighter(GameObj* obj){ enemyFighters.push_back(obj);}
    void addEnemyBase(GameObj* obj){ enemyBases.push_back(obj);}
    void addFriendlyBase(GameObj* obj){ friendlyBases[obj] = ExtraObjectData();}
    void addFriendlyUnit(GameObj* obj){ friendlyUnits[obj] = ExtraObjectData();}
    void addBuildableTile(sint4 x, sint4 y){buildabletile[x][y]=BSimul::FREE; }
    void addMinerals(GameObj* obj){minerals.push_back(obj);}
    void addBarracks(GameObj* obj){friendlyBarracks.push_back(obj);}
    void addConCen(GameObj* obj){friendlyControlCenters.push_back(obj); sint4 x,y; obj->get_center(x,y); home=TerrainBase::Loc(x,y);}
    void addFactory(GameObj* obj){friendlyFactories.push_back(obj);}
    void addWorker(GameObj* obj){friendlyWorkers.push_back(obj);}
    void addFighter(GameObj* obj){friendlyFighters.push_back(obj);}
    void addNewBuilding(GameObj* obj){newBuildings.push_back(obj); incompletes=true;}

 
    void removeFriendlyBase(GameObj* obj){friendlyBases.erase(obj);}
    void removeFriendlyUnit(GameObj* obj){friendlyUnits.erase(obj);}
    void removeBuildableTile(sint4 x, sint4 y){buildabletile[x][y]=OCCUPIED;}
    void removeEnemyWorker(GameObj* obj){
      FORALL(enemyWorkers, it){
	if(*it==obj){ 
	  enemyWorkers.erase(it); 
	  return;
	}
      }
    }
    void removeEnemyFighter(GameObj* obj){
      FORALL(enemyFighters, it){
	if(*it==obj){ 
	  enemyFighters.erase(it); 
	  return;
	}
      }
    }
    void removeEnemyBase(GameObj* obj){
      //sint4 x,y; 
      //obj->get_center(x,y); 
      //TerrainBase::Loc baseloc(x,y);
      FORALL(enemyBases, it){
	if(*it==obj){ 
	  enemyBases.erase(it); 
	  return;
	}
      }
    }
    void removeMinerals(GameObj* obj){
      FORALL(minerals, it){
	if(*it==obj){ 
	  minerals.erase(it); 
	  return;
	}
      }
    }
    void removeBarracks(GameObj* obj){
      FORALL(friendlyBarracks, it){
	if(*it==obj){ 
	  friendlyBarracks.erase(it); 
	  return;
	}
      }
    }
    void removeConCen(GameObj* obj){
      FORALL(friendlyControlCenters, it){
	if(*it==obj){
	  friendlyControlCenters.erase(it); 
	  return;
	}
      }
    }
    void removeFactory(GameObj* obj){
      FORALL(friendlyFactories, it){
	if(*it==obj){ 
	  friendlyFactories.erase(it); 
	  return;
	}
      }
    }
    void removeWorker(GameObj* obj){
      FORALL(friendlyWorkers, it){
	if(*it==obj){
	  friendlyWorkers.erase(it); 
	  return;
	}
      }
    }
    void removeFighter(GameObj* obj){
      FORALL(friendlyFighters, it){
	if(*it==obj){
	  friendlyFighters.erase(it); 
	  return;
	}
      }
    }
    void removeNewBuilding(GameObj* obj){
      FORALL(newBuildings, it){
	if(*it==obj){
	  newBuildings.erase(it); 
	  return;}
      } 
      if(!newBuildings.size()) 
	incompletes=false;
    }
    void removeDimensions(sint4 l, sint4 t, sint4 w, sint4 h){
      sint4 tempx, tempy;
      for(int i=0; i<(w/16); i++){
	for(int j=0; j<(h/16); j++){
	  tempx=l+i*16;
	  tempy=t+j*16;
	  buildabletile[(tempx/16)][(tempy/16)]=BSimul::OCCUPIED;
	}
      }
    }

    void printMap(){
      uint4 j,k;
      for(k=0; k<64; k++)      
	for(j=0; j<64; j++){
	  fprintf(stderr, "%d ", buildabletile[j][k]);
	  if(j==63)
	    fprintf(stderr, "\n");
	}
    }

    BuildTasksAI* simAI;
    AutoGatherAI* autogatherer;

    TerrainBase::Loc getCenterOfObjects(const ObjectVector& objects);
    GameObj* getNearestEnemyControlCenter(const ObjectVector& objects);
    GameObj* getNearestEnemyWorker(const ObjectVector& objects);
    bool noEnemyInRange(const GameObj* obj);
    GameObj* enemyNearBases();

    //for now
    sint4 buildabletile[64][64];
    GameStateModule *gsm;
    Random rand;
    TerrainBase::Loc home;
    bool massAttack, attackSet;
    mapTerrain::TerrainMap* terrMap;

  private:
    // can be made better later, but for now, this is much simpler
    ObjectDataMap friendlyBases;
    ObjectDataMap friendlyUnits;
    ObjectVector minerals;
    ObjectVector friendlyBarracks;
    ObjectVector friendlyControlCenters;
    ObjectVector friendlyFactories;
    ObjectVector friendlyWorkers;
    ObjectVector friendlyFighters;
    ObjectVector newBuildings;   
    ObjectVector harassers;
    ObjectVector enemyWorkers;
    ObjectVector enemyFighters;
    ObjectVector enemyBases;
    //Vector<TerrainBase::Loc> enemyBases;    
    bool incompletes, bqueued, freezeQueue;
    
    uint4 width, height, currentMins, targetMins, queueMarine, queueWorker, queueTank, intrainMarine, intrainWorker, intrainTank;
  };


}





#endif



