// $Id: Gfx2d2.C,v 1.6 2007/05/21 03:44:41 orts_furtak Exp $
 
// This is an ORTS file 
// (c) Michael Buro  
// (c) Tim Furtak    
// (c) Keith Yerex   
// licensed under the GPL

//===================================================================
// 2D Script Graphics Functions
//===================================================================

#include "GfxGlobal.H"
#include "Global.H"
#include "GameStateModule.H"
#include "GameObj.H"
#include "GfxModule.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "GfxObject.H"
#include "GfxTerrain.H"
#include "GfxAnimesh.H"
#include "Minimap.H"
#include "GameTile.H"
#include "Token.H"
#include "LayoutHV.H"
#include "Options.H"
#include "Image.H"
#include "GfxModule.H"
#include "ScriptModel.H"
#include "ScriptVector.H"

using namespace std;

#define PARSE_ERR(s) {							\
  errstr << endl << ": parse error at line " << tokenizer.cur_line();\
  ERR(s);\
}



//===================================================================

/** Prints a formatted printf string along with parameters to the screen.
    A questionable use of typecasting to a variable parameter list when forming
    the c string to print, along with the assumption that parameters are integers.
    If the formatting string tries to print 8 byte values it can get out of sync
    and walk off the end.  Note that variable arg lists upgrade floats to doubles.
*/

stype GUI_printf(const SActParam &p)
{
  VALGEN_CHECK(3);  
  const char *format = p.val[2].to_cstr();
  //  const string text = stype::print_va_list(format, p.val, 3);
  const string text = stype::my_printf(format, p.val, 3);

  glDisable(GL_TEXTURE_2D);
  render_bitmap_string(p.val[0].to_int(), p.val[1].to_int(), GLUT_BITMAP_HELVETICA_12, text.c_str());
  glEnable(GL_TEXTURE_2D);

  return stype();
}

//-------------------------------------------------------------------

stype GUI_text(const SActParam &p)
{
  VALN_CHECK(3);
  const char *text = p.val[2].to_cstr();
  render_bitmap_string(p.val[0].to_int(), p.val[1].to_int(), GLUT_BITMAP_HELVETICA_12, text);
  return stype();
}

//-------------------------------------------------------------------

stype GUI_textf(const SActParam &p)
{
  VALN_CHECK(4);  
  const char *text = p.val[2].to_cstr();
  void *font = get_GLUT_Font(p.val[3].to_int());
  glDisable(GL_TEXTURE_2D);
  render_bitmap_string(p.val[0].to_int(), p.val[1].to_int(), font, text);
  glEnable(GL_TEXTURE_2D);
  return stype();
}

//===================================================================

stype GUI_Color(const SActParam &p)
{
  VALN_CHECK(4);    

  real4 r = p.val[0].to_int() / 255.0;
  real4 g = p.val[1].to_int() / 255.0;
  real4 b = p.val[2].to_int() / 255.0;
  real4 a = p.val[3].to_int() / 255.0;

  glColor4f(r, g, b, a);
  return stype();
}

//===================================================================

stype GUI_Bitmap(const SActParam &p)
{
  VALN_CHECK(5);      

  //  glDisable(GL_CULL_FACE);

  if (p.val[0].to_int() == 0) return stype();

  sint4 x1 = p.val[1].to_int();
  sint4 y1 = p.val[2].to_int();
  sint4 x2 = p.val[1].to_int()+p.val[3].to_int();
  sint4 y2 = p.val[2].to_int()+p.val[4].to_int();

  glBindTexture(GL_TEXTURE_2D, p.val[0].to_int());
  glBegin(GL_QUADS);
  glTexCoord2f(1, 1); glVertex2f(x2, y1);
  glTexCoord2f(1, 0); glVertex2f(x2, y2);
  glTexCoord2f(0, 0); glVertex2f(x1, y2);
  glTexCoord2f(0, 1); glVertex2f(x1, y1);
  glEnd();

//  glEnable(GL_CULL_FACE);

  return stype(1);
}

//===================================================================

stype GUI_Rect(const SActParam &p)
{
  VALN_CHECK(4);

  sint4 x1 = p.val[0].to_int();
  sint4 y1 = p.val[1].to_int();
  sint4 x2 = p.val[0].to_int() + p.val[2].to_int();
  sint4 y2 = p.val[1].to_int() + p.val[3].to_int();

  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glVertex2f(x2, y1);
  glVertex2f(x2, y2);
  glVertex2f(x1, y2);
  glVertex2f(x1, y1);
  glEnd();
  glEnable(GL_TEXTURE_2D);
  
  return stype();
}

//===================================================================

stype GfxModule::GUI_Overlay(const SActParam &p)
{
  VALN_CHECK(5);
  OBJN_CHECK(4);  
  
  real4 x1 = p.val[1].to_int();
  real4 y1 = p.val[2].to_int();
  real4 x2 = p.val[3].to_int();
  real4 y2 = p.val[4].to_int();

  glBindTexture(GL_TEXTURE_2D, p.val[0].to_int());

  ScriptObj *gob = p.obj[0];

#if 1
  stype v = gob->get_val("GfxModel");
  GfxObject *o = stype_to_GfxObject(v);
  
#else
  sintptr *i = gob->get_int_ptr("GfxModel");
  if (!i) { cout << "Overlay: No 'GfxModel' variable" << endl; return stype(1); }
  GfxObject *o = (GfxObject*)(*i);
#endif
  
  glPushMatrix();
  glTranslatef(o->pose.center.x, o->pose.center.y, o->pose.center.z);

  overlay_mode(); // remove rotation

  glBegin(GL_QUADS);

  glTexCoord2f(0, 1); glVertex3f(x1, y1, 0);
  glTexCoord2f(1, 1); glVertex3f(x2, y1, 0);
  glTexCoord2f(1, 0); glVertex3f(x2, y2, 0);
  glTexCoord2f(0, 0); glVertex3f(x1, y2, 0);

  // hack to draw over the crease from the smoothed quad
  glTexCoord2f(0, 0); glVertex3f(x1, y2, 0);
  glTexCoord2f(0, 1); glVertex3f(x1, y1, 0);
  glTexCoord2f(1, 1); glVertex3f(x2, y1, 0);
  glTexCoord2f(1, 0); glVertex3f(x2, y2, 0);

  glEnd();

  glPopMatrix();
  //  overlay_mode_end(); // pops the modelview
  //  glDisable(GL_POLYGON_SMOOTH);

  return stype(1);
}

//-------------------------------------------------------------------

/**
   This is a subset of GUI_Overlay:
   pushes a matrix starts overlay mode and translates
   so that subsequent drawing goes above the given unit
   @see GUI_end_overlay
*/
stype GfxModule::GUI_start_overlay(const SActParam &p)
{
  OBJN_CHECK(1);
  ScriptObj *gob = p.obj[0];
  GfxObject *o = 0;
  ScriptObj *foo = gob->component("model");
  if (foo) {
    ScriptModel *sm = dynamic_cast<ScriptModel*>(foo);
    SCRIPT_ASSERT(sm && "no ScriptModel");
    o = sm->get_mesh();
  } else {

#if 1    
    stype v = gob->get_val("GfxModel");
    o = stype_to_GfxObject(v);
#else    
    sintptr *i = gob->get_int_ptr("GfxModel");
    if (!i) { cout << "start_overlay: No 'GfxModel' variable" << endl; return stype(1); }
    o = (GfxObject*)(*i);
#endif    
  }

  glPushMatrix();  
  glTranslatef(o->pose.center.x, o->pose.center.y, o->pose.center.z);
  overlay_mode(); // remove rotation

  GfxView &v = *cur_view;
  glScalef(v.scale,v.scale,1/v.scale);

  return stype();
}

//-------------------------------------------------------------------

/**
   to be used after GUI_start_overlay()
   just pops the matrix pushed by  GUI_start_overlay()
   @see GUI_start_overlay
*/
stype GfxModule::GUI_end_overlay(const SActParam &)
{ 
  glPopMatrix();
  return stype(1);
}

//===================================================================

stype GfxModule::GUI_Screen2World(const SActParam &p)
{
  VALN_CHECK(3);
  sint4 sx = p.val[0].to_int();
  sint4 sy = p.val[1].to_int();
  Vector<stype> &v = *ScriptVector::stype_to_vecstype(p.val[2]);
  v.clear();

  sint4 w = 3;
  SDL_Rect r = { sx - w/2, sy - w/2, w, w };
  Vec3<real4> inter;
  Vector<IdZ> ids;
  
  if (pick(views[active_view], r, inter, ids, false) > 0) {

    v.push_back(stype(int(inter.x * get_game().get_tile_points())));
    v.push_back(stype(int(inter.z * get_game().get_tile_points())));
    return stype(1);

  } else {

    v.push_back(stype(-1));
    v.push_back(stype(-1));
    return stype();

  }
}

//===================================================================

stype GfxModule::GUI_Screen2Obj(const SActParam &p)
{
  VALN_CHECK(2);  
  sint4 sx = p.val[0].to_int();
  sint4 sy = p.val[1].to_int();

  sint4 w = 3;
  SDL_Rect r = { sx - w/2, sy - w/2, w, w };
  Vec3<real4> inter;
  Vector<IdZ> ids;
  
  pick(views[active_view], r, inter, ids);

  FORALL (ids, i) {
    GfxObject *gfx = GfxObject::id2obj((*i).id >> GfxObject::SUB_ID_LOG);
    SCRIPT_ASSERT(gfx && "GfxObject ptr 0");
    GameObj *gob = GfxGameObject::get_gob(gfx);
    if (!gob) continue;

    return stype(gob);
  }

  return stype((GameObj*)0);
}

//===================================================================

stype GfxModule::GUI_MinimapPos(const SActParam &p)
{
  VALN_CHECK(4);    
  mm_rect.x = p.val[0].to_int();
  mm_rect.y = p.val[1].to_int();
  mm_rect.w = p.val[2].to_int();
  mm_rect.h = p.val[3].to_int();
 
  return stype(1);
}

//===================================================================

stype GUI_Outline(const SActParam &p)
{
  VALN_CHECK(1);    
  GfxTerrainTile::outline = p.val[0].to_int() != 0;
  return stype(1);
}

//===================================================================
/*
stype GUI_SetTexture(const SActParam &p)
{
  ScriptObj *gob = p.obj[0];

  sint4 *i = gob->get_int_ptr("GfxModel");
  if (!i) return 1;

  GfxObject *o = (GfxObject*)(*i);

  o->main_tex = p.val[0];
  return stype();
}
*/

//===================================================================

stype GfxModule::GUI_DrawModel2D(const SActParam &p)
{
  VALN_CHECK(4);  
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_COLOR_MATERIAL);

  GfxAnimesh::beginDraw();

  glClear(GL_DEPTH_BUFFER_BIT);
  GfxObject *gob = stype_to_GfxObject(p.val[0]);
  SCRIPT_ASSERT(gob && "GfxObject ptr 0");
  sint4 x0 = p.val[1].to_int();
  sint4 y0 = p.val[2].to_int();
  // negate object scale (called in draw())
  // normalize size by dividing by radius
  real4 s = 0.6 * p.val[3].to_int() / gob->pose.scale / gob->radius;
  
  glPushMatrix();

  glTranslatef(x0, y0, 0);

  // match perspective of the active view
  glRotatef(views[active_view].x_angle, 1.0, 0.0, 0.0);
  glRotatef(views[active_view].y_angle, 0.0, 1.0, 0.0);

  glScalef(s,s,s);

  glTranslatef(-gob->pose.center.x, -gob->pose.center.y, -gob->pose.center.z);
  real4 foo = gob->pose.scale;
  glRotatef(gob->pose.y_rot,0,1,0);
  glTranslatef(-gob->centroid.x * foo, -gob->centroid.y * foo, -gob->centroid.z * foo);
  glRotatef(-gob->pose.y_rot,0,1,0);
  gob->draw();
  glPopMatrix();

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // particle effects could have changed this

  //  GfxAnimesh::endDraw();

  glDisable(GL_LIGHTING);
  glCullFace(GL_BACK);

  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);

  return stype();
}

//Sami=========================================================================

void* get_GLUT_Font(sint4 script_font)
{
  void *font = 0;
  switch(script_font) {
    case 2: font = GLUT_BITMAP_9_BY_15;        break;
    case 3: font = GLUT_BITMAP_8_BY_13;        break;
    case 4: font = GLUT_BITMAP_TIMES_ROMAN_10; break;
    case 5: font = GLUT_BITMAP_TIMES_ROMAN_24; break;
    case 6: font = GLUT_BITMAP_HELVETICA_10;   break;
    case 7: font = GLUT_BITMAP_HELVETICA_12;   break;
    case 8: font = GLUT_BITMAP_HELVETICA_18;   break;
    default: ERR("NO FONT");
  }
  return font;
}

//Sami=========================================================================

stype GUI_error(const SActParam &p)
{
  VALGEN_CHECK(1);
  const char *format = p.val[0].to_cstr();

  cerr << "\nScriptError : \n\t"
       << stype::print_va_list(format, p.val, 1)
       << endl;

  return stype();
}

//Sami=========================================================================

stype GUI_set_cursor(const SActParam &p)
{
  VALN_CHECK(1);  
  sint4 c = p.val[0].to_int();
  if (c < 0 || (c > 19 && c < 100) || c > 102) c = GLUT_CURSOR_LEFT_ARROW;
  glutSetCursor(c);
  return stype();
}


//===================================================================
// old layout code
//===================================================================

#if 1

// split view

struct NamedLayout {
  const char *name;
  const char *layout;
};


// predefined layouts

static NamedLayout layouts[] = {
  { "1x1",  "H(0)" },
  { "2x1",  "H(2, V(0), V(0))" },
  { "1x2",  "V(2, H(0), H(0))" },
  { "2x2",  "V(2, H(2, V(0), V(0)), H(2, V(0), V(0)))" },
  { "3x1",  "H(3, V(0), V(0), V(0))" },
  { "4x1",  "H(4, V(0), V(0), V(0), V(0))" },
  { "1l2r", "H(2, V(0), V(2, H(0), H(0)))" },
  { "2t1b", "V(2:2:3, H(2, V(0),V(0)), H(0))" },  // 2:3 ratio
  { "3t2b", "V(2:2:3, H(3, V(0), V(0), V(0)), H(2, V(0), V(0)))" } // 2:3 ratio
};


// layout a rectangle using inter-window space and a named layout

static bool layout(const string &layout_name, SDL_Rect &rect, sint4 space,
                   Vector<SDL_Rect> &rects)
{
  FORU (i, sizeof(layouts)/sizeof(layouts[0])) {
    if (layout_name == layouts[i].name) {
      LayoutHV::parse(layouts[i].layout, rect, space, rects);
      return false;
    }
  }
  return true;
}


// update playfield layout
// force: update even if name hasn't changed (new window size)

void GfxModule::update_layout(const string &layout_name, bool force)
{
  // update gfx layout

  views_rect.x = views_rect.y = 0;
  views_rect.w = opt.win_w;
  views_rect.h = opt.win_h;
  
  mm_rect.x = opt.win_w - 200;
  mm_rect.y = 0;
  mm_rect.h = mm_rect.w = 200;

  if (!force && layout_name == prev_layout_name) return;

  if (force) {
    glClearColor(0, 0, 0, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    slow_frames = 0;
  }

  Vector<SDL_Rect> rects;

  SDL_Rect r = views_rect;

  r.x += border;
  r.y += border;
  r.w -= 2*border;
  r.h -= 2*border;
  
  if (layout(layout_name, r, border, rects)) ERR("layout problem");

  if (layout_name != prev_layout_name) { // fixme: see below

    prev_layout_name = layout_name;
    views.clear();

    //    if (rects.size() > view_colors.size()-1) ERR("too few view colors");

    sint4 j=1;
    
    FORALL (rects, i) {

      (*i).y = opt.win_h - (*i).y - (*i).h;
      GfxView v(*i);
      v.id = j;
      
      if (get_game().is_rotated()) { v.x_angle = 45; v.y_angle = 45; }
      else                         { v.x_angle = 45; v.y_angle = 0;  }
      
      //      v.frame_color = view_colors[j++];
      v.frame_color = view_colors[1];
      
      v.vp_scale = max(v.viewport.w, v.viewport.h);

      cout << "NEW VIEW: "; v.write();
      views.push_back(v);
    }

    //    glScissor(0, 0, win_w, win_h);    
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    
    slow_frames = 0; // start over

  } else {

    // keep previous view parameters, only change viewport
    assert(rects.size() == views.size());
    
    FORU (i, rects.size()) {
      rects[i].y = opt.win_h - rects[i].y - rects[i].h;
      views[i].viewport = rects[i];
      views[i].vp_scale = max(rects[i].w, rects[i].h);
      views[i].write();
    }
  }

  update_active_view();
}

#endif

