#include "MapEvalHandler.h"
#include "GameChanges.H"
#include "Server.H"
#include "GameStateModule.H"

using namespace std;

MapEvalHandler::MapEvalHandler(ClientState &c,
			       HeuristicMapWidget* widget,
			       uint4 immobileObjectInfluenceRange,
			       real8 rangeModifier, real8 decay) : state(c)
 // : EventHandler()
{
 // state = c;
  //this->gfxm = gfxm;
  //this->gsm = gsm;
  this->widget = widget;

  setImmobileObjectInfluenceRange(immobileObjectInfluenceRange);
  setRangeModifier(rangeModifier);
  setDecayRate(decay);
  immobileObjects = NULL;
  firstTime = true;

}

MapEvalHandler::~MapEvalHandler()
{
  clearImmobileObjects();
}

bool MapEvalHandler::handle_event(const Event& e)
{
  // this needs to be done on the first iteration.  Not sure if I should keep it or not.  There may be
  // a way to add it to the gui via a bp file. It can't be done in constructor though

  // if (e.get_what() == GfxModule::ADD_CWIDGETS_MSG) {
  if(firstTime){
    printf("FRANO: adding cwidget and stuff\n");
    //		if (widget->isOverlay()) {
    //			gfxm->add_mm_overlay(*widget);
    //		}else
   // {

      state.gfxm->add_cwidget(*widget);
      widget->set_position(widget->getPos().x, widget->getPos().y);

      widget->set_size((uint4)(state.gsm->get_game().get_map().get_width()*widget->mapScaleSize),
		       (uint4)(state.gsm->get_game().get_map().get_height()*widget->mapScaleSize));

  //  }

    widget->setGfxModule(state.gfxm);
    widget->setGameStateModule(state.gsm);

    widget->setRows((uint4)(state.gsm->get_game().get_map().get_height()*state.gsm->get_game().
			    get_tile_points()/widget->getGridSquareHeight()));
    widget->setColumns((uint4)(state.gsm->get_game().get_map().get_width()*state.gsm->get_game().
			       get_tile_points()/widget->getGridSquareWidth()));

    //widget->setRows((uint4)gsm->get_game().get_map().get_height());
    //widget->setColumns((uint4)gsm->get_game().get_map().get_width());
    //widget->setRows(30);
    //widget->setColumns(30);
    //real8 mapHeight = gsm->get_game().get_map().get_height()*gsm->get_game().get_map().get_tile_points();

    widget->calculateGridDimensions();
    widget->initializeMap();

    widget->turn_on();
    //first = false;

    playerID = state.gsm->get_game().get_client_player();

    numPlayers = state.gsm->get_game().get_player_num();

    immobileObjects = new Vector<ImmobileObject>*[widget->getColumns()];

    // set up vectors for the immobile objects... to speed up searching, a vector is made for each grid square.
    FORU (i, widget->getColumns()) {
      immobileObjects[i] = new Vector<ImmobileObject>[widget->getRows()];
    }
  firstTime = false;
}

   // return 0;
  //  }

  // only handle view messages
  if (e.get_what() != GameStateModule::VIEW_MSG) return 0;

  const GameChanges &changes = state.gsm->get_changes();

  //printf("SIZE: %d\n",changes.changed_objs.size());

  // TEMPORARY, until widget is fixed.
  if(widget->getGridHeight() < 1 || widget->getGridWidth() < 1) return 0;

  FORALL (changes.changed_objs, it) {
    GameObj *g = (*it)->get_GameObj();
    if (g) processChangedObject(g);
  }

  FORALL (changes.new_objs, it) {
    GameObj *g = (*it)->get_GameObj();
    if (g) processNewObject(g);
/*
      Vector<sint4> params;

      params.push_back(0);
      params.push_back(0);

      //params.push_back(g->sod.max_speed);

      g->set_action("move",params);
*/
  }

  FORALL (changes.vanished_objs, it) {
    GameObj *g = (*it)->get_GameObj();
    if (g) processVanishedObject(g);
  }

  FORALL (changes.dead_objs, it) {
    GameObj *g = (*it)->get_GameObj();
    if (g) processDeadObject(g);
  }

  // now we have to add to the age of all the squares
  FORU (i, widget->getColumns()) {
    FORU (j, widget->getRows()) {
      //widget->ageMap[i][j]++;
      FORU (k,3) {
	FORU (l,widget->getNumAttributes()) {
	  widget->oldHeuristicMap[i][j][k][l] *= 0.999; // fixme: magic constant?
	}
      }
    }
  }


  // THESE STOPPED WORKING???
  /*
    FORALL_CONST(changes.changed_objs,l) {
    processChangedObject((*l));
    }

    FORALL_CONST(changes.new_objs,i) {
    processNewObject((*i));
    }
  */
  // don't need to do this. since dead objects are also part of changed objects...

  /*
    FORALL_CONST(changes.dead_objs,j) {
    processDeadObject((*j));
    }
  */

  /*
    FORALL_CONST(changes.vanished_objs, k) {
    processVanishedObject((*k));
    }
  */
  //state.client->send_actions();
  //state.gsm->send_actions();
  return 0;
}


// zero out the old values if we see the map, regardless of whether we have the attributes or not
// it's a duplication of code, but I can't think of a much better way to do this without not clearing values of
// old map for attributes that the object does not have
// ie) object with no weapon, but with sight will not clear old weapon values of map without this code

void MapEvalHandler::updateFogOfWarMap(sint4 xIndex, sint4 yIndex, uint4 range)
{
  sint4 finalRangeX = (uint4)((range*rangeModifier * state.gsm->get_game().get_tile_points()) /
			      widget->getGridWidth());
  sint4 finalRangeY = (uint4)((range*rangeModifier*state.gsm->get_game().get_tile_points()) /
			      widget->getGridHeight());

  FORU (j, widget->getNumAttributes()) {

    for (sint4 k = max(0, (int)(xIndex-finalRangeX));
	 k < min((int)widget->getColumns(), (int)(xIndex+finalRangeX+1));
	 k++) {

      for (sint4 l = max(0, (int)(yIndex-finalRangeY));
	   l < min((int)widget->getRows(), (int)(yIndex+finalRangeY+1));
	   l++) {

	real8 distance = sqrt(pow((double)(k-xIndex), 2) + pow((double)(l-yIndex), 2));

  	if (distance <= (real8)finalRangeX && distance <= (real8)finalRangeY) {
	  FORU (m,3) {
	    widget->oldHeuristicMap[k][l][m][j] = 0.0; // since this is within our visible range, we get rid of old values
	  }
        }
      }
    }
  }
}

void MapEvalHandler::updateValuesInRange(sint4 xIndex, sint4 yIndex,
					 uint4 attribute, uint4 type,
					 sint4 value, uint4 range,
					 bool vanished, bool immobile)
{
  sint4 finalRangeX = (uint4)(((real8)range*rangeModifier*state.gsm->get_game().get_tile_points()) /
			      widget->getGridWidth());
  sint4 finalRangeY = (uint4)(((real8)range*rangeModifier*state.gsm->get_game().get_tile_points()) /
			      widget->getGridHeight());


  for (sint4 k = max(0, (int)(xIndex-finalRangeX));
       k < min((int)widget->getColumns(), (int)(xIndex+finalRangeX+1));
       k++) {

    for (sint4 l = max(0, (int)(yIndex-finalRangeY));
	 l < min((int)widget->getRows(), (int)(yIndex+finalRangeY+1));
	 l++) {

      // fixme: * cheaper than pow!
      real8 distance = sqrt(pow((real8)(k-xIndex), 2) + pow((real8)(l-yIndex), 2));

      if (distance <= (real8)finalRangeX && distance <= (real8)finalRangeY) {
	if (vanished) {
	  widget->heuristicMap[k][l][type][attribute] -= value*pow(decayRate, distance);
	  widget->oldHeuristicMap[k][l][type][attribute] += value*pow(decayRate, distance);
	}else if (immobile) {
	  widget->immobileHeuristicMap[k][l][type][attribute] += value*pow(decayRate, distance);
	} else {
	  widget->heuristicMap[k][l][type][attribute] += value*pow(decayRate, distance);
	}
      }// end of update if
    }
  }
  //	printf("end");
}

void MapEvalHandler::resetAgeMap(sint4 xIndex,sint4 yIndex,uint4 range) {
  sint4 finalRangeX = (uint4)(((real8)range*rangeModifier*state.gsm->get_game().get_tile_points()) /
			      widget->getGridWidth());
  sint4 finalRangeY = (uint4)(((real8)range*rangeModifier*state.gsm->get_game().get_tile_points()) /
			      widget->getGridHeight());

  for (sint4 k = max(0, (int)(xIndex-finalRangeX));
      k < min((int)widget->getColumns(), (int)(xIndex+finalRangeX+1));
      k++) {

    for (sint4 l = max(0, (int)(yIndex-finalRangeY));
	 l < min((int)widget->getRows(), (int)(yIndex+finalRangeY+1));
	 l++) {

      // fixme: pow -> *
      real8 distance = sqrt(pow((real8)(k-xIndex),2) + pow((real8)(l-yIndex),2));

    if (distance <= (real8)finalRangeX && distance <= (real8)finalRangeY) {
	widget->ageMap[k][l] = 0;
      }
    }
  }
}


void MapEvalHandler::processNewObject(GameObj* gob) {

  //if (gob->has_attr("x") && gob->has_attr("y")) {
  if (!gob->sod.in_game || !gob->has_attr("hp")) return;

  uint4 xPos,yPos;

  uint4 ownerID = gob->get_int("owner");

  xPos = gob->get_int("x");
  yPos = gob->get_int("y");


  sint4 xIndex = (sint4)(xPos / widget->getGridWidth());
  sint4 yIndex = (sint4)(yPos / widget->getGridHeight());

  uint4 type;

  if (playerID ==ownerID) {
    type = 0;
  } else if (ownerID == numPlayers) {
    type = 2;
  } else {
    type = 1;
  }
  /*
    if (gob->get_int("is_mobile") == 0) {
    printf("NEW Immobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    } else {
    printf("NEW Mobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    }
  */

  // if the new object is an immobile object, we can add it to a list we keep track of for the rest of the game
  // immobile objects are treated differently than mobile objects...
  if (gob->get_int("is_mobile") == 0) {
    sint4 index = getImmobileObjectIndex(immobileObjects[xIndex][yIndex],xPos,yPos);

    // only do something if it's an object we have not seen before
    if (index < 0) {
      //printf("FOUND NEW IMMOBILE OBJECT AT: %d,%d\n",xPos,yPos);
      ImmobileObject newObject;
      newObject.x = xPos;
      newObject.y = yPos;

      newObject.attributes = new sint4[widget->getNumAttributes()];
      FORU (i,widget->getNumAttributes()) {
	if (widget->attributes[i].isComponent) {
	  if (gob->component(widget->attributes[i].componentName) == NULL)
	    continue;
	  // if no such component attr, continue
  	if (!gob->component(widget->attributes[i].componentName)->
	      has_attr(widget->attributes[i].attributeName)) continue;

  	newObject.attributes[i] = gob->component(widget->attributes[i].componentName)->get_int(widget->attributes[i].attributeName);

	} else {
	  if (!gob->has_attr(widget->attributes[i].attributeName)) continue;
	  newObject.attributes[i] = gob->get_int(widget->attributes[i].attributeName);
	}

	updateValuesInRange(xIndex, yIndex, i, type, newObject.attributes[i],
			    immobileObjectInfluenceRange, false, true);
      }

      immobileObjects[xIndex][yIndex].push_back(newObject);

    } else {
      FORU (i,widget->getNumAttributes()) {
	sint4 value;
	bool changed = false;

	if (widget->attributes[i].isComponent) {
	  if (gob->component(widget->attributes[i].componentName) == NULL)
	    continue;
	  // if no such component attr, continue
  	if (!gob->component(widget->attributes[i].componentName)->
	      has_attr(widget->attributes[i].attributeName)) continue;

  	value = gob->component(widget->attributes[i].componentName)->
	    get_int(widget->attributes[i].attributeName);

	  // if the value is unchanged from last time, we don't need to do anything
	  if (immobileObjects[xIndex][yIndex][index].attributes[i] != value) changed = true;

	} else {
	  if (!gob->has_attr(widget->attributes[i].attributeName)) continue;
	  value = gob->get_int(widget->attributes[i].attributeName);
	  if (immobileObjects[xIndex][yIndex][index].attributes[i] != value) changed = true;
	}
	// if we have a changed value, we have to first subtract the old value
	if (changed) {
	  updateValuesInRange(xIndex, yIndex, i, type, -immobileObjects[xIndex][yIndex][index].
			      attributes[i],immobileObjectInfluenceRange,false,true);

	  immobileObjects[xIndex][yIndex][index].attributes[i] = value;
	  updateValuesInRange(xIndex, yIndex, i, type, value, immobileObjectInfluenceRange,
			      false, true);
	}
      }
    }
  } else {
    FORU (j,widget->getNumAttributes()) {
      sint4 value;

      if (widget->attributes[j].isComponent) {
	// check if this component exists. check is fast, since component list is a map
	if (gob->component(widget->attributes[j].componentName) == NULL)
	  continue;
	// if no such component attr, continue
        if (!gob->component(widget->attributes[j].componentName)->
	    has_attr(widget->attributes[j].attributeName)) continue;
        value = gob->component(widget->attributes[j].componentName)->
	  get_int(widget->attributes[j].attributeName);
      } else {
	if (!gob->has_attr(widget->attributes[j].attributeName)) continue;
	value = gob->get_int(widget->attributes[j].attributeName);
      }

      updateValuesInRange(xIndex,yIndex,j,type,value,gob->get_int("max_speed"),false,false);
    }
  }
}

void MapEvalHandler::processDeadObject(GameObj* gob)
{
  //if (gob->has_attr("x") && gob->has_attr("y")) {
  if (!gob->sod.in_game || !gob->has_attr("hp")) return;

  uint4 xPos,yPos;

  xPos = gob->get_int("x");
  yPos = gob->get_int("y");

  sint4 xIndex = (sint4)(xPos / widget->getGridWidth());
  sint4 yIndex = (sint4)(yPos / widget->getGridHeight());

  uint4 ownerID = gob->get_int("owner");

  uint4 type;
  if (playerID ==ownerID) {
    type = 0;
  } else if (ownerID == numPlayers) {
    type = 2;
  } else {
    type = 1;
  }

  // remove all stats from dead objects
  FORU (j,widget->getNumAttributes()) {
    sint4 oldValue;

    if (widget->attributes[j].isComponent) {
      // check if this component exists. check is fast, since component list is a map
      if (gob->component(widget->attributes[j].componentName) == NULL)
	continue;
      // if no such component attr, continue
      if (!gob->component(widget->attributes[j].componentName)->
	  has_attr(widget->attributes[j].attributeName)) continue;
      oldValue = gob->component(widget->attributes[j].componentName)->
	get_int_prev(widget->attributes[j].attributeName);
    } else {
      if (!gob->has_attr(widget->attributes[j].attributeName)) continue;
      oldValue = gob->get_int_prev(widget->attributes[j].attributeName);
    }

    if (gob->get_int("is_mobile") == 0) {
      updateValuesInRange(xIndex,yIndex,j,type,-oldValue,immobileObjectInfluenceRange,false,true);
    } else {
      updateValuesInRange(xIndex,yIndex,j,type,-oldValue,gob->get_int("max_speed"),false,false);
    }
  }

  if (gob->get_int("is_mobile") == 0) {
    sint4 index = getImmobileObjectIndex(immobileObjects[xIndex][yIndex],xPos,yPos);

    if (index < 0) {
      printf("WARNING: you have a dead immobile object that has not been added to immobile objects!!!\n");

    } else {

      delete immobileObjects[xIndex][yIndex][index].attributes;

      // is there a better way to do this??? yes ...

      Vector<ImmobileObject>::iterator it = immobileObjects[xIndex][yIndex].begin() + index;

      assert(it < immobileObjects[xIndex][yIndex].end());
      // for(int i=0; i < index; i++) { it++; }
      immobileObjects[xIndex][yIndex].erase(it);
    }
  }

}

void MapEvalHandler::processVanishedObject(GameObj* gob) {
  //if (gob->has_attr("x") && gob->has_attr("y")) {
  if (!gob->sod.in_game || !gob->has_attr("hp")) return;

  // DO NOTHING for immobile objects... they are handled differently when they disappear
  if (gob->get_int("is_mobile") == 0) return;

  uint4 xPos,yPos;

  uint4 ownerID = gob->get_int("owner");

  xPos = gob->get_int("x");
  yPos = gob->get_int("y");

  sint4 xIndex = (sint4)(xPos / widget->getGridWidth());
  sint4 yIndex = (sint4)(yPos / widget->getGridHeight());

  uint4 type;
  if (playerID ==ownerID) {
    type = 0;
  } else if (ownerID == numPlayers) {
    type = 2;
  } else {
    type = 1;
  }
  /*
    if (gob->get_int("is_mobile") == 0) {
    printf("VANISHED Immobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    } else {
    printf("VANISHED Mobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    }
  */
  FORU (j,widget->getNumAttributes()) {
    sint4 value;

    if (widget->attributes[j].isComponent) {
      // check if this component exists. check is fast, since component list is a map
      if (gob->component(widget->attributes[j].componentName) == NULL)
	continue;
      // if no such component attr, continue
      if (!gob->component(widget->attributes[j].componentName)->
	  has_attr(widget->attributes[j].attributeName)) continue;
      value = gob->component(widget->attributes[j].componentName)->
	get_int(widget->attributes[j].attributeName);
    } else {
      if (!gob->has_attr(widget->attributes[j].attributeName)) continue;
      value = gob->get_int(widget->attributes[j].attributeName);
    }

    updateValuesInRange(xIndex,yIndex,j,type,value,gob->get_int("max_speed"),true,false);

  }
}

void MapEvalHandler::processChangedObject(GameObj* gob) {
  if (!gob->sod.in_game || !gob->has_attr("hp")) return;

  uint4 xPos,yPos,prevXPos,prevYPos,ownerID;
  bool positionChanged = false;

  xPos = gob->get_int("x");
  yPos = gob->get_int("y");

  ownerID = gob->get_int("owner");
  // these are not objects on the map, not sure WHAT they are...
  //if (xPos<1 || yPos < 1|| xPos >= width || yPos >= height) return;
  prevXPos = gob->get_int_prev("x");
  prevYPos = gob->get_int_prev("y");

  sint4 xIndex = (sint4)(xPos / widget->getGridWidth());
  sint4 yIndex = (sint4)(yPos / widget->getGridHeight());

  sint4 oldXIndex = (sint4)(prevXPos / widget->getGridWidth());
  sint4 oldYIndex = (sint4)(prevYPos / widget->getGridHeight());

  /*
    if (gob->get_int("is_mobile") == 0) {
    printf("CHANGED Immobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    } else {
    printf("CHANGED Mobile object at %d,%d: previous vals were %d %d\n",gob->get_int("x"),gob->get_int("y"),gob->get_int_prev("x"),gob->get_int_prev("y"));
    }
  */
  if (xIndex != oldXIndex || yIndex != oldYIndex) {
    positionChanged = true;

    // added just for ability to see query results
    widget->lastChangedObjectPos.x = xIndex;
    widget->lastChangedObjectPos.y = yIndex;

  }

  uint4 type;
  if (playerID ==ownerID) {
    type = 0;
    // we have to do this regardless of whether we move or not, since we ALWAYS need to update the
    // fog of war map with respect to what we see
    updateFogOfWarMap(xIndex,yIndex,gob->get_int("max_speed"));
    /*
      if (positionChanged) {
      // only reset the age map if we move, since if we don't move, then the vanished map doesn't change
      // however, once we move, we have to make sure that the places we saw now start counting from 0
      resetAgeMap(oldXIndex,oldYIndex,gob->get_int("max_speed"));
      }
    */
  } else if (ownerID == numPlayers) {
    type = 2;
  } else {
    type = 1;
  }

  FORU (j, widget->getNumAttributes()) {

    sint4 value = 0,oldValue = 0;
    bool attributeChanged = false;

    if (widget->attributes[j].isComponent) {
      // check if this component exists. check is fast, since component list is a map
      if (gob->component(widget->attributes[j].componentName) == NULL)
	continue;
      // if no such component attr, continue
      if (!gob->component(widget->attributes[j].componentName)->
	  has_attr(widget->attributes[j].attributeName)) continue;
      if (gob->component(widget->attributes[j].componentName)->
	  attr_changed(widget->attributes[j].attributeName)) {
  	  oldValue = gob->component(widget->attributes[j].componentName)->
	  get_int_prev(widget->attributes[j].attributeName);
	attributeChanged = true;
      }
      value = gob->component(widget->attributes[j].componentName)->
	get_int(widget->attributes[j].attributeName);
    } else {
      if (!gob->has_attr(widget->attributes[j].attributeName)) continue;
      if (gob->attr_changed(widget->attributes[j].attributeName)) {
	oldValue = gob->get_int_prev(widget->attributes[j].attributeName);
	attributeChanged = true;
      }
      value = gob->get_int(widget->attributes[j].attributeName);
    }

    if (!attributeChanged && !positionChanged) continue; // if nothing has changed, do nothing. this should be the most common case

    uint4 maxSpeed = gob->get_int("max_speed");

    bool isImmobile = false;
    if (gob->get_int("is_mobile") == 0) {
      sint4 index = getImmobileObjectIndex(immobileObjects[xIndex][yIndex],xPos,yPos);
      if (index < 0) {
	printf("WARNING: changed immobile object is not in immobileObjects list. It should be!\n");
      } else {
	immobileObjects[xIndex][yIndex][index].attributes[j] = value;
      }
      isImmobile = true;
      maxSpeed = immobileObjectInfluenceRange;
    }

    // we need to get rid of the influence we projected last iteration.
    if (positionChanged) {
      if (attributeChanged) {
	updateValuesInRange(oldXIndex, oldYIndex, j, type, -oldValue, maxSpeed, false, isImmobile);

      } else {
	updateValuesInRange(oldXIndex, oldYIndex, j, type, -value, maxSpeed, false, isImmobile);
      }
    } else {
      updateValuesInRange(xIndex, yIndex, j, type, -oldValue, maxSpeed, false, isImmobile);
    }

    // now we add the new influence
    updateValuesInRange(xIndex, yIndex, j, type, value, maxSpeed, false, isImmobile);


  }// end attribute loop
}

sint4 MapEvalHandler::getImmobileObjectIndex(const Vector<ImmobileObject> &objectList,
					     uint4 x, uint4 y)
{
  sint4 counter = 0;
  //if (objectList == NULL) return -1;
  FORALL (objectList, i) {
    if ((*i).x == x && (*i).y == y) return counter;
    counter++;
  }

  return -1;
}

void MapEvalHandler::setImmobileObjectInfluenceRange(uint4 range) {
  this->immobileObjectInfluenceRange = range;
}
void MapEvalHandler::setRangeModifier(real8 modifier) {
	this->rangeModifier = modifier;
}
void MapEvalHandler::setDecayRate(real8 decayRate) {
  this->decayRate = decayRate;
}

void MapEvalHandler::clearImmobileObjects() {
  if (immobileObjects != NULL) {
    FORU (i, widget->getColumns()) {
      FORU (j, widget->getRows()) {
	FORU (k, immobileObjects[i][j].size()) {
	  delete immobileObjects[i][j][k].attributes;
	}
      }
      delete[] immobileObjects[i];
    }
    delete immobileObjects;
  }
}


