// $Id: GfxObject.C,v 1.3 2006/07/10 22:51:52 packer Exp $

#include "GfxObject.H"
#include "GfxModule.H"
#include "GameStateModule.H"
#include "GfxTerrain.H"
#include "GfxTerrainTile.H"
#include "GameTile.H"
#include "Game.H"
#include "PlayerInfo.H"
#include "GameObj.H"
#include "Minimap.H"

using namespace std;

const sint4 AIR_OBJ_HEIGHT = 3; // in tiles (was 4)

uint4 GfxObject::id0 = 2; // > 1
GfxObject::Id2GfxObjMap GfxObject::id2obj_map;
GfxTerrain* GfxObject::terrain = 0;
GfxModule* GfxObject::gfxm = 0;

GfxGameObject::Gfx2GfxGameObj GfxGameObject::gfx2gfxgame_map;
GfxGameObject::Gfx2GameObj    GfxGameObject::gfx2game_map;

//===================================================================

GfxObject::GfxObject(){
  drawn = 0;
  my_id = ++id0;
  if (my_id >= (2000000000 >> SUB_ID_LOG)) ERR("id out of bounds"); // fixme: recycle ids?
  id2obj_map[my_id] = this;
  selected = false;
  dead = false;
  moved = true;
  zcat = O_ON_LAND;

  poses.resize(1);
}

//-------------------------------------------------------------------

GfxObject::~GfxObject(){
  FIND (id2obj_map, i, my_id);
  if (i != id2obj_map.end()) id2obj_map.erase(i);
  FORALL (sub_objects, it) delete *it;
  FORALL (emitters, it) delete *it;
}

//===================================================================

void GfxObject::match_id(uint4 id){
  FIND (id2obj_map, i, my_id);
  if (i != id2obj_map.end()) id2obj_map.erase(i);
  my_id = id;
}

//===================================================================

void GfxObject::draw_at(const Mat4<real4> &mat){
  /*glPushMatrix();
  glMultMatrixf(mat.ptr());
  draw();
  glPopMatrix();*/
  cout << "GfxObject draw_at command requested" << endl;
}

void GfxObject::draw_emitters_at(const Mat4<real4> &mat){ 
  /*glPushMatrix();  
  glMultMatrixf(mat.ptr());
  draw_emitters();
  glPopMatrix();*/
  cout << "GfxObject draw_emitters_at requested" << endl;
}

//===================================================================

void GfxObject::draw_emitters(){
  /*glTranslatef(pose.center.x, pose.center.y, pose.center.z);
  glRotatef(pose.y_rot, 0, 1, 0);
  FORALL (sub_objects, i) (*i)->draw_emitters();
  FORALL (emitters, i)    (*i)->draw_emitters();
  glPopMatrix();*/
  cout << "GfxObject draw_emitters requested" << endl;
}

//===================================================================

void GfxObject::attachGfxObject(GfxObject *o)
{
  sub_objects.push_back(o);
}

//-------------------------------------------------------------------

void GfxObject::attachEmitter(GfxObject *o)
{
  emitters.push_back(o);
}

//===================================================================

void GfxObject::update(real4 dt)
{
  /*
  aabb_o.x = center.x - radius;
  aabb_o.y = center.y - radius;
  aabb_o.z = center.z - radius;
  aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;
  */

  FORALL (sub_objects, it) (*it)->update(dt);
  FORALL (emitters, it)    (*it)->update(dt);
}

//===================================================================

void GfxObject::set_position(R x, R y, R z, uint4 p)
{
  poses[p].center.x = x;
  poses[p].center.y = y;
  poses[p].center.z = z;

  pose.center = Vec3<R>(0,0,0);
  FORALL (poses, i) pose.center += (*i).center;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_rotation(real4 y, uint4 p)
{
  poses[p].y_rot = y;

  pose.y_rot = 0;
  FORALL (poses, i) pose.y_rot += (*i).y_rot;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_offset(real4 y, uint4 p)
{
  poses[p].y_off = y;

  pose.y_off = 0;
  FORALL (poses, i) pose.y_off += (*i).y_off;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_pitch(real4 x, uint4 p)
{
  poses[p].x_rot = x;

  pose.x_rot = 0;
  FORALL (poses, i) pose.x_rot += (*i).x_rot;
  moved = true;
}

//-------------------------------------------------------------------

void GfxObject::set_scale(real4 s, uint4 p)
{
  poses[p].scale = s;
  pose.scale = 1.0;
  FORALL (poses, i) pose.scale *= (*i).scale;
  moved = true;
}

//===================================================================

GfxObject* GfxObject::id2obj(uint4 id)
{
  FIND (id2obj_map, it, id);
  if (it != id2obj_map.end()) return it->second;
  return 0;
}

//===================================================================
//===================================================================
//===================================================================

GfxGameObject::GfxGameObject(GameObj *g) : gob(g)
{
  gfx_obj = 0;
  pose_num = -1;
  y_rot = 0.0;
  x_rot = 0.0;
  heading[0] = 0.0;
  pitch[0] = 0.0;
}

//-------------------------------------------------------------------

GfxGameObject::~GfxGameObject()
{
  FIND (gfx2game_map, it, gfx_obj);
  if (it != gfx2game_map.end()) gfx2game_map.erase(it);

  FIND (gfx2gfxgame_map, it2, gfx_obj);
  if (it2 != gfx2gfxgame_map.end()) gfx2gfxgame_map.erase(it2);
}

//===================================================================

void GfxGameObject::attach_gfxobj(GfxObject *gfx)
{
  pose_num = gfx->poses.size();
  gfx->poses.resize(pose_num+1);

  gfx_obj = gfx;
  gfx2gfxgame_map[gfx] = this;
  gfx2game_map[gfx] = gob;

  game_tick();
}

//===================================================================

void GfxGameObject::game_tick()
{
  assert(this->gob);
  GameObj &gob = *this->gob;

  const sint4 TP  = GUI_gfxm->get_game().get_tile_points();
  const real4 ITP = 1.0/TP;
  
  sint4 ox = *gob.sod.x;
  sint4 oy = *gob.sod.y;
  sint4 x = ox;
  sint4 y = oy;

  if (gfx_obj) {
    if (x < 0) {
      if (gfx_obj->hidden < 2) gfx_obj->hidden = 3;
      return;
    } else if (gfx_obj->hidden) {
      gfx_obj->hidden = 1;
      interp = 0; // don't slide to the new position
    }
  }

  const Map<GameTile> &map = GUI_gfxm->get_game().get_map();
  sint4 mw = map.get_width();
  sint4 mh = map.get_height();  
  
  if (x < 0) x = 0;
  else if (x >= mw * TP) x = mw * TP - 1;

  if (y < 0) y = 0;
  else if (y >= mh * TP) y = mh * TP - 1;

  const GameTile &tile = map(x / TP, y / TP);
  real4 inv_h = 1.0 / GameConst::HEIGHT_MULT;
  real4 h = inv_h * tile.get_min_h();
  real4 h_diff;
  real4 noise = 0.f;
 
  position[1] = position[0];
  heading[1]  = heading[0];
  pitch[1]    = pitch[0];
  
  switch (*gob.sod.zcat) {

  case Object::UNDER_WATER:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_ON_LAND;
    --h;
    break;

  case Object::ON_WATER:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_ON_LAND;
    break;

  case Object::ON_LAND:
  
    if (gfx_obj) {
      const double M_180_PI = 1/0.01745329251994329576;	//!< \f$ 180 / \pi \f$
      //!< The smallest X of type float such that 1.0 + X != 1.0.
      const double M_EPSILON = 0.00001;
    
      gfx_obj->zcat = GfxObject::O_ON_LAND;
      
      enum TILT { TILT_NONE=0, 
                  TILT_RAMPS=1, 
                  TILT_NOISE=2, 
                  TILT_MODE_SINGLE=4,
                  TILT_MODE_LINEAR=16 
                };
      sint4 tilt = 0;
      sintptr *c = gob.get_int_ptr("tilt_mode");
      if (c) tilt = *c;
      
      if(tilt & TILT_NONE)
      {
        pitch[0] = 0.0f;
        
        // Height of the terrain directly below us     
        h = height_above(&noise);
        h += noise;
        break;
      }
      else if(tilt & (TILT_RAMPS | TILT_NOISE))
      {
        // Get the tile sample range
        c = gob.get_int_ptr("tilt_range");
        real4 range = 2.f;
        if (c) range = static_cast<real4>(*c);
        
        // Sample 3 colinear points
        if(tilt & TILT_MODE_LINEAR)
        {          
          // Height of the terrain in front and behind the object
          real4 noise_front, noise_back;
          real4 h_front = height_above(&noise_front, range);
          real4 h_back  = height_above(&noise_back, -range);   

          // Set the objects height to the center position
          h = height_above(&noise); 
          h += noise;

          if(!(tilt & TILT_RAMPS))
            h_front = h_back;

          if(tilt & TILT_NOISE)
          {
            h_front += noise_front;
            h_back  += noise_back;
          }

          // Set the object at a weighted combination of the sample heights
          h = (2.f*h + h_front + h_back) / 4.f;
          
          h_diff = (h_front - h_back) * TP;
        
        // Sample a point ahead of the center (TILT_MODE_SINGLE)
        }else
        {
          // Height of the terrain directly below us     
          h = height_above(&noise);

          // Get the tile sample range
          c = gob.get_int_ptr("tilt_range");
          real4 range = 2.f;
          if (c) range = static_cast<real4>(*c);

          // Height of the terrain in front of the object
          real4 h_ahead, noise_ahead;
          h_ahead = height_above(&noise_ahead, range);

          if(!(tilt & TILT_RAMPS))
            h_ahead = h;

          if(tilt & TILT_NOISE)
            h_ahead += noise_ahead;
          else
            h_ahead += noise;

          h += noise;    // Adjust for noise in the terrain

          h_diff = (h_ahead - h) * TP;
        }

        if (abs(h_diff) > M_EPSILON)
          // Pitch to the slope of the terrain under us
          pitch[0] = atan2(h_diff, range) * M_180_PI;
        else
          pitch[0] = 0.0f;
      }
    }
    break;

  case Object::IN_AIR:
    if (gfx_obj) gfx_obj->zcat = GfxObject::O_IN_AIR;
    h = AIR_OBJ_HEIGHT;
    break;

  default: ERR("unknown zcat"); break;
  }
  
  //------------------------------

  Object::Shape shape = static_cast<Object::Shape>(*gob.sod.shape);

  if (shape == Object::CIRCLE) {

    position[0].x = ox * ITP;
    position[0].y = h;
    position[0].z = oy * ITP;

    //    real4 radius = *gob.sod.radius * ITP;
    //    aabb_o.x = position[0].x - radius;
    //    aabb_o.y = position[0].y - radius;
    //    aabb_o.z = position[0].z - radius;
    //    aabb_d.x = aabb_d.y = aabb_d.z = 2*radius;

    sintptr *face_target = gob.get_int_ptr("face_target");
    
    heading[0] = 0;
    sintptr *ld = gob.get_int_ptr("last_direction");  // 0 = not attacking, 1...GameConst::ATTACK_N
    if ((!face_target || *face_target) &&
        ld && (*ld || gob.attr_changed("last_direction"))) {
      // if the unit has stopped shooting, keep looking in the same direction
      sint4 foo = *ld ? *ld : gob.get_int_prev("last_direction");
      heading[0] = GameConst::angle_from_dir(foo, GameConst::ATTACK_N);
    } else {
      const sintptr *hd = gob.get_int_ptr("heading");  // 0 = stopped, 1...GAME_CONST::HEADING_N
      if (hd) heading[0] = GameConst::angle_from_dir(*hd, GameConst::HEADING_N);
    }

  } else if (shape == Object::RECTANGLE) {

    heading[0] = 0;

    //    sint4 width  = gob.get_int("width");
    //    sint4 height = gob.get_int("height");

    //    aabb_o.x = (ox-width*0.5) * ITP;
    //    aabb_o.y = h;
    //    aabb_o.z = (oy-height*0.5) * ITP;

    //    aabb_d.x = width  * ITP;
    //    aabb_d.z = height * ITP;
    //    aabb_d.y = (aabb_d.z + aabb_d.x) * 0.5;

    position[0].x = ox * ITP;
    position[0].y = h;
    position[0].z = oy * ITP;

  } else {

    ERR2("unknown shape ", shape);

  }

  if (interp > 0) interp = 0;
  else {
    //first time
    interp      = 0;
    position[1] = position[0];
    heading[1]  = heading[0];
    pitch[1]    = pitch[0];
  }
}

real4 GfxGameObject::height_above(real4 *noise, real4 x_offset)
{
  const double M_PI_180 = 0.01745329251994329576;	  //!< \f$ \pi /180 \f$
  const sint4 TP  = GUI_gfxm->get_game().get_tile_points();
  const real4 ITP = 1.0/TP;
  
  const Map<GameTile> &map = GUI_gfxm->get_game().get_map();
  
  GameObj &gob = *this->gob;  
  sint4 x = *gob.sod.x;
  sint4 y = *gob.sod.y;
  GfxTerrainTile *gtt = GUI_gfxm->terrain->tile(x / TP, y / TP);
  sint1 *f = gtt->flattened;

  sint4 mw = map.get_width();
  sint4 mh = map.get_height();  
  
  if (x < 0) x = 0;
  else if (x >= mw * TP) x = mw * TP - 1;

  if (y < 0) y = 0;
  else if (y >= mh * TP) y = mh * TP - 1;

  // todo: apply y_offset the test location 

  sint4 sx = x + static_cast<sint4>((x_offset * cos(y_rot * M_PI_180)) + 0.5);
  sint4 sy = y - static_cast<sint4>((x_offset * sin(y_rot * M_PI_180)) + 0.5);

  sint4 tx = sx / TP;
  sint4 ty = sy / TP;
  if (tx < 0) tx = 0;
  if (tx >= mw-1) tx = mw-1;
  if (ty < 0) ty = 0;
  if (ty >= mh-1) ty = mh-1;

  // Get the tile at location tx, ty
  const GameTile &tile = map(tx, ty);
  // Determine the terrain height
  real4 dx = std::min(std::max(real4(sx % TP) / TP, 0.0f), 1.0f);
  real4 dy = std::min(std::max(real4(sy % TP) / TP, 0.0f), 1.0f);
  real4 h = tile.height_at(dx, dy) / GameConst::HEIGHT_MULT;
  
  // Get terrain noise at the same location
  if (noise) {
    if(!f[0] && !f[1] && !f[2] && !f[3])
      *noise = GfxTerrainTile::get_noise_at(sx * ITP, sy * ITP);
    else
      *noise = 0.f;
  }

  return h;
}

//-------------------------------------------------------------------

void GfxGameObject::update(real4 dt)
{
  interp=min(interp+8.0*dt,1.0); // fixme: use game framerate

  center = position[0]*interp + position[1]*(1.0-interp);

  real4 hdiff = heading[0] - heading[1];
  if (hdiff > 180)
    y_rot = (heading[0])*interp + (heading[1]+360)*(1.0-interp);
  else if (hdiff < -180)
    y_rot = (heading[0]+360)*interp + (heading[1])*(1.0-interp);
  else
    y_rot = heading[0]*interp + heading[1]*(1.0-interp);

  if (gfx_obj) {
    gfx_obj->set_position(center.x, center.y, center.z, pose_num);
    gfx_obj->set_rotation(y_rot, pose_num);
    
    x_rot = pitch[0]*interp + pitch[1]*(1.0-interp);
    gfx_obj->set_pitch(x_rot, pose_num);
  }
}

//===================================================================

GameObj* GfxGameObject::get_gob(GfxObject *gfx)
{
  FIND (gfx2game_map, it, gfx);
  if (it != gfx2game_map.end()) return it->second;
  return 0;
}

//-------------------------------------------------------------------

GfxGameObject* GfxGameObject::get_gfx(GfxObject *gfx)
{
  FIND (gfx2gfxgame_map, it, gfx);
  if (it != gfx2gfxgame_map.end()) return it->second;
  return 0;
}

//===================================================================

void GfxGameObject::draw_damage()
{
  if (gob && gob->dir_dmg) {

    real4 r = gob->get_int("radius")/GameConst::TILE_POINTS + 0.5;
    
    // Determine the classification of the object damaged
    sint4 type = 0;
    sintptr *c = gob->get_int_ptr("classification");
    if (c) type = *c;
    enum CLASSIFICATION { UNKNOWN=0, MECHANICAL=1, BIOLOGICAL=2, BUG=4, ORGANIC=8, INORGANIC=16 };

    glPushMatrix();
    glTranslatef(center.x, center.y, center.z);

    real8 ang = -360.0 / GameConst::DAMAGE_N; // amount for each slice

    real4 j = -0.5 + GameConst::DAMAGE_N / 2.0;

    FORS (i, GameConst::DAMAGE_N) {

      if (gob->dir_dmg & (1<<i)) {
        glRotatef(ang*(i-j), 0, 1, 0);
        j = static_cast<real4>(i);
        
        if      (type & MECHANICAL)  glColor3f(0, 0, 1);
        else if (type & BIOLOGICAL)  glColor3f(1, 0, 0);
        else if (type & BUG)         glColor3f(0, 1, 0);
        else if (type & ORGANIC)     glColor3f(1, 1, 0);
        else if (type & INORGANIC)   glColor3f(1, 0, 1);
        else                         glColor3f(1, 1, 1);
        
        glBegin(GL_TRIANGLES);
        glVertex3f(r+0.05, 0.5,  0.0);
        glVertex3f(r+0.25, 0.5,  0.2);
        glVertex3f(r+0.25, 0.5, -0.2);
        
        glVertex3f(r+0.05, 0.5,  0.00);
        glVertex3f(r+0.50, 0.5,  0.15);
        glVertex3f(r+0.50, 0.5, -0.15);
        glEnd();
        
      }
    }
    
    glPopMatrix();
  }

  glerror("GfxObject::draw_damage");
}

//===================================================================

void GfxGameObject::draw_shape()
{
  glerror("GfxAnimesh::draw_shape 0");

  const sint4 TP  = GUI_gfxm->get_game().get_tile_points();
  const real4 ITP = 1.0/TP;

  assert(gob);
  const GameObj &g = *gob;
  
  glEnable(GL_LIGHTING);

  glDisable(GL_BLEND);
  glActiveTextureARB(GL_TEXTURE0);
  glDisable(GL_TEXTURE_2D);

  glDisable(GL_CULL_FACE);
  //  glLoadName(my_id << GfxObject::SUB_ID_LOG);

  glPushMatrix();
  Vec3<R> &c = center;
  glTranslatef(c.x, c.y, c.z);
  glRotatef(y_rot, 0, 1, 0);

  glColor4f(1, 1, 1, 1);

  Object::Shape shape = static_cast<Object::Shape>(*g.sod.shape);
  
  if (shape == Object::CIRCLE) {
    real4 r = *g.sod.radius * ITP;

    glBegin(GL_TRIANGLE_FAN);

    glVertex3f(0, 0.1, 0);

    glVertex3f(r * 1.0000, 0, r * 0.0000);
    glVertex3f(r * 0.8660, 0, r * 0.5000);
    glVertex3f(r * 0.5000, 0, r * 0.8660);
    glVertex3f(r * 0.0000, 0, r * 1.0000);
    glVertex3f(r *-0.5000, 0, r * 0.8660);
    glVertex3f(r *-0.8660, 0, r * 0.5000);
    glVertex3f(r *-1.0000, 0, r * 0.0000);
    glVertex3f(r *-0.8660, 0, r *-0.5000);
    glVertex3f(r *-0.5000, 0, r *-0.8660);
    glVertex3f(r * 0.0000, 0, r *-1.0000);
    glVertex3f(r * 0.5000, 0, r *-0.8660);
    glVertex3f(r * 0.8660, 0, r *-0.5000);
    glVertex3f(r * 1.0000, 0, r *-0.0000);

    glEnd();

  } else if (shape == Object::RECTANGLE) {
    real4 w2 = (*g.sod.x2 - *g.sod.x1) / 2.0 * ITP;
    real4 h2 = (*g.sod.y2 - *g.sod.y1) / 2.0 * ITP;

    glBegin(GL_QUADS);
    glVertex3f(-w2, 0,  h2);
    glVertex3f(-w2, 0, -h2);
    glVertex3f(-w2, 1, -h2);
    glVertex3f(-w2, 1,  h2);

    glVertex3f( w2, 0,  h2);
    glVertex3f( w2, 0, -h2);
    glVertex3f( w2, 1, -h2);
    glVertex3f( w2, 1,  h2);

    glVertex3f( w2, 0, -h2);
    glVertex3f(-w2, 0, -h2);
    glVertex3f(-w2, 1, -h2);
    glVertex3f( w2, 1, -h2);

    glVertex3f( w2, 0,  h2);
    glVertex3f(-w2, 0,  h2);
    glVertex3f(-w2, 1,  h2);
    glVertex3f( w2, 1,  h2);
    glEnd();
  }

  glPopMatrix();
  glEnable(GL_CULL_FACE);
  glDisable(GL_LIGHTING);

  glEnable(GL_TEXTURE_2D);

  glerror("GfxAnimesh::draw_shape 1");
}

//===================================================================

