// $Id: OSL_Context.C,v 1.2 2004/10/29 16:58:40 orts_jyip Exp $

// (c) Michael Buro, mburo@cs.ualberta.ca, licensed under the GPL

/*

result types:

sint - uint = sint !
uint + uint = uint !
sint + sint = sint !
sint - sint = sint !

uint + sint = uint ?
sint + uint = uint ?
uint - sint = uint ?
uint - uint = sint ?

sint * / % sint = sint !
sint * / % uint = sint !
uint * / % sint = sint !
uint * / % uint = uint !

checks:

- (unary)

+ - * /

*/


#include "OSL.H"
#include "OSL_Opcodes.H"

using namespace std;

namespace OSL {

  Context::Context(Program &prog_, sint4 sl) : prog(prog_), stack_len(sl) {
    if (sl < 10) throw MemException(where() + "Context(): stack len < 10");
    stack = new Value[stack_len];
    sp = fp = 0; pc = 0;
    code = 0; code_len = 0;;
  }

  
  Context::~Context() { delete [] stack; }

  
  string Context::where() const {

    if (!fu) return string("at ?:?:") + to_string(pc) + " ";

    return string("at ") +
      (fu->class_id < prog.classes.size() ? prog.classes[fu->class_id].name : "?") + ":"
      + fu->name + ":" + to_string(pc) + " ";
  }

  
  void Context::init_run(FuncMember *new_fu, vector<Value> &par) {
    sp = 0; 
    fu = 0; // indicates external call

    if (par.size()+10 > stack_len)
      throw MemException(where() + "init_run: out of stack memory");

#if 0    
    // fixme: check parameter types!
    
    if (par.size() != new_fu->par_tinds.size())
      throw ParException(where() + "init_run: parameter number mismatch");

    // check parameter types fixme: do more thorough check
    { FORU (i, par.size())
	if (par[i].get_cat() != new_fu->par_tinds[i])
	  throw ParException(where() + "init_run: parameter type mismatch");
    }
#endif
    
    // push parameters in reverse order
    FORALL_REV (par, i) push(*i);

    init_frame(new_fu);
  }

  
  void Context::set_fu(FuncMember *new_fu)
  {
    fu = new_fu;
    code = &fu->code[0];
    code_len = fu->code.size();
  }

  
  void Context::init_frame(FuncMember *new_fu) {

    if (sp + 10 >= stack_len)
      throw MemException(where() + "init_frame: out of stack memory");

    // save current context on stack
    { Value &v = stack[sp]; v.set_cat(Type::TC_fp); v.v_fp = fp; ++sp; }
    { Value &v = stack[sp]; v.set_cat(Type::TC_fu); v.v_fu = fu; ++sp; }
    { Value &v = stack[sp]; v.set_cat(Type::TC_pc); v.v_pc = pc; ++sp; }    
    fp = sp;
    pc = 0;
    set_fu(new_fu);
    
    // create local variables, fixme: speed up with memcopy from new_fu->loc_init 
    FORALL (fu->loc_types, i) {
      Value v;
      if (*i >= Type::TC_END) v.set_cat(Type::TC_object);
      else v.set_cat(Type::Category(*i));
      //      if (i->cat >= Type::TC_END) v.set_cat(Type::TC_object);
      //      else v.set_cat(i->cat);
      push(v);
    }
  }
  

  Value Context::result() const {
    if (sp != 1) throw MemException(where() + "result: no result on stack");
    return stack[0];
  }

  
  // assumption: tind of simple types == TC_ value
  // -> types[] must be initialized correctly!

  inline void Context::check_type(Type::Category t1)
  {
    check_sp(1);
    if (stack[sp-1].get_cat() != t1) {
      stringstream s1; Type::write_cat(t1, s1);
      stringstream s2; Type::write_cat(stack[sp-1].get_cat(), s2);      
      throw TypeException(where() + " expected type " + s1.str() +
			  " instead of " + s2.str());
    }
  }
  
  inline void Context::check_types(Type::Category t1, Type::Category t2)
  {
    check_sp(2);
    if (stack[sp-2].get_cat() != t1 || stack[sp-1].get_cat() != t2) {
      stringstream s1; Type::write_cat(t1, s1);
      stringstream s2; Type::write_cat(t2, s2);
      stringstream s3; Type::write_cat(stack[sp-2].get_cat(), s3);
      stringstream s4; Type::write_cat(stack[sp-1].get_cat(), s4);            
      throw TypeException(where() + " expected types " + s1.str() + " " +
			  s2.str() + " instead of " + s3.str() + " " + 
			  s4.str());
    }
  }

  inline void Context::check_ssize(uint4 n)
  {
    if (sp + n > stack_len)
      throw MemException(where() + "out of stack memory");
  }

  inline void Context::check_sp(uint4 n)
  {
    if (sp < n) throw OutOfRangeException(where() + "sp too small");
  }

  inline void Context::push(const Value &v)
  {
    check_ssize(1);
    stack[sp++] = v;
  }
  
  inline void Context::push_sint4(sint4 v)
  {
    check_ssize(1);  
    Value &val = stack[sp++];
    val.v_sint4 = v;
    val.set_cat(Type::TC_sint4);
  }

  inline void Context::push_uint4(uint4 v)
  {
    check_ssize(1);  
    Value &val = stack[sp++];
    val.v_uint4 = v;
    val.set_cat(Type::TC_uint4);
  }
 
  inline void Context::push_sint8(sint8 v)
  {
    check_ssize(2);  
    Value &val = stack[sp+=2];
    val.v_sint8 = v;
    val.set_cat(Type::TC_sint8);
  }

  inline void Context::push_uint8(uint8 v)
  {
    check_ssize(2);  
    Value &val = stack[sp+=2];
    val.v_uint8 = v;
    val.set_cat(Type::TC_uint8);
  }

  inline void Context::push_real4(real4 v)
  {
    check_ssize(1);  
    Value &val = stack[sp++];
    val.v_real4 = v;
    val.set_cat(Type::TC_real4);
  }

  inline void Context::push_real8(real8 v)
  {
    check_ssize(2);  
    Value &val = stack[sp+=2];
    val.v_real8 = v;
    val.set_cat(Type::TC_real8);
  }

  inline void Context::push_null()
  {
    check_ssize(POINTER_WORDS);
    Value &val = stack[sp+=POINTER_WORDS];
    val.v_object = 0;
    val.set_cat(Type::TC_object);
  }


  Class *Context::get_class(const string &op)
  {
    Value &v = stack[sp-1];
    
    write_stack();

    if (v.get_cat() != Type::TC_object)
      throw OutOfRangeException(where() + op + ": not an object");

    if (v.v_object == 0)
      throw OutOfRangeException(where() + op + ": ref=0");

    uint4 tind = v.v_object->tind;

    if (tind >= prog.types.size()) 
      throw OutOfRangeException(where() + op + ": type index");

    Type &type = prog.types[tind];
    
    if (type.cat != Type::TC_class)
      throw TypeException(where() + op + ": no class");
    
    if ((uint4)type.class_index >= prog.classes.size())
      throw OutOfRangeException(where() + op + ": corrupt class index");
    
    return &prog.classes[type.class_index];
  }


#define BIN_OP(T1, OP, T2)			\
  check_types(Type::TC_##T1, Type::TC_##T2);	\
  stack[sp-2].v_##T1 OP##= stack[sp-1].v_##T2;	\
  pc++; sp--; return CONT;

#define BIN_OP_CHANGE(T1, OP, T2, T1NEW)	\
  check_types(Type::TC_##T1, Type::TC_##T2);	\
  stack[sp-2].v_##T1 OP##= stack[sp-1].v_##T2;	\
  stack[sp-2].cat = Type::TC_##T1NEW;		\
  pc++; sp--; return CONT;

#define BIN_OP_INT4(OP)						\
  if (stack[sp-2].cat != Type::TC_sint4 &&			\
      stack[sp-2].cat != Type::TC_uint4 &&			\
      stack[sp-1].cat != Type::TC_sint4 &&			\
      stack[sp-1].cat != Type::TC_uint4)                        \
    throw TypeException(where() + "2 int4s expected");          \
  stack[sp-2].v_uint4 OP##= stack[sp-1].v_uint4;		\
  pc++; sp--; return CONT;

#define BIN_OP_INT8(OP)						\
  if (stack[sp-2].cat != Type::TC_sint8 &&			\
      stack[sp-2].cat != Type::TC_uint8 &&			\
      stack[sp-1].cat != Type::TC_sint8 &&			\
      stack[sp-1].cat != Type::TC_uint8)                        \
    throw TypeException(where() + "2 int8s expected");          \
  stack[sp-2].v_uint8 OP##= stack[sp-1].v_uint8;		\
  pc++; sp--; return CONT;

#define BIN_REL_OP_SAME(T, OP)						\
  if (stack[sp-2].cat != Type::TC_##T ||				\
      stack[sp-1].cat != Type::TC_##T)                                  \
    throw TypeException(where() + "2 identical types expected");        \
  stack[sp-2].v_sint4 = stack[sp-2].v_##T OP stack[sp-1].v_##T;		\
  pc++; sp--; return CONT;

  // sets result type to sint4
#define BIN_REL_OP_SAME_ADJUST(T, OP)					\
  if (stack[sp-2].cat != Type::TC_##T ||				\
      stack[sp-1].cat != Type::TC_##T)                                  \
   throw TypeException(where() + "2 int4s expected");	                \
  stack[sp-2].v_sint4 = (stack[sp-2].v_##T OP stack[sp-1].v_##T);	\
  stack[sp-2].cat = Type::TC_sint4;					\
  pc++; sp--; return CONT;

  // T1 REL-OP T2 (SIGNLESS = result if signed value < 0
#define BIN_REL_OP_SU(T1, OP, T2, SIGNLESS)				\
  if (stack[sp-2].cat != Type::TC_##T1 ||				\
      stack[sp-1].cat != Type::TC_##T2)                                 \
    throw TypeException(where() +"2 int4s expected");	                \
  if (stack[sp-2].v_##T1 < 0)						\
    stack[sp-2].v_sint4 = SIGNLESS;					\
  else 									\
    stack[sp-2].v_sint4 = (stack[sp-2].v_##T2 OP stack[sp-1].v_##T2);	\
  pc++; sp--; return CONT;

  // T1 REL-OP T2 (SIGNLESS = result if signed value < 0
#define BIN_REL_OP_US(T1, OP, T2, SIGNLESS)				\
  if (stack[sp-2].cat != Type::TC_##T1 ||				\
      stack[sp-1].cat != Type::TC_##T2)                                 \
    throw TypeException(where() + "2 int4s expected");              	\
  if (stack[sp-1].v_##T2 < 0)						\
    stack[sp-2].v_sint4 = SIGNLESS;					\
  else 									\
    stack[sp-2].v_sint4 = (stack[sp-2].v_##T1 OP stack[sp-1].v_##T1);	\
  stack[sp-2].cat = Type::TC_##T1;                                      \
  pc++; sp--; return CONT;


  // convert signed integer into integer
 
#define CONV_SI(F,T,R)							\
{									\
  check_type(Type::TC_##F);						\
  Value &v = stack[sp-1];						\
  if (v.v_##F < min_##T || v.v_##F > max_##T)				\
    throw OutOfRangeException(where() + "conversion "#F" to "#T);	\
  v.cat = Type::TC_##R; v.v_##R = v.v_##F;				\
  pc++; return CONT;							\
}

  // convert unsigned integer into integer
  
#define CONV_UI(F,T,R)							\
{									\
  check_type(Type::TC_##F);						\
  Value &v = stack[sp-1];						\
  if (v.v_##F > (F)max_##T)						\
    throw OutOfRangeException(where() + "conversion "#F" to "#T);	\
  v.cat = Type::TC_##R; v.v_##R = v.v_##F;				\
  pc++; return CONT;							\
}

// convert without checks

#define CONV_WIDEN(F,T)				\
{ 						\
  check_type(Type::TC_##F);			\
  Value &v = stack[sp-1];			\
  v.cat = Type::TC_##T; v.v_##T = v.v_##F;	\
  pc++; return CONT;				\
}

// convert floating point value into integer

#define CONV_RI4(F,T,R)							\
{									\
  check_type(Type::TC_##F);						\
  Value &v = stack[sp-1];						\
  if (v.v_##F-0.5 < min_##T || v.v_##F+0.5 > max_##T)			\
    throw OutOfRangeException(where() + "conversion "#F" to "#T);	\
  /* fixme: is this using the right rounding mode? */                   \
  v.cat = Type::TC_##R; v.v_##R = (R)my_round(v.v_##F);			\
  pc++; return CONT;							\
}



// STORE macros
//   local, parameter, static, instance variable, array element
// 
// S4 version:  top variable must be sint4, value range check
// E version:   top variable must have same type as variable
// R version:   compatible reference expected


///////////////////// store local variable

// variable type must be sint4

#define STORE_LS4(T)							\
{									\
  check_type(Type::TC_sint4);						\
  									\
  /* check range */							\
									\
  Value &v = stack[sp-1];						\
  if (v.v_sint4 < min_##T || v.v_sint4 > max_##T)			\
    throw OutOfRangeException(where() + "STORE_LS4 "#T": value");	\
									\
  uint4 ind = code[pc+1];						\
  if (ind >= fu->loc_types.size())					\
    throw OutOfRangeException(where() + "STORE_LS4 "#T": index");	\
									\
  /* fixme: check whether local variable's real type is T */		\
									\
  if (stack[get_loc_ind(ind)].cat != Type::TC_sint4)			\
    throw TypeException(where() + "STORE_LS4 "#T": var type");		\
  									\
  stack[get_loc_ind(ind)].v_sint4 = v.v_sint4;				\
  sp--; pc += 2; return CONT;						\
}


// variable type must be T

#define STORE_LE(T)							\
{									\
  check_type(Type::TC_##T);						\
		                           				\
  uint4 ind = code[pc+1];                                               \
  if (ind >= fu->loc_types.size())					\
    throw OutOfRangeException(where() + "STORE_LE "#T": index");	\
                                                                        \
  if (stack[get_loc_ind(ind)].cat != Type::TC_##T)			\
    throw TypeException(where() + "STORE_LE "#T": var type");		\
  stack[get_loc_ind(ind)].v_##T = stack[sp-1].v_##T;			\
  sp--; pc += 2; return CONT;						\
}

// variable type must be compatible objref

#define STORE_LR 							\
{									\
  check_type(Type::TC_object);						\
									\
  uint4 ind = code[pc+1];						\
  if (ind >= fu->loc_types.size())					\
    throw OutOfRangeException(where() + "STORE_LR: index");	        \
									\
  if (stack[get_loc_ind(ind)].cat != Type::TC_object)			\
    throw TypeException(where() + "STORE_L: var type");		        \
  									\
  /* fixme: are references compatible? (to test compiler) */            \
  stack[get_loc_ind(ind)].v_object = stack[sp-1].v_object;		\
  sp--; pc += 2; return CONT;						\
}

/////////////////////////////// store parameter

// variable type must be sint4

#define STORE_PS4(T)							\
{									\
  check_type(Type::TC_sint4);						\
  									\
  /* check range */							\
									\
  Value &v = stack[sp-1];						\
  if (v.v_sint4 < min_##T || v.v_sint4 > max_##T)			\
    throw OutOfRangeException(where() + "STORE_PS4 "#T": value");	\
									\
  uint4 ind = code[pc+1];						\
  if (ind >= fu->par_types.size())					\
    throw OutOfRangeException(where() + "STORE_PS4 "#T": index");	\
									\
  /* fixme: check whether variable's real type is T */		        \
									\
  if (stack[get_par_ind(ind)].cat != Type::TC_sint4)			\
    throw TypeException(where() + "STORE_PS4 "#T": var type");		\
  									\
  stack[get_par_ind(ind)].v_sint4 = v.v_sint4;				\
  sp--; pc += 2; return CONT;						\
}


// variable type must be T

#define STORE_PE(T)							\
{									\
  check_type(Type::TC_##T);						\
									\
  uint4 ind = code[pc+1];						\
  if (ind >= fu->par_types.size())					\
    throw OutOfRangeException(where() + "STORE_PE "#T": index");	\
									\
  if (stack[get_par_ind(ind)].cat != Type::TC_##T)			\
    throw TypeException(where() + "STORE_PE "#T": var type");		\
  									\
  stack[get_par_ind(ind)].v_##T = stack[sp-1].v_##T;			\
  sp--; pc += 2; return CONT;						\
}


// variable type must be compatible objref

#define STORE_PR							\
{									\
  check_type(Type::TC_object);						\
									\
  uint4 ind = code[pc+1];						\
  if (ind >= fu->par_types.size())					\
    throw OutOfRangeException(where() + "STORE_PR: index");	        \
									\
  if (stack[get_par_ind(ind)].cat != Type::TC_object)			\
    throw TypeException(where() + "STORE_PR: var type");		\
  									\
  /* fixme: are references compatible? (to test compiler) */            \
  stack[get_par_ind(ind)].v_object = stack[sp-1].v_object;		\
  sp--; pc += 2; return CONT;						\
}


////////////////////// store static variable

// variable type must be sint4

#define STORE_SS4(T)						       	\
{									\
  check_type(Type::TC_sint4);						\
									\
  /* check range */							\
									\
  Value &v = stack[sp-1];						\
  if (v.v_sint4 < min_##T || v.v_sint4 > max_##T)			\
    throw OutOfRangeException(where() + "STORE_SS4 "#T": value");	\
									\
  uint4 cind = code[pc+1], dind = code[pc+2];				\
									\
  if (cind >= prog.classes.size())					\
    throw OutOfRangeException(where() + "STORE_SS4 "#T": class index");	\
									\
  Class &cl = prog.classes[cind];					\
									\
  if (dind >= cl.static_values.size())					\
    throw OutOfRangeException(where() + "STORE_SS4 "#T": data index");	\
									\
  Value &stv = cl.static_values[dind];					\
									\
  /* fixme: check whether variable's real type is T */			\
  									\
  if (stv.cat != Type::TC_sint4)					\
    throw TypeException(where() + "STORE_SS4 "#T": var type");		\
									\
  stv.v_sint4 = v.v_sint4;						\
  sp--; pc += 3; return CONT;						\
}


// variable type must be T

#define STORE_SE(T)							\
{									\
  check_type(Type::TC_##T);						\
									\
  uint4 cind = code[pc+1], dind = code[pc+2];				\
									\
  if (cind >= prog.classes.size())					\
    throw OutOfRangeException(where() + "STORE_SE "#T": class index");	\
									\
  Class &cl = prog.classes[cind];					\
									\
  if (dind >= cl.static_values.size())					\
    throw OutOfRangeException(where() + "STORE_SE "#T": data index");	\
									\
  Value &stv = cl.static_values[dind];					\
									\
  if (stv.cat != Type::TC_##T)						\
    throw TypeException(where() + "STORE_SE "#T": var type");		\
									\
  stv.v_##T = stack[sp-1].v_##T;					\
  sp--; pc += 3; return CONT;						\
}


// variable type must be compatible objref

#define STORE_SR							\
{									\
  check_type(Type::TC_object);						\
									\
  uint4 cind = code[pc+1], dind = code[pc+2];				\
									\
  if (cind >= prog.classes.size())					\
    throw OutOfRangeException(where() + "STORE_SR: class index");	\
									\
  Class &cl = prog.classes[cind];					\
									\
  if (dind >= cl.static_values.size())					\
    throw OutOfRangeException(where() + "STORE_SR: data index");	\
									\
  Value &stv = cl.static_values[dind];					\
									\
  if (stv.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_SR: var type");		\
									\
  /* fixme: are references compatible? (to test compiler) */            \
  stv.v_object = stack[sp-1].v_object;					\
  sp--; pc += 3; return CONT;						\
}


////////////////////// store instance variable

// variable type must be sint4

#define STORE_IS4(T)							\
{									\
  check_types(Type::TC_sint4, Type::TC_object);				\
									\
  /* check range */							\
									\
  Value &v = stack[sp-2];						\
  if (v.v_sint4 < min_##T || v.v_sint4 > max_##T)			\
    throw OutOfRangeException(where() + "STORE_IS4 "#T": value");	\
									\
  uint4 dind = code[pc+1];						\
  Class &cl = *get_class("STORE_IS4 "#T);				\
									\
  if (dind >= cl.inst_data.size())					\
    throw OutOfRangeException(where() + "STORE_IS4 "#T": data index");	\
									\
  Value &ival = static_cast<ClassInstance*>(stack[sp-1].v_object)->data[dind];	\
									\
  /* fixme: check whether variable's real type is T */			\
                                                                        \
  if (ival.cat != Type::TC_sint4)					\
    throw TypeException(where() + "STORE_IS4 "#T": var type");		\
									\
  ival.v_sint4 = v.v_sint4;						\
  sp -= 2; pc += 2; return CONT;					\
}


// variable type must be T

#define STORE_IE(T)							\
{									\
  check_types(Type::TC_##T, Type::TC_object);				\
									\
  uint4 dind = code[pc+1];						\
  Class &cl = *get_class("STORE_IE "#T);				\
									\
  if (dind >= cl.inst_data.size())					\
    throw OutOfRangeException(where() + "STORE_IE "#T": data index");	\
  cout << "CLASS NOW IS AT " << stack[sp-1].v_object << endl;		\
  Value &ival = static_cast<ClassInstance*>(stack[sp-1].v_object)->data[dind];	\
cout << "FOO " << sint4(ival.cat) << " " << Type::TC_##T << " " << dind << endl;\
									\
  if (ival.cat != Type::TC_##T)						\
    throw TypeException(where() + "STORE_IE "#T": var type");		\
  ival = stack[sp-2];							\
  sp -= 2; pc += 2; return CONT;					\
}

// variable type must be compatible objref

#define STORE_IR							\
{									\
  check_types(Type::TC_object, Type::TC_object);			\
									\
  uint4 dind = code[pc+1];						\
  Class &cl = *get_class("STORE_IR");					\
									\
  if (dind >= cl.inst_data.size())					\
    throw OutOfRangeException(where() + "STORE_IR: data index");	\
									\
  Value &ival = static_cast<ClassInstance*>(stack[sp-1].v_object)->data[dind];	\
									\
  if (ival.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_IR: var type");		\
                                                                        \
  /* fixme: are references compatible? (to test compiler) */            \
  ival.v_object = stack[sp-2].v_object;					\
  sp -= 2; pc += 2; return CONT;					\
}


////////////////////// store array element
// top = a ind value

// variable type must be sint4

#define STORE_AS4(T)							\
{									\
  check_sp(3);								\
  check_type(Type::TC_sint4);						\
									\
  /* check range */							\
									\
  Value &v = stack[sp-1];						\
  if (v.v_sint4 < min_##T || v.v_sint4 > max_##T)			\
    throw OutOfRangeException(where() + "STORE_AS4"#T": value");	\
									\
  Value &aref = stack[sp-3];						\
									\
  if (aref.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_AS4"#T": not an object");	\
									\
  if (aref.v_object == 0)						\
    throw TypeException(where() + "STORE_AS4"#T": ref=0");		\
  									\
  if (stack[sp-2].cat != Type::TC_uint4)				\
    throw TypeException(where() + "STORE_AS4"#T": index type != uint4");\
									\
  uint4 ind = stack[sp-2].v_uint4;					\
  									\
  if (aref.v_object->tind >= prog.types.size())				\
    throw OutOfRangeException(where() + "STORE_AS4"#T": aref type index");\
  									\
  Type &type = prog.types[aref.v_object->tind];				\
  if (type.cat != Type::TC_array)					\
    throw TypeException(where() + "STORE_AS4"#T": aref expected");	\
  									\
/*  if (ind < 0 || ind >= static_cast<ArrayInstance*>(aref.v_object)->n)*/\
  if (ind >= static_cast<ArrayInstance*>(aref.v_object)->n)      	\
    throw OutOfRangeException(where() + "STORE_AS4"#T": index");	\
  									\
  Value &aval = static_cast<ArrayInstance*>(aref.v_object)->data[ind];	\
									\
  /* fixme: check whether variable's real type is T */			\
									\
  if (aval.cat != Type::TC_sint4)					\
    throw TypeException(where() + "STORE_AS4"#T": var type");		\
									\
  aval.v_sint4 = v.v_sint4;						\
  sp -= 3; pc++; return CONT;						\
}


// variable type must be T

#define STORE_AE(T)							\
{ \
  write_stack(); 							\
  check_sp(3);								\
/*  check_type(Type::TC_sint4);	why here?*/					\
									\
  Value &v = stack[sp-1];						\
  Value &aref = stack[sp-3];						\
									\
  if (aref.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_AE"#T": not an object");	\
									\
  if (aref.v_object == 0)						\
    throw TypeException(where() + "STORE_AE"#T": ref=0");		\
									\
  if (stack[sp-2].cat != Type::TC_uint4)				\
    throw TypeException(where() + "STORE_AE"#T": index type != uint4");	\
									\
  uint4 ind = stack[sp-2].v_uint4;					\
									\
  if (aref.v_object->tind >= prog.types.size())				\
    throw OutOfRangeException(where() + "STORE_AE"#T": aref type index");\
									\
  Type &type = prog.types[aref.v_object->tind];				\
  if (type.cat != Type::TC_array)					\
    throw TypeException(where() + "STORE_AE"#T": aref expected");	\
									\
/*  if (ind < 0 || ind >= static_cast<ArrayInstance*>(aref.v_object)->n) */ \
  if (ind >= static_cast<ArrayInstance*>(aref.v_object)->n)	        \
    throw OutOfRangeException(where() + "STORE_AE"#T": index");		\
									\
  Value &aval = static_cast<ArrayInstance*>(aref.v_object)->data[ind];	\
									\
  if (aval.cat != Type::TC_##T)						\
    throw TypeException(where() + "STORE_AE "#T": var type");		\
									\
  aval = v;								\
  sp -= 3; pc++; return CONT;						\
}


// variable type must be compatible objref

#define STORE_AR							\
{									\
  check_sp(3);								\
  check_type(Type::TC_object);						\
									\
  Value &v = stack[sp-1];						\
  Value &aref = stack[sp-3];						\
									\
  if (aref.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_AR: not an object");		\
									\
  if (aref.v_object == 0)						\
    throw TypeException(where() + "STORE_AR: ref=0");			\
									\
  if (stack[sp-2].cat != Type::TC_uint4)				\
    throw TypeException(where() + "STORE_AR: index type != uint4");	\
									\
  uint4 ind = stack[sp-2].v_uint4;					\
									\
  if (aref.v_object->tind >= prog.types.size())				\
    throw OutOfRangeException(where() + "STORE_AR: aref type index");	\
									\
  Type &type = prog.types[aref.v_object->tind];				\
  if (type.cat != Type::TC_array)					\
    throw TypeException(where() + "STORE_AR: aref expected");		\
									\
/*  if (ind < 0 || ind >= static_cast<ArrayInstance*>(aref.v_object)->n)*/ \
  if (ind >= static_cast<ArrayInstance*>(aref.v_object)->n)	        \
    throw OutOfRangeException(where() + "STORE_AR: index");		\
									\
  Value &aval = static_cast<ArrayInstance*>(aref.v_object)->data[ind];	\
									\
  if (aval.cat != Type::TC_object)					\
    throw TypeException(where() + "STORE_AR: var type");		\
									\
  /* fixme: check if type is compatible */				\
  aval.v_object = v.v_object;						\
  sp -= 3; pc++; return CONT;						\
}


  FuncMember *Context::remove_frame()
  {
    // params + locals + fp + fu + pc
    if (sp < fu->loc_types.size() + fu->par_types.size() + 3)
      throw OutOfRangeException(where() + "RETURN: sp too small");
    
    // remove frame
    sp = fp;
    
    if (stack[sp-1].cat != Type::TC_pc ||
	stack[sp-2].cat != Type::TC_fu ||
	stack[sp-3].cat != Type::TC_fp)
      throw TypeException(where() + "RETURN: corrupt frame info");
    
    pc = stack[--sp].v_pc;
    FuncMember *prev_fu = stack[--sp].v_fu;
    fp = stack[--sp].v_fp;
    sp -= fu->par_types.size();
    return prev_fu;
  }

  
  Context::StepResult Context::step()
  {
    if (pc >= code_len) throw OutOfRangeException(where() + "pc >= code_len");
    //cout << "Current code " << code[pc] << endl;
    //    write_stack();
    switch (code[pc]) {

    case Opcodes::NOP:
      ++pc; return CONT;

    case Opcodes::CALL_VIRT: // find=fetch() a=pop call a->virt[find]
      {
	uint4 f_ind = code[pc+1];
	Class &cl = *get_class("CALL_VIRT");

	if (f_ind >= cl.virt_funcs.size())
	  throw OutOfRangeException(where() + "CALL_VIRT: func index");	  

	//	init_frame(&cl.virt_funcs[f_ind]);
	//	pc += 2; return CONT;
	pc += 2;
	init_frame(&cl.virt_funcs[f_ind]);
	return CONT;
      }
      
    case Opcodes::CALL_STAT: // cind=fetch() find=fetch() call cind::stat[find]
      {
	uint4 c_ind = code[pc+1];
	uint4 f_ind = code[pc+2];

	if (c_ind < prog.classes.size())
	  throw OutOfRangeException(where() + "CALL_STAT: class index");
	  
	if (f_ind >= prog.classes[c_ind].static_funcs.size())
	  throw OutOfRangeException(where() + "CALL_STAT: func index");	  

	pc += 3;
	init_frame(&prog.classes[c_ind].static_funcs[f_ind]);
	return CONT;

      }
      
    case Opcodes::CALL_EXT: // find=fetch() push(call ext[find] (pop(),...pop()))
      // fixme: no return value?
      {
	uint4 f_ind = code[pc+1];

	if (f_ind >= prog.extfs.size())
	  throw OutOfRangeException(where() + "CALL_EXT: func index");	  

	uint4 pn = prog.extfs[f_ind].par_types.size();
	Value v;
	check_sp(pn);
	write_stack();
	// copy parameters from stack to vector (reverse order)
	vector<Value> params;
	FORU (i, pn) params.push_back(stack[--sp]);   // fixme: type check
	v = prog.extfs[f_ind].call(params);
	if (prog.extfs[f_ind].ret_tind > Type::TC_object)
	  v.cat = Type::TC_object;
	else
	  v.cat = (Type::Category)prog.extfs[f_ind].ret_tind;

	// hack
	v.cat = Type::TC_sint4;

	stack[sp++] = v;
	pc += 2; return CONT;
      }
      
    case Opcodes::RETURN:
      {
	FuncMember *prev_fu = remove_frame();
	if (!prev_fu) return END;
	set_fu(prev_fu);
	return CONT;
      }

    case Opcodes::RETURN_VAL:
      {
        Value v = stack[--sp]; // sp check is done in remove_frame()
	FuncMember *prev_fu = remove_frame();
	stack[sp++] = v;
	if (!prev_fu) return END;
	set_fu(prev_fu);
	return CONT;
      }

    case Opcodes::NEW_CLASS:
      {
	uint4 c_ind = code[pc+1]; 

	if (c_ind >= prog.classes.size())
	  throw OutOfRangeException(where() + "NEW_CLASS: class index");

	Class &cl = prog.classes[c_ind];
	uint4 len = prog.get_class_data_offset() + cl.inst_data_size;

	// fixme is this really necessary?
	assert(cl.inst_data_size != 0);
	
	Heap::mem_unit* loc = prog.heap->alloc(len);

	ClassInstance *ci = reinterpret_cast<ClassInstance*>(loc);

	cout << "HERE" << cl.ref_tind << endl;
	cout << "CLASS LOCATED AT " << ci << " " << (void*)loc << endl;

	ci->tind = cl.ref_tind;

	for (uint4 i=0; i<prog.classes[prog.types[ci->tind].class_index].inst_data.size(); ++i) {
	  TypeIndex ti = prog.classes[prog.types[ci->tind].class_index].inst_data[i].type;
	  Value v;
	  v.cat = prog.types[ti].cat;
	  v.v_uint8 = 0;
	  ci->data[i] = v;
	}

	FORU(i, prog.types.size()) cout << i << " is " << prog.types[i].cat << endl;

	check_ssize(1);	
	Value &v = stack[sp];
	v.v_object = ci;
	v.cat = Type::TC_object;
	sp++; pc += 2; return CONT;
      }

    case Opcodes::NEW_ARRAY:
      {
	uint4 t_ind = code[pc+1];
	uint4 size = code[pc+2];

	uint4 len = prog.get_array_data_offset() + size * sizeof(Value);
	Heap::mem_unit* loc = prog.heap->alloc(len);
	ArrayInstance *ai = reinterpret_cast<ArrayInstance*>(loc);

	ai->tind = t_ind;
	ai->n = size;

	Type::Category cat = prog.types[t_ind].T_cat;

	for (uint4 i=0; i<size; ++i) {
	  Value v;
	  v.cat = cat;
	  v.v_uint8 = 0;
	  ai->data[i] = v;
	}

	check_ssize(1);
	Value &v = stack[sp];
	v.v_object = ai;
	v.cat = Type::TC_object;
	write_stack();
	++sp; pc += 3; return CONT;
      }      

    case Opcodes::DUP:
      check_ssize(1);
      check_sp(1);
      stack[sp] = stack[sp-1];
      ++sp; pc++; return CONT;

    case Opcodes::JUMP:
      {
	uint4 newpc = (uint4)code[pc+1];
	if (newpc >= code_len)
	  throw OutOfRangeException(where() + "JUMP: address out of range");
	pc = newpc; return CONT;
      }
    
    case Opcodes::JUMP_TRUE:
      check_type(Type::TC_sint4);
      if (stack[--sp].v_sint4 != 0) {
	uint4 newpc = (uint4)code[pc+1];
        if (newpc >= code_len)
	  throw OutOfRangeException(where() + "JUMP_TRUE: address out of range");
	pc = newpc; return CONT;
      }
      pc += 2; return CONT;
    
    case Opcodes::JUMP_FALSE:
      check_type(Type::TC_sint4);
      if (stack[--sp].v_sint4 == 0) {
	uint4 newpc = (uint4)code[pc+1];
        if (newpc >= code_len)
	  throw OutOfRangeException(where() + "JUMP_FALSE: address out of range");
	pc = newpc; return CONT;
      }
      pc += 2; return CONT;
    
    case Opcodes::LINE: line_num = code[pc+1]; pc += 2; return CONT;

      // ----------- LOAD -------------

      // constant
    
    case Opcodes::PUSH_CS4: push_sint4(code[pc+1]); pc += 2; return CONT;
    case Opcodes::PUSH_CU4: push_uint4(code[pc+1]); pc += 2; return CONT;
    case Opcodes::PUSH_CS8: push_sint8(*(sint8*)(&code[pc+1])); pc += 3;
      return CONT;
    case Opcodes::PUSH_CU8: push_uint8(*(uint8*)(&code[pc+1])); pc += 3; 
      return CONT;
    case Opcodes::PUSH_CR4: push_real4(*(real4*)(&code[pc+1])); pc += 2; 
      return CONT;
    case Opcodes::PUSH_CR8: push_real8(*(real8*)(&code[pc+1])); pc += 3; 
      return CONT;
    case Opcodes::PUSH_CARR:
      ++pc;
      FORU(i, code[pc])
	push_uint4(code[++pc]);
      ++pc;
      return CONT;

      // 0
    
    case Opcodes::PUSH_ZS4:  push_sint4(0);    pc++; return CONT;
    case Opcodes::PUSH_ZU4:  push_uint4(0);    pc++; return CONT;
    case Opcodes::PUSH_ZS8:  push_sint8(0);    pc++; return CONT;
    case Opcodes::PUSH_ZU8:  push_uint8(0);    pc++; return CONT;
    case Opcodes::PUSH_ZR4:  push_real4(0.0f); pc++; return CONT;
    case Opcodes::PUSH_ZR8:  push_real8(0.0);  pc++; return CONT;
    case Opcodes::PUSH_NULL: push_null();      pc++; return CONT;
    
      // load local variable
      
    case Opcodes::LOAD_LOCAL:  // fetch(i) push(local[i])
      {
	uint4 ind = code[pc+1];
	if (ind >= fu->loc_types.size())
	  throw OutOfRangeException(where() + "LOAD_LOCAL: ind");

	push(stack[get_loc_ind(ind)]);
	pc += 2; return CONT;
      }
      
      // load parameter

    case Opcodes::LOAD_PARAM:  // fetch(i) push(param[i])
      {
	uint4 ind = code[pc+1];

	if (ind >= fu->par_types.size())
	  throw OutOfRangeException(where() + "LOAD_PARAM: ind");

	push(stack[get_par_ind(ind)]);
	pc += 2; return CONT;

      }

      // load static class value

    case Opcodes::LOAD_STATIC: // fetch(cind) fetch(dind) push(cid::stat[dind])
      {
	uint4 cind = code[pc+1], dind = code[pc+2];
	
	if (cind >= prog.classes.size())
	  throw OutOfRangeException(where() + "LOAD_STATIC: class index");

	const Class &cl = prog.classes[cind];

	if (dind >= cl.static_values.size())
	  throw OutOfRangeException(where() + "LOAD_STATIC: data index");

	push(cl.static_values[dind]);
	pc += 3; return CONT;
      }
    
    // load class instance value
    
    case Opcodes::LOAD_INST:   // fetch(dind) a=pop() push(a->memb[dind])
      {
	uint4 dind = code[pc+1];
	Class &cl = *get_class("LOAD_INST");

	if (dind >= cl.inst_data.size())
	  throw OutOfRangeException(where() + "LOAD_INST: data index");

	check_type(Type::TC_object);
	stack[sp-1] = static_cast<ClassInstance*>(stack[sp-1].v_object)->data[dind];
	pc += 2; return CONT;
      }
      
      // load array element
      
    case Opcodes::LOAD_ARRAY:  // i=pop() a=pop() push(a[i])
      {
	check_sp(2);
	Value &aref = stack[sp-2];

	if (aref.cat != Type::TC_object)
	  throw TypeException(where() + "LOAD_ARRAY: not an object");

	if (aref.v_object == 0)
	  throw TypeException(where() + "LOAD_ARRAY: ref=0");
	
	if (stack[sp-1].cat != Type::TC_uint4)
	  throw TypeException(where() + "LOAD_ARRAY: index type != uint4");

	uint4 ind = stack[sp-1].v_uint4;

	if (aref.v_object->tind >= prog.types.size())
	  throw OutOfRangeException(where() + "LOAD_ARRAY: aref type index");
 
	Type &type = prog.types[aref.v_object->tind];
	if (type.cat != Type::TC_array)
	  throw TypeException(where() + "LOAD_ARRAY: aref expected");

	//	if (ind < 0 || ind >= static_cast<ArrayInstance*>(aref.v_object)->n)
	if (ind >= static_cast<ArrayInstance*>(aref.v_object)->n)
	  throw OutOfRangeException(where() + "LOAD_ARRAY: index");

	sp--;
	stack[sp-1] = static_cast<ArrayInstance*>(aref.v_object)->data[ind];
	pc++; return CONT;
      }

    case Opcodes::STORE_LS1: STORE_LS4(sint1); 
    case Opcodes::STORE_LU1: STORE_LS4(uint1);
    case Opcodes::STORE_LS2: STORE_LS4(sint2);
    case Opcodes::STORE_LU2: STORE_LS4(uint2);
    case Opcodes::STORE_LS4: STORE_LE(sint4);
    case Opcodes::STORE_LU4: STORE_LE(uint4);
    case Opcodes::STORE_LS8: STORE_LE(sint8);
    case Opcodes::STORE_LU8: STORE_LE(uint8);
    case Opcodes::STORE_LR4: STORE_LE(real4);
    case Opcodes::STORE_LR8: STORE_LE(real8);
    case Opcodes::STORE_LREF: STORE_LR;

    case Opcodes::STORE_PS1: STORE_PS4(sint1); 
    case Opcodes::STORE_PU1: STORE_PS4(uint1);
    case Opcodes::STORE_PS2: STORE_PS4(sint2);
    case Opcodes::STORE_PU2: STORE_PS4(uint2);
    case Opcodes::STORE_PS4: STORE_PE(sint4);
    case Opcodes::STORE_PU4: STORE_PE(uint4);
    case Opcodes::STORE_PS8: STORE_PE(sint8);
    case Opcodes::STORE_PU8: STORE_PE(uint8);
    case Opcodes::STORE_PR4: STORE_PE(real4);
    case Opcodes::STORE_PR8: STORE_PE(real8);
    case Opcodes::STORE_PREF: STORE_PR;      
      
    case Opcodes::STORE_SS1: STORE_SS4(sint1); 
    case Opcodes::STORE_SU1: STORE_SS4(uint1);
    case Opcodes::STORE_SS2: STORE_SS4(sint2);
    case Opcodes::STORE_SU2: STORE_SS4(uint2);
    case Opcodes::STORE_SS4: STORE_SE(sint4);
    case Opcodes::STORE_SU4: STORE_SE(uint4);
    case Opcodes::STORE_SS8: STORE_SE(sint8);
    case Opcodes::STORE_SU8: STORE_SE(uint8);
    case Opcodes::STORE_SR4: STORE_SE(real4);
    case Opcodes::STORE_SR8: STORE_SE(real8);
    case Opcodes::STORE_SREF: STORE_SR;
      
    case Opcodes::STORE_IS1: STORE_IS4(sint1); 
    case Opcodes::STORE_IU1: STORE_IS4(uint1);
    case Opcodes::STORE_IS2: STORE_IS4(sint2);
    case Opcodes::STORE_IU2: STORE_IS4(uint2);
    case Opcodes::STORE_IS4: STORE_IE(sint4);
    case Opcodes::STORE_IU4: STORE_IE(uint4);
    case Opcodes::STORE_IS8: STORE_IE(sint8);
    case Opcodes::STORE_IU8: STORE_IE(uint8);
    case Opcodes::STORE_IR4: STORE_IE(real4);
    case Opcodes::STORE_IR8: STORE_IE(real8);
    case Opcodes::STORE_IREF: STORE_IR;
      
    case Opcodes::STORE_AS1: STORE_AS4(sint1); 
    case Opcodes::STORE_AU1: STORE_AS4(uint1);
    case Opcodes::STORE_AS2: STORE_AS4(sint2);
    case Opcodes::STORE_AU2: STORE_AS4(uint2);
    case Opcodes::STORE_AS4: STORE_AE(sint4);
    case Opcodes::STORE_AU4: STORE_AE(uint4);
    case Opcodes::STORE_AS8: STORE_AE(sint8);
    case Opcodes::STORE_AU8: STORE_AE(uint8);
    case Opcodes::STORE_AR4: STORE_AE(real4);
    case Opcodes::STORE_AR8: STORE_AE(real8);
    case Opcodes::STORE_AREF: STORE_AR;
      
    case Opcodes::POP:
      --sp; ++pc; return CONT;
      
      /*
	conversions:

	all conversions create values of type u4,s4,u8,s8,r4, or r8:
	s1,u1,s2,u2 -> s4
	u4          -> u4
	s8          -> s8
	u8          -> u8
	r4          -> r4
	r8          -> r8
	
	real4 -> real8

	round to nearest real8 value (no exception)
	
	int -> real:

	round to nearest real value (no exception)

	real -> int:

	round to nearest int value; exception on overflow

	sintK -> sintL:

	K < L: copy (sign extend)
	K > L: throw exception if out of range, copy otherwise

	uintK -> sintL:

	K < L: copy (sign extend)
	K > L: throw excpetion if out of range, copy otherwise

	sintK -> uintL:

	if (sint value < 0) throw error
	if (sint value > max_uintL) throw error
	copy

	uintK -> uintL:

	if (uintK value > max_uintL) throw error
	copy
      */

    case Opcodes::CONV_S4S1: CONV_SI(sint4,sint1,sint4);
    case Opcodes::CONV_U4S1: CONV_UI(uint4,sint1,sint4);
    case Opcodes::CONV_S8S1: CONV_SI(sint8,sint1,sint4);
    case Opcodes::CONV_U8S1: CONV_UI(uint8,sint1,sint4); 
    case Opcodes::CONV_R4S1: CONV_RI4(real4,sint1,sint4);
    case Opcodes::CONV_R8S1: CONV_RI4(real8,sint1,sint4);
      
    case Opcodes::CONV_S4U1: CONV_SI(sint4,uint1,sint4); 
    case Opcodes::CONV_U4U1: CONV_UI(uint4,uint1,sint4); 
    case Opcodes::CONV_S8U1: CONV_SI(sint8,uint1,sint4); 
    case Opcodes::CONV_U8U1: CONV_UI(uint8,uint1,sint4); 
    case Opcodes::CONV_R4U1: CONV_RI4(real4,uint1,sint4);
    case Opcodes::CONV_R8U1: CONV_RI4(real8,uint1,sint4);
      
    case Opcodes::CONV_S4S2: CONV_SI(sint4,sint2,sint4); 
    case Opcodes::CONV_U4S2: CONV_UI(uint4,sint2,sint4); 
    case Opcodes::CONV_S8S2: CONV_SI(sint8,sint2,sint4); 
    case Opcodes::CONV_U8S2: CONV_UI(uint8,sint2,sint4); 
    case Opcodes::CONV_R4S2: CONV_RI4(real4,sint2,sint4);    
    case Opcodes::CONV_R8S2: CONV_RI4(real8,sint2,sint4);   
			     						   
    case Opcodes::CONV_S4U2: CONV_SI(sint4,uint2,sint4); 
    case Opcodes::CONV_U4U2: CONV_UI(uint4,uint2,sint4); 
    case Opcodes::CONV_S8U2: CONV_SI(sint8,uint2,sint4); 
    case Opcodes::CONV_U8U2: CONV_UI(uint8,uint2,sint4); 
    case Opcodes::CONV_R4U2: CONV_RI4(real4,uint2,sint4);
    case Opcodes::CONV_R8U2: CONV_RI4(real8,uint2,sint4);
       
    case Opcodes::CONV_U4S4: CONV_UI(uint4,sint4,sint4); 
    case Opcodes::CONV_S8S4: CONV_SI(sint8,sint4,sint4);
    case Opcodes::CONV_U8S4: CONV_UI(uint8,sint4,sint4); 
    case Opcodes::CONV_R4S4: CONV_RI4(real4,sint4,sint4);
    case Opcodes::CONV_R8S4: CONV_RI4(real8,sint4,sint4);
 
    case Opcodes::CONV_S4U4:
      {
	check_type(Type::TC_sint4);
	Value &v = stack[sp-1];
	if (v.v_sint4 < 0)
	  throw OutOfRangeException(where() + "conversion sint4 to uint4");
	v.cat = Type::TC_uint4;
	pc++; return CONT;
      }

    case Opcodes::CONV_S8U4: CONV_SI(sint8,uint4,uint4); 
    case Opcodes::CONV_U8U4: CONV_UI(uint8,uint4,uint4); 
    case Opcodes::CONV_R4U4: CONV_RI4(real4,uint4,uint4);
    case Opcodes::CONV_R8U4: CONV_RI4(real8,uint4,uint4);
      
    case Opcodes::CONV_S4S8: CONV_WIDEN(sint4,sint8); 
    case Opcodes::CONV_U4S8: CONV_WIDEN(uint4,sint8);
    case Opcodes::CONV_U8S8: CONV_UI(uint8,sint8,sint8); 
    case Opcodes::CONV_R4S8: ERR("CONV_R4S8 not implemented"); // fixme
    case Opcodes::CONV_R8S8: ERR("CONV_R8S8 not implemented"); // fixme

    case Opcodes::CONV_S4U8:
      {
	check_type(Type::TC_sint4);
	Value &v = stack[sp-1];
	if (v.v_sint4 < 0)	
	  throw OutOfRangeException(where() + "conversion sint4 to uint8");
	v.cat = Type::TC_uint8; v.v_uint8 = v.v_sint4;
	pc++; return CONT;
      }
      
    case Opcodes::CONV_U4U8: CONV_WIDEN(uint4,uint8); 
    case Opcodes::CONV_S8U8:
      {
	check_type(Type::TC_sint8);
	Value &v = stack[sp-1];
	if (v.v_sint8 < 0)	
	  throw OutOfRangeException(where() + "conversion sint8 to uint8");
	v.cat = Type::TC_uint8;
	pc++; return CONT;
      }

    case Opcodes::CONV_R4U8: ERR("CONV_R4U8 not implemented"); // fixme
    case Opcodes::CONV_R8U8: ERR("CONV_R8U8 not implemented"); // fixme

    case Opcodes::CONV_S4R4: CONV_WIDEN(sint4,real4);
    case Opcodes::CONV_U4R4: CONV_WIDEN(uint4,real4); 
    case Opcodes::CONV_S8R4: CONV_WIDEN(sint8,real4); 
    case Opcodes::CONV_U8R4: CONV_WIDEN(uint8,real4); 
    case Opcodes::CONV_R8R4: 

    case Opcodes::CONV_S4R8: CONV_WIDEN(sint4,real8);
    case Opcodes::CONV_U4R8: CONV_WIDEN(uint4,real8);
    case Opcodes::CONV_S8R8: CONV_WIDEN(sint8,real8);
    case Opcodes::CONV_U8R8: CONV_WIDEN(uint8,real8);
    case Opcodes::CONV_R4R8: CONV_WIDEN(real4,real8); 
      
      // binary operators
      
      // ---------- ADD ----------    

    case Opcodes::ADD_R4:
      BIN_OP(real4, +, real4);
    
    case Opcodes::ADD_SS4: // sint + sint = sint
      BIN_OP(sint4, +, sint4);
      
    case Opcodes::ADD_SU4: // sint + uint = uint !!!
      BIN_OP_CHANGE(sint4, +, uint4, uint4);
      
    case Opcodes::ADD_US4: // uint + sint = uint
      BIN_OP(uint4, +, sint4);
      
    case Opcodes::ADD_UU4: // uint + uint = uint
      BIN_OP(uint4, +, uint4);

    case Opcodes::ADD_SS8: // sint + sint = sint
      BIN_OP(sint8, +, sint8);
      
    case Opcodes::ADD_SU8: // sint + uint = uint !!!
      BIN_OP_CHANGE(sint8, +, uint8, uint8);
      
    case Opcodes::ADD_US8: // uint + sint = uint
      BIN_OP(uint8, +, sint8);
      
    case Opcodes::ADD_UU8: // uint + uint = uint
      BIN_OP(uint8, +, uint8);

      
      // ---------- SUB ----------

    case Opcodes::SUB_R4:
      BIN_OP(real4, -, real4);

    case Opcodes::SUB_SS4: // sint - sint = sint
      BIN_OP(sint4, -, sint4);

    case Opcodes::SUB_SU4: // sint - uint = sint
      BIN_OP(sint4, -, uint4);

    case Opcodes::SUB_US4: // uint - sint = uint
      BIN_OP(uint4, -, sint4);

    case Opcodes::SUB_UU4: // uint - uint = sint !!!
      BIN_OP_CHANGE(uint4, -, uint4, sint4);

    case Opcodes::SUB_SS8: // sint - sint = sint
      BIN_OP(sint8, -, sint8);

    case Opcodes::SUB_SU8: // sint - uint = sint
      BIN_OP(sint8, -, uint8);

    case Opcodes::SUB_US8: // uint - sint = uint
      BIN_OP(uint8, -, sint8);

    case Opcodes::SUB_UU8: // uint - uint = sint !!!
      BIN_OP_CHANGE(uint8, -, uint8, sint8);

      // ---------- MUL ----------    
    
    case Opcodes::MUL_R4:
      BIN_OP(real4, *, real4);

    case Opcodes::MUL_SS4: // sint * sint = sint
      BIN_OP(sint4, *, sint4);

    case Opcodes::MUL_SU4: // sint * uint = sint
      BIN_OP(sint4, *, uint4);

    case Opcodes::MUL_US4: // uint * sint = sint !!!
      BIN_OP_CHANGE(uint4, *, sint4, sint4);

    case Opcodes::MUL_UU4: // uint + uint = uint
      BIN_OP(uint4, *, uint4);

    case Opcodes::MUL_SS8: // sint * sint = sint
      BIN_OP(sint8, *, sint8);

    case Opcodes::MUL_SU8: // sint * uint = sint
      BIN_OP(sint8, *, uint8);

    case Opcodes::MUL_US8: // uint * sint = sint !!!
      BIN_OP_CHANGE(uint8, *, sint8, sint8);

    case Opcodes::MUL_UU8: // uint + uint = uint
      BIN_OP(uint8, *, uint8);


      // ---------- DIV ----------    
    
    case Opcodes::DIV_R4:
      if (stack[sp-1].v_real4 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(real4, /, real4);

    case Opcodes::DIV_SS4: // sint / sint = sint
      if (stack[sp-1].v_sint4 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(sint4, /, sint4);

    case Opcodes::DIV_SU4: // sint / uint = sint
      if (stack[sp-1].v_uint4 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(sint4, /, uint4);

    case Opcodes::DIV_US4: // uint / sint = sint !!!
      if (stack[sp-1].v_sint4 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP_CHANGE(uint4, /, sint4, sint4);

    case Opcodes::DIV_UU4: // uint / uint = uint
      if (stack[sp-1].v_uint4 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(uint4, /, uint4);

    case Opcodes::DIV_SS8: // sint / sint = sint
      if (stack[sp-1].v_sint8 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(sint8, /, sint8);

    case Opcodes::DIV_SU8: // sint / uint = sint
      if (stack[sp-1].v_uint8 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(sint8, /, uint8);

    case Opcodes::DIV_US8: // uint / sint = sint !!!
      if (stack[sp-1].v_sint8 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP_CHANGE(uint8, /, sint8, sint8);

    case Opcodes::DIV_UU8: // uint / uint = uint
      if (stack[sp-1].v_uint8 == 0)
	throw ParException(where() + "DIV: division by 0");
      BIN_OP(uint8, /, uint8);

      // ---------- MOD ----------    
    
    case Opcodes::MOD_SS4: // sint / sint = sint
      if (stack[sp-1].v_sint4 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(sint4, %, sint4);

    case Opcodes::MOD_SU4: // sint / uint = sint
      if (stack[sp-1].v_uint4 == 0) 
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(sint4, %, uint4);

    case Opcodes::MOD_US4: // uint / sint = sint !!!
      if (stack[sp-1].v_sint4 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP_CHANGE(uint4, %, sint4, sint4);

    case Opcodes::MOD_UU4: // uint / uint = uint
      if (stack[sp-1].v_uint4 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(uint4, %, uint4);

    case Opcodes::MOD_SS8: // sint / sint = sint
      if (stack[sp-1].v_sint8 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(sint8, %, sint8);

    case Opcodes::MOD_SU8: // sint / uint = sint
      if (stack[sp-1].v_uint8 == 0) 
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(sint8, %, uint8);

    case Opcodes::MOD_US8: // uint / sint = sint !!!
      if (stack[sp-1].v_sint8 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP_CHANGE(uint8, %, sint8, sint8);

    case Opcodes::MOD_UU8: // uint / uint = uint
      if (stack[sp-1].v_uint8 == 0)
	throw ParException(where() + "MOD: division by 0");
      BIN_OP(uint8, %, uint8);

      // ---------- AND ----------    

    case Opcodes::AND_I4: BIN_OP_INT4(&);
    case Opcodes::AND_I8: BIN_OP_INT8(&);
    case Opcodes::OR_I4:  BIN_OP_INT4(|);
    case Opcodes::OR_I8:  BIN_OP_INT8(|);
    case Opcodes::XOR_I4: BIN_OP_INT4(^);        
    case Opcodes::XOR_I8: BIN_OP_INT8(^);        

    case Opcodes::SHL_I4:

      // 2nd value can't be negative
      if (stack[sp-1].cat == Type::TC_sint4 && stack[sp-1].v_sint4 < 0)
	throw OutOfRangeException(where() + "SHL_I4: 2nd value < 0");

      BIN_OP_INT4(<<);

      // fixme: need to distinguish between signed/unsigned for SHR!

    case Opcodes::SHR_I4:

      // 2nd value can't be negative
      if (stack[sp-1].cat == Type::TC_sint4 && stack[sp-1].v_sint4 < 0)
	throw OutOfRangeException(where() + "SHR_I4: 2nd value < 0");

      BIN_OP_INT4(>>);

      
      // Relational operators (result sint4: 0 or 1)
      // correctly handle signed/unsigned comparisons
    
      // ---------- == ----------    
    
    case Opcodes::EQ_SS4: BIN_REL_OP_SAME(sint4, ==);
    case Opcodes::EQ_SU4: BIN_REL_OP_SU(sint4, ==, uint4, 0);
    case Opcodes::EQ_US4: BIN_REL_OP_US(uint4, ==, sint4, 0);
    case Opcodes::EQ_UU4: BIN_REL_OP_SAME_ADJUST(uint4, ==);
    case Opcodes::EQ_SS8: BIN_REL_OP_SAME(sint8, ==);
    case Opcodes::EQ_SU8: BIN_REL_OP_SU(sint8, ==, uint8, 0);
    case Opcodes::EQ_US8: BIN_REL_OP_US(uint8, ==, sint8, 0);
    case Opcodes::EQ_UU8: BIN_REL_OP_SAME_ADJUST(uint8, ==)

      // ---------- != ----------    
    
    case Opcodes::NEQ_SS4: BIN_REL_OP_SAME(sint4, !=);
    case Opcodes::NEQ_SU4: BIN_REL_OP_SU(sint4, !=, uint4, 1);
    case Opcodes::NEQ_US4: BIN_REL_OP_US(uint4, !=, sint4, 1);
    case Opcodes::NEQ_UU4: BIN_REL_OP_SAME_ADJUST(uint4, !=);
    case Opcodes::NEQ_SS8: BIN_REL_OP_SAME(sint8, !=);
    case Opcodes::NEQ_SU8: BIN_REL_OP_SU(sint8, !=, uint8, 1);
    case Opcodes::NEQ_US8: BIN_REL_OP_US(uint8, !=, sint8, 1);
    case Opcodes::NEQ_UU8: BIN_REL_OP_SAME_ADJUST(uint8, !=);
      
      // ---------- < ----------    
    
    case Opcodes::LT_SS4: BIN_REL_OP_SAME(sint4, <);
    case Opcodes::LT_SU4: BIN_REL_OP_SU(sint4, <, uint4, 1);
    case Opcodes::LT_US4: BIN_REL_OP_US(uint4, <, sint4, 0);
    case Opcodes::LT_UU4: BIN_REL_OP_SAME_ADJUST(uint4, <);
    case Opcodes::LT_SS8: BIN_REL_OP_SAME(sint8, <);
    case Opcodes::LT_SU8: BIN_REL_OP_SU(sint8, <, uint8, 1);
    case Opcodes::LT_US8: BIN_REL_OP_US(uint8, <, sint8, 0);
    case Opcodes::LT_UU8: BIN_REL_OP_SAME_ADJUST(uint8, <);

      // ---------- > ----------    
    
    case Opcodes::GT_SS4: BIN_REL_OP_SAME(sint4, >);
    case Opcodes::GT_SU4: BIN_REL_OP_SU(sint4, >, uint4, 0);
    case Opcodes::GT_US4: BIN_REL_OP_US(uint4, >, sint4, 1);
    case Opcodes::GT_UU4: BIN_REL_OP_SAME_ADJUST(uint4, >);
    case Opcodes::GT_SS8: BIN_REL_OP_SAME(sint8, >);
    case Opcodes::GT_SU8: BIN_REL_OP_SU(sint8, >, uint8, 0);
    case Opcodes::GT_US8: BIN_REL_OP_US(uint8, >, sint8, 1);
    case Opcodes::GT_UU8: BIN_REL_OP_SAME_ADJUST(uint8, >);

      // ---------- <= ----------    
    
    case Opcodes::LEQ_SS4: BIN_REL_OP_SAME(sint4, <=);
    case Opcodes::LEQ_SU4: BIN_REL_OP_SU(sint4, <=, uint4, 1);
    case Opcodes::LEQ_US4: BIN_REL_OP_US(uint4, <=, sint4, 0);
    case Opcodes::LEQ_UU4: BIN_REL_OP_SAME_ADJUST(uint4, <=);
    case Opcodes::LEQ_SS8: BIN_REL_OP_SAME(sint8, <=);
    case Opcodes::LEQ_SU8: BIN_REL_OP_SU(sint8, <=, uint8, 1);
    case Opcodes::LEQ_US8: BIN_REL_OP_US(uint8, <=, sint8, 0);
    case Opcodes::LEQ_UU8: BIN_REL_OP_SAME_ADJUST(uint8, <=);

      // ---------- >= ----------    
    
    case Opcodes::GEQ_SS4: BIN_REL_OP_SAME(sint4, >=);
    case Opcodes::GEQ_SU4: BIN_REL_OP_SU(sint4, >=, uint4, 0);
    case Opcodes::GEQ_US4: BIN_REL_OP_US(uint4, >=, sint4, 1);
    case Opcodes::GEQ_UU4: BIN_REL_OP_SAME_ADJUST(uint4, >=);
    case Opcodes::GEQ_SS8: BIN_REL_OP_SAME(sint8, >=);
    case Opcodes::GEQ_SU8: BIN_REL_OP_SU(sint8, >=, uint8, 0);
    case Opcodes::GEQ_US8: BIN_REL_OP_US(uint8, >=, sint8, 1);
    case Opcodes::GEQ_UU8: BIN_REL_OP_SAME_ADJUST(uint8, >=);


      // --------- unary operators -----------


      // --------- NEG ------------

    case Opcodes::NEG_S4: // -sint4 = sint4
      {
	check_type(Type::TC_sint4);
	Value &v = stack[sp-1];
        if (v.v_sint4 == min_sint4)
	  throw ParException(where() + "neg(>=max_sint4");
	v.v_sint4 = -v.v_sint4; pc++; return CONT;
      }
      
    case Opcodes::NEG_U4: // -uint4 = sint4
      {
	check_type(Type::TC_uint4);
	Value &v = stack[sp-1];
        if (v.v_uint4 > (uint4)max_sint4)
	  throw ParException(where() + "neg(>=max_sint4");
	v.v_sint4 = -v.v_uint4; v.cat = Type::TC_sint4; pc++; return CONT;
      }
    
    case Opcodes::NEG_R4: // -real4 = real4
      {
	check_type(Type::TC_real4);
	Value &v = stack[sp-1]; v.v_real4 = -v.v_real4; pc++; return CONT;
      }
    
    case Opcodes::NEG_R8: // -real8 = real8
      {
	check_type(Type::TC_real8);
	Value &v = stack[sp-1]; v.v_real8 = -v.v_real8; pc++; return CONT;
      }

      // --------- COMPL ------------
    
    case Opcodes::COMPL_S4: // ~sint4 = sint4
      {
	check_type(Type::TC_sint4);
	Value &v = stack[sp-1]; v.v_sint4 = ~v.v_sint4; pc++; return CONT;
      }
      
    case Opcodes::COMPL_U4: // ~uint4 = sint4
      {
	check_type(Type::TC_uint4);
	Value &v = stack[sp-1]; v.v_sint4 = ~v.v_uint4; pc++; return CONT;
      }
    
      // --------- INV ------------
    
    case Opcodes::INV_S4: // !sint4 = sint4
      {
	check_type(Type::TC_sint4);
	Value &v = stack[sp-1]; v.v_sint4 = !v.v_sint4; pc++; return CONT;
      }
      
    case Opcodes::INV_U4: // !uint4 = sint4
      {
	check_type(Type::TC_uint4); Value &v = stack[sp-1];
	v.v_sint4 = !v.v_uint4; v.cat = Type::TC_sint4; pc++; return CONT;
      }


      // --------- MISC ----------
    
    case Opcodes::PRINT :
      cout << "top= ";
      prog.write_value(stack[sp-1], cout);
      cout << endl;
      sp--; pc++; return CONT;

    case Opcodes::EXIT :
      cout << "exit" << endl; exit(0);
      return CONT;

    default :
      errstr << "unknown opcode! pc=" << pc << " " << (sint4)code[pc] << endl;
      exit(-1);
      pc++;
      return END;
    }

    ERR("here?");
    return END;
  }


  void Context::write_stack(ostream &os) const
  {
    os << "stack[" << sp << " deep] : " << endl;
    
    if (sp >= 8) os << "..." << endl;

    for (uint4 i=(sp<8)? 0 :(sp-8); i < sp; ++i) {
      form(os, "%d", i);
      prog.write_value(stack[i], os);
      os << endl;
    }
  }


  Context::StepResult Context::run(sint4 n)
  {
    StepResult res;

    if (n < 0) {

      // only stop at END

      FOREVER {
	res = step();
	if (res == END) break;
      }
      return res;
    

    } else if (n > 0) {

      // run n steps, stop at END and YIELD

      res = END;
      while (n--) {
	res = step();
	if (res != CONT) break;
      }
      return res;

    } else {

      // stop at END or YIELD
    
      FOREVER {
	res = step();
	if (res != CONT) break;
      }
      return res;
    } 
  }
}
