#ifndef GameStateModule_H
#define GameStateModule_H

// $Id: GameStateModule.H,v 1.14 2006/06/19 13:11:12 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "ClientModule.H"

class Game;
class GameChanges;
class ActionChanges;
class Options;
class TCP_Client;
class UniRandSrc;

/** Wrapper for a Game's connection to the server.  Calls functions that
    send and receive game views, then calls event handlers to deal with
    any changes.
*/

class GameStateModule : public ClientModule
{
public:

  struct Options
  {
    std::string host;
    sint4 port;
    sint4 dmin;  // random lag (msec) minimum
    sint4 dmax;  // random lag (msec) maximum
    sint4 smin;  // random sleep (msec) minimum
    sint4 smax;  // random sleep (msec) maximum
    sint4 clev;  // compression level
    bool  act;   // generate actions
    bool  nfog;
    std::string save_file;
    std::string load_file;
    std::string id; // client authentication string sent to server

    Options();
    static void add();
  };
  
  // messages sent out when the client receives data from the server
  static const sint4 FROM;            // identifies GameStateModule as the sender
  static const sint4 FINISHED_MSG;    // game finished
  static const sint4 STOP_MSG;        // client dead (game end, i/o error)
  static const sint4 READ_ERROR_MSG;  // socket error
  static const sint4 VIEW_MSG;        // view received
  
  GameStateModule(const GameStateModule::Options &o);
  virtual ~GameStateModule();

  // connect to server and wait for game to begin
  // returns true iff connected
  bool connect();

  // handle next server event
  // returns true in case of a server event
  bool recv_view();

  // get raw message
  const std::string &get_server_msg() const { return server_msg; }

  // access to game
  Game &get_game() { return *game; }
  const Game &get_game() const { return *game; }

  // access to changes in last frame
  const GameChanges &get_changes() const { return *changes; }
  const ActionChanges &get_action_changes() const { return *action_changes; }

  // send all created actions to server
  // <0:  error
  // >=0: number of messages in buffer
  sint4 send_actions();

  bool error() const { return client_died; }
  bool exit() const { return quit; }
  
protected:
  
  enum State { ST_IDLE, ST_RECV_VIEW };

  Options opt;
  State state;            // client state
  TCP_Client  *tcp_client;
  UniRandSrc *rand_delay; // network delay random source, =0: no delay
  UniRandSrc *rand_sleep; // sleep andom source, =0: no sleep
  Game *game;             // game state
  GameChanges *changes;   // world changes during last simulation frame
  ActionChanges *action_changes;
  sint4 last_action_msg_sent;
  bool client_died;       // read error
  bool quit;              // read error or game end
  std::string server_msg; // message sent from server

private:

  // not implemented yet
  GameStateModule(const GameStateModule &) : ClientModule() { ERR("what"); }
  GameStateModule &operator=(const GameStateModule &) { ERR("what"); }  
};

#endif
