/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Element.H"
#include "CompAction.H"

Element::~Element() { if (comp_act) delete comp_act; }

/*
  Update a running Adler-32 checksum with the bytes buf[0..len-1]
  and return the updated checksum.  The Adler-32 checksum should be
  initialized to 1.
  
  Usage example:
  
  unsigned long adler = 1L;
  
  while (read_buffer(buffer, length) != EOF) {
  adler = update_adler32(adler, buffer, length);
  }
  if (adler != original_adler) error();
*/


static uint4 update_adler32(uint4 adler, uint1 *buf, sint4 len)
{
  uint4 s1 = adler & 0xffff;
  uint4 s2 = (adler >> 16) & 0xffff;
  const uint4 BASE = 65521; // largest prime smaller than 65536

  FORS (i, len) {
    s1 += buf[i]; if (s1 >= BASE) s1 -= BASE;
    s2 += s1;     if (s2 >= BASE) s2 -= BASE;
  }
  return (s2 << 16) + s1;
}

// Return the adler32 of the bytes buf[0..len-1]

uint4 adler32(uint1 *buf, sint4 len)
{
  return update_adler32(1, buf, len);
}


uint4 Prop::checksum(uint4 id)
{
  uint4 tid = sv_id;
  sv_id = id;

  uint4 cs = adler32((uint1*)&PROP_FIRST,
		     ((uint1*)&PROP_LAST)+sizeof(PROP_LAST)-((uint1*)&PROP_FIRST));
  sv_id = tid;
  return cs;
}
